/*
								+----------------------------------+
								|                                  |
								|   ***  3D modelling tools  ***   |
								|                                  |
								|   Copyright  -tHE SWINe- 2005   |
								|                                  |
								|           Geometry.cpp           |
								|                                  |
								+----------------------------------+
*/

#include <crtdbg.h>
#include <stdlib.h>
#include <windows.h>
#include <gl/gl.h>

#if defined(_MSC_VER) && !defined(__MWERKS__) // msvc
#pragma warning(4:4786)
// switch-off warning 4786 (too long class name truncated to 255 chars in debug info)
#endif

#include <vector>
#include <stack>
#include <math.h>
#include "glext.h"
#include "vector.h"
#include "geometry.h"

/*
 *								=== TRefVertex ===
 */

TRefVertex TRefVertex::Lerp(float t, const TRefVertex &r_right) const
{
	TVertex t_temp;

	_ASSERTE(p_vertex);
	_ASSERTE(p_geom);

	t_temp = p_vertex->Lerp(t, *r_right.p_vertex);
	// we have lerp written already

	TRefVertex t_vertex;

	t_vertex = p_geom->t_AddRefVertex(t_temp);
	if(!t_vertex.p_vertex)
		throw CCantInsertVertex_Exception();

	return t_vertex;
}

/*
 *								=== ~TRefVertex ===
 */

/*
 *								=== CGeometry ===
 */

int CGeometry::m_n_surface_space = 0;

CGeometry::CGeometry(int n_polys_per_hash_record, int n_vertex_hash_table_size,
	float f_vertex_cell_overlap_ratio, float f_min_vertex_cell_overlap)
		:m_n_polys_per_hash_record(n_polys_per_hash_record),
		 m_n_vertex_hash_table_size(n_vertex_hash_table_size),
		 m_f_vertex_cell_overlap_ratio(f_vertex_cell_overlap_ratio),
		 m_f_min_vertex_cell_overlap(f_min_vertex_cell_overlap),
		 m_p_vertex_hash_table(0),
		 m_p_poly_hash_table(0),
		 m_b_bounding_box_dirty_flag(false),
		 m_b_poly_hash_dirty_flag(false),
		 m_b_vertex_hash_dirty_flag(false),
		 m_b_vertex_normal_smooth_group_dirty_flag(false),
		 m_b_vertex_normal_smooth_all_dirty_flag(false),
		 m_b_poly_neighbours_dirty_flag(false)
{
}

CGeometry::~CGeometry()
{
	Delete();
}

Vector3f CGeometry::v_Bound_Min()
{
	UpdateBoundingBox();
	return m_v_bounding_box[0];
}

Vector3f CGeometry::v_Bound_Max()
{
	UpdateBoundingBox();
	return m_v_bounding_box[1];
}

void CGeometry::UpdateBoundingBox()
{
	if(!m_b_bounding_box_dirty_flag)
		return;

	_ASSERTE(m_b_vertex_hash_dirty_flag);
	// vertex hash is dependent on bounding box dimensions ...

	if(m_vertex_list.size())
		m_v_bounding_box[0] = m_v_bounding_box[1] = m_vertex_list[0]->v_pos;
	else
		m_v_bounding_box[0] = m_v_bounding_box[1] = Vector3f(0, 0, 0);
	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin() + 1;
	   p_vertex < m_vertex_list.end(); p_vertex ++) {
		for(int i = 0; i < 3; i ++) {
			if((*p_vertex)->v_pos[i] < m_v_bounding_box[0][i])
				m_v_bounding_box[0][i] = (*p_vertex)->v_pos[i];
			if((*p_vertex)->v_pos[i] > m_v_bounding_box[1][i])
				m_v_bounding_box[1][i] = (*p_vertex)->v_pos[i];
		}
	}

	m_b_bounding_box_dirty_flag = false;
}

// update vertices if necessary, clear flags if successful
bool CGeometry::UpdateVertexHash()
{
	if(m_b_vertex_hash_dirty_flag) {
		UpdateBoundingBox();
		// need to update bounding box to calc new size of hash table as it's spatially bound

		if(m_p_vertex_hash_table) {
			int n_size3 = m_n_vertex_hash_table_size *
				m_n_vertex_hash_table_size * m_n_vertex_hash_table_size;
			for(std::vector<TVertex*> **p_hash = m_p_vertex_hash_table, **p_end =
			   m_p_vertex_hash_table + n_size3; p_hash < p_end;)
				delete *(p_hash ++);
			delete[] m_p_vertex_hash_table;
			m_p_vertex_hash_table = 0;
		}
		m_f_vertex_hash_table_cell_size = (m_v_bounding_box[0] - m_v_bounding_box[1]).f_Length() /
			(m_n_vertex_hash_table_size * 1.732f); // sqrt(3)
		m_f_vertex_hash_table_cell_size = (m_f_vertex_hash_table_cell_size >
			2 * m_f_min_vertex_cell_overlap)? m_f_vertex_hash_table_cell_size :
			2 * m_f_min_vertex_cell_overlap;
		if(!(m_p_vertex_hash_table = p_HashVertices(m_n_vertex_hash_table_size,
			m_f_vertex_hash_table_cell_size, (m_f_vertex_hash_table_cell_size *
			m_f_vertex_cell_overlap_ratio > m_f_min_vertex_cell_overlap)?
			m_f_vertex_hash_table_cell_size * m_f_vertex_cell_overlap_ratio :
			m_f_min_vertex_cell_overlap)))
			return false;
	}

	m_b_vertex_hash_dirty_flag = false;

	return true;
}

// update polygons if necessary clear flags if successful
bool CGeometry::UpdatePolygonHash()
{
	if(m_b_poly_hash_dirty_flag) {
		if(m_p_poly_hash_table) {
			for(std::vector<int> **p_hash = m_p_poly_hash_table, **p_end =
			   m_p_poly_hash_table + m_n_poly_hash_table_size; p_hash < p_end;)
				delete *(p_hash ++);
			delete[] m_p_poly_hash_table;
			m_p_poly_hash_table = 0;
		}
		m_n_poly_hash_table_size = 1 + m_poly_list.size() / m_n_polys_per_hash_record;
		if(!(m_p_poly_hash_table = p_HashPolygons(m_n_poly_hash_table_size)))
			return false;
	}

	m_b_poly_hash_dirty_flag = false;

	return true;
}

bool CGeometry::UpdatePolygonNeighbours()
{
	if(m_b_poly_neighbours_dirty_flag) {
		if(!FindNeighbours())
			return false;
	}

	m_b_poly_neighbours_dirty_flag = false;

	return true;
}

bool CGeometry::UpdateVertexNormals_SmoothAll()
{
	if(m_b_vertex_normal_smooth_all_dirty_flag) {
		if(!CalcVertexNormals(true))
			return false;
	}

	m_b_vertex_normal_smooth_all_dirty_flag = false;
	m_b_vertex_normal_smooth_group_dirty_flag = true;

	return true;
}

bool CGeometry::UpdateVertexNormals_SmoothGroups()
{
	if(m_b_vertex_normal_smooth_group_dirty_flag) {
		if(!CalcVertexNormals(false))
			return false;
	}

	m_b_vertex_normal_smooth_all_dirty_flag = true;
	m_b_vertex_normal_smooth_group_dirty_flag = false;

	return true;
}

const std::vector<int> *CGeometry::p_PolygonNeighbours(int n_index, int n_edge)
{
	if(!UpdatePolygonNeighbours())
		return 0;
	return m_poly_list[n_index].p_neighbour_index_list[n_edge];
}

// delete polygons
void CGeometry::DeletePolygons(int n_index, int n_count)
{
	if(!n_count)
		return;

	if(!m_b_poly_hash_dirty_flag) {
		for(std::vector<int> **p_hash = m_p_poly_hash_table, **p_end =
		   m_p_poly_hash_table + m_n_poly_hash_table_size; p_hash < p_end; p_hash ++) {
			for(std::vector<int>::iterator p_poly2 = (*p_hash)->begin();
			   p_poly2 < (*p_hash)->end(); p_poly2 ++) {
				if(*p_poly2 >= n_index && *p_poly2 < n_index + n_count) {
					(*p_hash)->erase(&*(p_poly2 --));
					n_count --;
				} else if(*p_poly2 > n_index)
					(*p_poly2) --;
			}
		}
	}
	// keep poly hash up-to-date

	//m_b_poly_hash_dirty_flag = true;
	m_b_vertex_normal_smooth_group_dirty_flag = true;
	m_b_vertex_normal_smooth_all_dirty_flag = true;
	m_b_poly_neighbours_dirty_flag = true;

	m_poly_list.erase(&m_poly_list[n_index], &m_poly_list[n_index + n_count]);
}

bool CGeometry::DeleteVertices(int n_index, int n_count)
{
	if(!n_count)
		return true;

	if(!UpdatePolygonHash())
		return false;
	// create polygon hash table

	for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin() + n_index, p_end =
	   m_vertex_list.begin() + n_index + n_count; p_vertex < p_end; p_vertex ++) {
		std::vector<int> *p_hash = m_p_poly_hash_table[
			n_HashFunction(*p_vertex, m_n_poly_hash_table_size)];
		for(std::vector<int>::const_iterator p_poly = p_hash->begin();
		   p_poly < p_hash->end(); p_poly ++) {
			int n_poly_index;
			DeletePolygons(n_poly_index = *p_poly, 1);
			// delete polygon

			for(std::vector<int> **p_hash = m_p_poly_hash_table, **p_end =
			   m_p_poly_hash_table + m_n_poly_hash_table_size; p_hash < p_end; p_hash ++) {
				for(std::vector<int>::iterator p_poly2 = (*p_hash)->begin();
				   p_poly2 < (*p_hash)->end(); p_poly2 ++) {
					if(*p_poly2 == n_poly_index)
						(*p_hash)->erase(p_poly2 --);
					else if(*p_poly2 > n_poly_index)
						(*p_poly2) --;
				}
			}
			// update hash table so indices point to right polys
			// (can be slow - doing it per every deleted poly)
		}

		delete *p_vertex;
		// delete vertex
	}
	// delete polygons, referencing this vertex

	m_vertex_list.erase(&m_vertex_list[n_index], &m_vertex_list[n_index + n_count]);
	// delete vertex pointers

	m_b_bounding_box_dirty_flag = true;
	m_b_poly_hash_dirty_flag = false; // f_ixme - i'm keeping it up-to-date ... or not? yes. i'm
	m_b_vertex_hash_dirty_flag = true;
	m_b_vertex_normal_smooth_group_dirty_flag = true;
	m_b_vertex_normal_smooth_all_dirty_flag = true;
	m_b_poly_neighbours_dirty_flag = true;
	// set dirty flags

	return true;
}

/*
 *	void CGeometry::Delete()
 *		- delete contents of geometry structure
 */
void CGeometry::Delete()
{
	if(m_p_poly_hash_table) {
		for(std::vector<int> **p_hash = m_p_poly_hash_table, **p_end =
		   m_p_poly_hash_table + m_n_poly_hash_table_size; p_hash < p_end;)
			delete *(p_hash ++);
		delete[] m_p_poly_hash_table;
		m_p_poly_hash_table = 0;
	}
	if(m_p_vertex_hash_table) {
		int n_size3 = m_n_vertex_hash_table_size *
			m_n_vertex_hash_table_size * m_n_vertex_hash_table_size;
		for(std::vector<TVertex*> **p_hash = m_p_vertex_hash_table, **p_end =
		   m_p_vertex_hash_table + n_size3; p_hash < p_end;)
			delete *(p_hash ++);
		delete[] m_p_vertex_hash_table;
		m_p_vertex_hash_table = 0;
	}
	for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		delete *(p_vertex ++);
	m_vertex_list.clear();
	for(std::vector<TPolygon>::iterator p_poly = m_poly_list.begin();
	   p_poly < m_poly_list.end();)
		delete (*(p_poly ++)).p_poly;
	m_poly_list.clear();
	m_n_surface_space = 0; // we don't get so high numbers

	m_b_bounding_box_dirty_flag = true;
	m_b_poly_hash_dirty_flag = true;
	m_b_vertex_hash_dirty_flag = true;
	m_b_vertex_normal_smooth_group_dirty_flag = true;
	m_b_vertex_normal_smooth_all_dirty_flag = true;
	m_b_poly_neighbours_dirty_flag = true;
}

/*
 *	TVertex *CGeometry::p_AddVertex(TVertex t_vertex)
 *		- add new vertex (doesn't compare with existing vertices)
 */
TVertex *CGeometry::p_AddVertex(TVertex t_vertex)
{
	TVertex *p_new;
	if(!(p_new = new TVertex))
		return 0;
	*p_new = t_vertex;
	m_vertex_list.push_back(p_new);
	if(!m_vertex_list.size()) {
		Delete();
		delete p_new;
		return 0;
	}

	for(int i = 0; i < 3; i ++) {
		if(t_vertex.v_pos[i] < m_v_bounding_box[0][i] ||
		   t_vertex.v_pos[i] > m_v_bounding_box[1][i]) {
			if(!m_b_bounding_box_dirty_flag) {
				if(t_vertex.v_pos[i] < m_v_bounding_box[0][i])
					m_v_bounding_box[0][i] = t_vertex.v_pos[i];
				if(t_vertex.v_pos[i] > m_v_bounding_box[1][i])
					m_v_bounding_box[1][i] = t_vertex.v_pos[i];
			}
		}
	}
	// see if we extend bounding box ... in case we did, we invalidate
	// vertex hash table, based on it's dimensions

	if(!m_b_vertex_hash_dirty_flag && m_p_vertex_hash_table) {
		int n_index[8];
		int n_index_num = 0;
		float f_vertex_hash_table_cell_overlap = (m_f_vertex_hash_table_cell_size *
			m_f_vertex_cell_overlap_ratio > m_f_min_vertex_cell_overlap)?
			m_f_vertex_hash_table_cell_size * m_f_vertex_cell_overlap_ratio :
			m_f_min_vertex_cell_overlap;
		for(int i = 0; i < 8; i ++) {
			Vector3f v_shift((i & 1)? 0 : f_vertex_hash_table_cell_overlap,
				((i >> 1) & 1)? 0 : f_vertex_hash_table_cell_overlap,
				(i >> 2)? 0 : f_vertex_hash_table_cell_overlap);
			int n_hash = n_HashFunction(t_vertex.v_pos + v_shift,
				m_n_vertex_hash_table_size, m_f_vertex_hash_table_cell_size);
			for(int *p_index = n_index; p_index < n_index + n_index_num;) {
				if(*p_index ++ == n_hash) {
					n_hash = -1;
					break;
				}
			}
			if(n_hash != -1)
				n_index[n_index_num ++] = n_hash;
		}
		// calc indices ...

		for(int *p_index = n_index; p_index < n_index + n_index_num;) {
			m_p_vertex_hash_table[*p_index]->push_back(p_new);
			if(!m_p_vertex_hash_table[*p_index ++]->size()) {
				delete p_new;
				m_b_vertex_hash_dirty_flag = true;
				return false;
			}
		}
	} else
		m_b_vertex_hash_dirty_flag = true;
	// in case we already created hash table, just add vertex inside

	return p_new;
}

/*
 *	TRefVertex CGeometry::t_AddRefVertex(TVertex t_vertex)
 *		- add reference vertex to t_vertex (add t_vertex then create vertex,
 *		  referencing it, which is returned)
 */
TRefVertex CGeometry::t_AddRefVertex(TVertex t_vertex)
{
	TRefVertex t_ref_vertex;

	t_ref_vertex.p_geom = this;
	t_ref_vertex.p_vertex = p_AddVertex(t_vertex);

	return t_ref_vertex;
}

class COpenGLFormatTable {
public:
	/*
	 *	static Vector4i v_Color_Indices(GLenum n_color_format;)
	 *		- return vector of four integers, containing component indices of given color type
	 *		- if component isn't present, index is -1 (GL_RED = Vector4i(0, -1, -1, -1))
	 *		- supported color formats are GL_RED, GL_GREEN, GL_BLUE, GL_ALPHA, GL_RGB,
	 *		  GL_RGBA,GL_BGR, GL_BGRA, GL_LUMINANCE, GL_LUMINANCE_ALPHA, GL_INTENSITY
	 *		- for unsupported format returns Vector4i(-1, -1, -1, -1)
	 */
	static Vector4i v_Color_Indices(GLenum n_color_format)
	{
		switch(n_color_format) {
		case GL_RED:
			return Vector4i(0, -1, -1, -1);
			break;
		case GL_GREEN:
			return Vector4i(-1, 0, -1, -1);
			break;
		case GL_BLUE:
			return Vector4i(-1, -1, 0, -1);
			break;
		case GL_ALPHA:
			return Vector4i(-1, -1, -1, 0);
			break;
		case GL_RGB:
			return Vector4i(0, 1, 2, -1);
			break;
		case GL_RGBA:
			return Vector4i(1, 2, 3, 0);
			break;
		case GL_BGR:
			return Vector4i(2, 1, 0, -1);
			break;
		case GL_BGRA:
			return Vector4i(3, 2, 1, 0);
			break;
		case GL_LUMINANCE:
			return Vector4i(0, 0, 0, -1);
			break;
		case GL_LUMINANCE_ALPHA:
			return Vector4i(1, 1, 1, 0);
			break;
		case GL_INTENSITY:
			return Vector4i(0, 0, 0, 0);
			break;
		}
		return Vector4i(-1, -1, -1, -1);
	}

	struct TVertexFormat {
		GLenum n_format;
		int n_position_offset;
		int n_color_offset;
		int n_color_format;
		int n_color_type;
		int n_texcoord_offset;
		int n_texcoord_num;
		int n_texcoord_type;
		int n_normal_offset;
		int n_stride;
	};

	/*
	 *	TVertexFormat t_Vertex_Format(GLenum n_vertex_format)
	 *		- returns structure, describing requested OpenGL vertex format
	 *		- if it fails to find requested format, returns TVertexFormat with n_format set to 0
	 *		- supported vertex formats are  GL_V2F, GL_V3F, GL_C4UB_V2F, GL_C4UB_V3F, GL_C3F_V3F,
	 *		  GL_N3F_V3F, GL_C4F_N3F_V3F, GL_T2F_V3F, GL_T4F_V4F, GL_T2F_C4UB_V3F, GL_T2F_C3F_V3F,
	 *		  GL_T2F_N3F_V3F, GL_T2F_C4F_N3F_V3F, GL_T4F_C4F_N3F_V4F
	 */
	static TVertexFormat t_Vertex_Format(GLenum n_vertex_format)
	{
		TVertexFormat p_vf_list[] = {
			{GL_V2F, 0, -1, 0, 0, -1, 0, 0, -1, 2 * sizeof(GLfloat)},
			{GL_V3F, 0, -1, 0, 0, -1, 0, 0, -1, 3 * sizeof(GLfloat)},
			{GL_C4UB_V2F, 4 * sizeof(GLubyte), 0, GL_RGBA, GL_UNSIGNED_BYTE, -1, 0, 0, -1, 2 * sizeof(GLfloat) + 4 * sizeof(GLubyte)},
			{GL_C4UB_V3F, 4 * sizeof(GLubyte), 0, GL_RGBA, GL_UNSIGNED_BYTE, -1, 0, 0, -1, 3 * sizeof(GLfloat) + 4 * sizeof(GLubyte)},
			{GL_C3F_V3F, 3 * sizeof(GLfloat), 0, GL_RGB, GL_FLOAT, -1, 0, 0, -1, 6 * sizeof(GLfloat)},
			{GL_N3F_V3F, 3 * sizeof(GLfloat), -1, 0, 0, -1, 0, 0, 0, 6 * sizeof(GLfloat)},
			{GL_C4F_N3F_V3F, 7 * sizeof(GLfloat), 0, GL_RGBA, GL_FLOAT, -1, 0, 0, 4 * sizeof(GLfloat), 10 * sizeof(GLfloat)},
			{GL_T2F_V3F, 2 * sizeof(GLfloat), -1, 0, 0, 0, 2, GL_FLOAT, -1, 5 * sizeof(GLfloat)},
			{GL_T4F_V4F, 4 * sizeof(GLfloat), -1, 0, 0, 0, 4, GL_FLOAT, -1, 8 * sizeof(GLfloat)},
			{GL_T2F_C4UB_V3F, 4 * sizeof(GLubyte) + 2 * sizeof(GLfloat), 2 * sizeof(GLfloat), GL_RGBA, GL_UNSIGNED_BYTE, 0, 2, GL_FLOAT, -1, 5 * sizeof(GLfloat) + 4 * sizeof(GLubyte)},
			{GL_T2F_C3F_V3F, 5 * sizeof(GLfloat), 2 * sizeof(GLfloat), GL_RGB, GL_FLOAT, 0, 2, GL_FLOAT, -1, 8 * sizeof(GLfloat)},
			{GL_T2F_N3F_V3F, 5 * sizeof(GLfloat), -1, 0, 0, 0, 2, GL_FLOAT, 2 * sizeof(GLfloat), 8 * sizeof(GLfloat)},
			{GL_T2F_C4F_N3F_V3F, 5 * sizeof(GLfloat) + 4 * sizeof(GLubyte), 2 * sizeof(GLfloat), GL_RGBA, GL_FLOAT, 0, 2, GL_FLOAT, 2 * sizeof(GLfloat) + 4 * sizeof(GLubyte), 8 * sizeof(GLfloat) + 4 * sizeof(GLubyte)},
			{GL_T4F_C4F_N3F_V4F, 11 * sizeof(GLfloat), 4 * sizeof(GLfloat), GL_RGBA, GL_FLOAT, 0, 4, GL_FLOAT, 8 * sizeof(GLfloat), 15 * sizeof(GLfloat)}
		};
		// todo - test those constants, or at least check them

		for(int i = 0; i < sizeof(p_vf_list) / sizeof(p_vf_list[0]); i ++) {
			if(p_vf_list[i].n_format == n_vertex_format)
				return p_vf_list[i];
		}
		// try to find requested format

		TVertexFormat t_empty_format = {0};
		return t_empty_format;
	}
};

/*
 *	bool CGeometry::AddVertices(int n_stride, const void *p_vertex, int n_vertex_num,
 *		int n_position_offset, int n_texcoord_offset, int n_texcoord_num, GLenum n_texcoord_type,
 *		int n_normal_offset, int n_color_offset, GLenum n_color_format, int n_color_type,
 *		bool b_2d_vertices)
 *		- add vertices of certain format, the new vertices are added to back of m_vertex_list
 *		- n_*_offset is component offset in vertex structure (-1 = component not present;
 *		  positions must be always present)
 *		- positions and normals are always 3-component floats
 *		- n_texcoord_num is number of texture coordinates, it can be number
 *		  from 0 trough 4 * max_Vertex_Texcoords
 *		- n_color_format is format of color, supported formats are: GL_RED, GL_GREEN, GL_BLUE,
 *		  GL_ALPHA, GL_RGB, GL_RGBA, GL_BGR, GL_BGRA, GL_LUMINANCE, GL_LUMINANCE_ALPHA, GL_INTENSITY
 *		- color components that aren't specified are set to 0, unspecified alpha is set to 1
 *		- supported texcoord and color types include: GL_BYTE, GL_UNSIGNED_BYTE, GL_SHORT,
 *		  GL_UNSIGNED_SHORT, GL_INT, GL_UNSIGNED_INT, 
 */
bool CGeometry::AddVertices(const void *p_vertex, int n_vertex_num, int n_stride,
	int n_position_offset, int n_texcoord_offset, int n_texcoord_num, GLenum n_texcoord_type,
	int n_normal_offset, int n_color_offset, GLenum n_color_format, GLenum n_color_type,
	bool b_2d_vertices)
{
	if(n_position_offset == -1 || n_texcoord_num > 4 * max_Vertex_Texcoords)
		return false;

	Vector4i v_component_indices;
	if(n_color_offset != -1) {
		v_component_indices = COpenGLFormatTable::v_Color_Indices(n_color_format);
		if(v_component_indices == Vector4i(-1, -1, -1, -1))
			return false;
	}
	// recognize color format
	
	if(n_vertex_num) {
		m_b_vertex_normal_smooth_group_dirty_flag = true;
		m_b_vertex_normal_smooth_all_dirty_flag = true;
	}
	// invalidate normals on new vertices

	for(void *p_end = (void*)((int)p_vertex + n_stride * n_vertex_num); p_vertex < p_end;
	   p_vertex = (void*)((int)p_vertex + n_stride)) {
		TVertex t_vertex;
		t_vertex.Clear_ExtraFields();

		if(b_2d_vertices) {
			t_vertex.v_pos = Vector3f(
				((float*)((int)p_vertex + n_position_offset))[0],
				((float*)((int)p_vertex + n_position_offset))[1], 0);
		} else {
			t_vertex.v_pos = Vector3f(
				((float*)((int)p_vertex + n_position_offset))[0],
				((float*)((int)p_vertex + n_position_offset))[1],
				((float*)((int)p_vertex + n_position_offset))[2]);
		}
		// set position

		if(n_normal_offset != -1) {
			t_vertex.v_normal = Vector3f(
				((float*)((int)p_vertex + n_normal_offset))[0],
				((float*)((int)p_vertex + n_normal_offset))[1],
				((float*)((int)p_vertex + n_normal_offset))[2]);
		} else
			t_vertex.v_normal = Vector3f(0, 0, 0);
		// set normal

		if(n_color_offset != -1) {
			switch(n_color_type) {
			case GL_BYTE:
				for(int j = 0; j < 4; j ++) {
					t_vertex.v_color[j] = (v_component_indices[j] == -1)? (float)(j == 3) :
						(float)(((GLbyte*)((int)p_vertex +
						n_color_offset))[v_component_indices[j]]) / 127;
				}
				break;
			case GL_UNSIGNED_BYTE:
				for(int j = 0; j < 4; j ++) {
					t_vertex.v_color[j] = (v_component_indices[j] == -1)? (float)(j == 3) :
						(float)(((GLubyte*)((int)p_vertex +
						n_color_offset))[v_component_indices[j]]) / 255;
				}
				break;
			case GL_SHORT:
				for(int j = 0; j < 4; j ++) {
					t_vertex.v_color[j] = (v_component_indices[j] == -1)? (float)(j == 3) :
						(float)(((GLshort*)((int)p_vertex +
						n_color_offset))[v_component_indices[j]]) / 32767;
				}
				break;
			case GL_UNSIGNED_SHORT:
				for(int j = 0; j < 4; j ++) {
					t_vertex.v_color[j] = (v_component_indices[j] == -1)? (float)(j == 3) :
						(float)(((GLushort*)((int)p_vertex +
						n_color_offset))[v_component_indices[j]]) / 65535;
				}
				break;
			case GL_INT:
				for(int j = 0; j < 4; j ++) {
					t_vertex.v_color[j] = (v_component_indices[j] == -1)? (float)(j == 3) :
						(float)(((GLint*)((int)p_vertex +
						n_color_offset))[v_component_indices[j]]) / 2147483647;
				}
				break;
			case GL_UNSIGNED_INT:
				for(int j = 0; j < 4; j ++) {
					t_vertex.v_color[j] = (v_component_indices[j] == -1)? (float)(j == 3) :
						(float)(((GLuint*)((int)p_vertex +
						n_color_offset))[v_component_indices[j]]) / 4294967295UL;
				}
				break;
			case GL_FLOAT:
				for(int j = 0; j < 4; j ++) {
					t_vertex.v_color[j] = (v_component_indices[j] == -1)? (float)(j == 3) :
						/*(float)*/(((GLfloat*)((int)p_vertex +
						n_color_offset))[v_component_indices[j]]);
				}
				break;
			default:
				return false;
			}
			// gather colors
		} else
			t_vertex.v_color = Vector4f(0, 0, 0, 1);
		// set color

		if(n_texcoord_offset != -1 && n_texcoord_num) {
			switch(n_texcoord_type) {
			case GL_BYTE:
				for(int j = 0; j < (n_texcoord_num + 3) / 4; j ++) {
					t_vertex.v_texcoord[j / 4][j % 4] = (j > n_texcoord_num)?
						(float)((j % 4) == 3) : (float)(((GLbyte*)((int)p_vertex +
						n_texcoord_offset))[j % 4]) / 127;
				}
				break;
			case GL_UNSIGNED_BYTE:
				for(int j = 0; j < (n_texcoord_num + 3) / 4; j ++) {
					t_vertex.v_texcoord[j / 4][j % 4] = (j > n_texcoord_num)?
						(float)((j % 4) == 3) : (float)(((GLubyte*)((int)p_vertex +
						n_texcoord_offset))[j % 4]) / 255;
				}
				break;
			case GL_SHORT:
				for(int j = 0; j < (n_texcoord_num + 3) / 4; j ++) {
					t_vertex.v_texcoord[j / 4][j % 4] = (j > n_texcoord_num)?
						(float)((j % 4) == 3) : (float)(((GLshort*)((int)p_vertex +
						n_texcoord_offset))[j % 4]) / 32767;
				}
				break;
			case GL_UNSIGNED_SHORT:
				for(int j = 0; j < (n_texcoord_num + 3) / 4; j ++) {
					t_vertex.v_texcoord[j / 4][j % 4] = (j > n_texcoord_num)?
						(float)((j % 4) == 3) : (float)(((GLushort*)((int)p_vertex +
						n_texcoord_offset))[j % 4]) / 65535;
				}
				break;
			case GL_INT:
				for(int j = 0; j < (n_texcoord_num + 3) / 4; j ++) {
					t_vertex.v_texcoord[j / 4][j % 4] = (j > n_texcoord_num)?
						(float)((j % 4) == 3) : (float)(((GLint*)((int)p_vertex +
						n_texcoord_offset))[j % 4]) / 2147483647;
				}
				break;
			case GL_UNSIGNED_INT:
				for(int j = 0; j < (n_texcoord_num + 3) / 4; j ++) {
					t_vertex.v_texcoord[j / 4][j % 4] = (j > n_texcoord_num)?
						(float)((j % 4) == 3) : (float)(((GLuint*)((int)p_vertex +
						n_texcoord_offset))[j % 4]) / 4294967295UL;
				}
				break;
			case GL_FLOAT:
				for(int j = 0; j < ((n_texcoord_num + 3) & (~3)); j ++) {
					t_vertex.v_texcoord[j / 4][j % 4] = (j > n_texcoord_num)?
						(float)((j % 4) == 3) : (float)(((GLfloat*)((int)p_vertex +
						n_texcoord_offset))[j % 4]);
				}
				break;
			default:
				return false;
			}
		} else
			t_vertex.v_texcoord[0] = Vector4f(0, 0, 0, 1);
		// set texture coord

		if(!p_AddVertex(t_vertex))
			return false;
		// add vertex
	}
	// add vertices

	return true;
}

/*
 *	bool CGeometry::AddGLVertices(const void *p_vertex, int n_vertex_num, GLenum n_vertex_format)
 *		- add vertices, vertex format specified by OpenGL enum
 *		- n_vertex_format is one of GL_V2F, GL_V3F, GL_C4UB_V2F, GL_C4UB_V3F, GL_C3F_V3F,
 *		  GL_N3F_V3F, GL_C4F_N3F_V3F, GL_T2F_V3F, GL_T4F_V4F, GL_T2F_C4UB_V3F, GL_T2F_C3F_V3F,
 *		  GL_T2F_N3F_V3F, GL_T2F_C4F_N3F_V3F, GL_T4F_C4F_N3F_V4F
 *		- note if n_vertex_format is GL_*V4F, three position coordinates only will be used,
 *		  the fourth one is discarded
 */
bool CGeometry::AddGLVertices(const void *p_vertex, int n_vertex_num, GLenum n_vertex_format)
{
	COpenGLFormatTable::TVertexFormat t_format;
	t_format = COpenGLFormatTable::t_Vertex_Format(n_vertex_format);
	if(!t_format.n_format)
		return false;
	// find format

	return AddVertices(p_vertex, n_vertex_num, t_format.n_stride,
	   t_format.n_position_offset,
	   t_format.n_texcoord_offset, t_format.n_texcoord_num,
	   t_format.n_texcoord_type, t_format.n_normal_offset,
	   t_format.n_color_offset, t_format.n_color_format,
	   t_format.n_color_type, n_vertex_format == GL_V2F || n_vertex_format == GL_C4UB_V2F);
	// add vertices
}

/*
 *	bool CGeometry::GetVertices(void *&r_p_vertex, int &r_n_length, int n_stride,
 *		int n_position_offset = 0, int n_texcoord_offset = -1, int n_texcoord_num = 2,
 *		GLenum n_texcoord_type = GL_FLOAT, int n_normal_offset = -1, int n_color_offset = -1,
 *		GLenum n_color_format = GL_RGB, GLenum n_color_type = GL_UNSIGNED_BYTE,
 *		int n_vertex_coord_dimension = 3)
 *		- get vertices in required custom format (n_vertex_coord_dimension must be 2 to 4)
 *		- r_p_vertex is vertex buffer pointer, if r_n_length is sufficient do nothing,
 *		  otherwise (re)alloc; always writes length of used space into r_n_length
 *		  (so keep variable with number of bytes allocated and next one with number of bytes used)
 */
bool CGeometry::GetVertices(void *&r_p_vertex, int &r_n_length, int n_stride,
	int n_position_offset, int n_texcoord_offset, int n_texcoord_num, GLenum n_texcoord_type,
	int n_normal_offset, int n_color_offset, GLenum n_color_format, GLenum n_color_type,
	int n_vertex_coord_dimension)
{
	if(n_vertex_coord_dimension < 2 || n_vertex_coord_dimension > 4 ||
	   n_texcoord_num > 4 * max_Vertex_Texcoords)
		return false;

	int n_required_length = n_stride * m_vertex_list.size();
	if(r_n_length < n_required_length) {
		if(r_n_length > 0)
			delete[] r_p_vertex;
		if(!(r_p_vertex = (void*)new char[n_required_length])) {
			r_n_length = 0;
			return false;
		}
	}
	r_n_length = n_required_length;
	// make enough space for vertex data

	Vector4i v_component_indices;
	if(n_color_offset != -1) {
		v_component_indices = COpenGLFormatTable::v_Color_Indices(n_color_format);
		if(v_component_indices == Vector4i(-1, -1, -1, -1))
			return false;
	}
	// recognize color format (a bit different table than in complementary AddVertices() function)

	{void *p_vertex_data = r_p_vertex;
	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end(); p_vertex ++,
	   p_vertex_data = (void*)((int)p_vertex_data + n_stride)) {
		// p_vertex_data contains buffer where vertex data should be stored,
		// p_vertex is vertex struct

		if(n_vertex_coord_dimension == 3) {
			((float*)((int)p_vertex_data + n_position_offset))[0] = (*p_vertex)->v_pos.x;
			((float*)((int)p_vertex_data + n_position_offset))[1] = (*p_vertex)->v_pos.y;
			((float*)((int)p_vertex_data + n_position_offset))[2] = (*p_vertex)->v_pos.z;
		} else if(n_vertex_coord_dimension == 2) {
			((float*)((int)p_vertex_data + n_position_offset))[0] = (*p_vertex)->v_pos.x;
			((float*)((int)p_vertex_data + n_position_offset))[1] = (*p_vertex)->v_pos.y;
		} else /*if(n_vertex_coord_dimension == 4)*/ {
			((float*)((int)p_vertex_data + n_position_offset))[0] = (*p_vertex)->v_pos.x;
			((float*)((int)p_vertex_data + n_position_offset))[1] = (*p_vertex)->v_pos.y;
			((float*)((int)p_vertex_data + n_position_offset))[2] = (*p_vertex)->v_pos.z;
			((float*)((int)p_vertex_data + n_position_offset))[3] = 1.0f;
		}
		// get position

		if(n_normal_offset != -1) {
			((float*)((int)p_vertex_data + n_normal_offset))[0] = (*p_vertex)->v_normal.x;
			((float*)((int)p_vertex_data + n_normal_offset))[1] = (*p_vertex)->v_normal.y;
			((float*)((int)p_vertex_data + n_normal_offset))[2] = (*p_vertex)->v_normal.z;
		}
		// get normal

		if(n_color_offset != -1) {
			switch(n_color_type) {
			case GL_BYTE:
				for(int j = 0; j < 4; j ++) {
					if(v_component_indices[j] == -1)
						continue;
					(((GLbyte*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_color[j] > 1)? 1 : (((*p_vertex)->v_color[j] >= 0)?
						(*p_vertex)->v_color[j] : 0)) * 127;
				}
				break;
			case GL_UNSIGNED_BYTE:
				for(int j = 0; j < 4; j ++) {
					if(v_component_indices[j] == -1)
						continue;
					(((GLubyte*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_color[j] > 1)? 1 : (((*p_vertex)->v_color[j] >= 0)?
						(*p_vertex)->v_color[j] : 0)) * 255;
				}
				break;
			case GL_SHORT:
				for(int j = 0; j < 4; j ++) {
					if(v_component_indices[j] == -1)
						continue;
					(((GLshort*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_color[j] > 1)? 1 : (((*p_vertex)->v_color[j] >= 0)?
						(*p_vertex)->v_color[j] : 0)) * 32767;
				}
				break;
			case GL_UNSIGNED_SHORT:
				for(int j = 0; j < 4; j ++) {
					if(v_component_indices[j] == -1)
						continue;
					(((GLushort*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_color[j] > 1)? 1 : (((*p_vertex)->v_color[j] >= 0)?
						(*p_vertex)->v_color[j] : 0)) * 65535;
				}
				break;
			case GL_INT:
				for(int j = 0; j < 4; j ++) {
					if(v_component_indices[j] == -1)
						continue;
					(((GLint*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_color[j] > 1)? 1 : (((*p_vertex)->v_color[j] >= 0)?
						(*p_vertex)->v_color[j] : 0)) * 2147483647;
				}
				break;
			case GL_UNSIGNED_INT:
				for(int j = 0; j < 4; j ++) {
					if(v_component_indices[j] == -1)
						continue;
					(((GLuint*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_color[j] > 1)? 1 : (((*p_vertex)->v_color[j] >= 0)?
						(*p_vertex)->v_color[j] : 0)) * 4294967295UL;
				}
				break;
			case GL_FLOAT:
				for(int j = 0; j < 4; j ++) {
					if(v_component_indices[j] == -1)
						continue;
					(((GLfloat*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						/*(((*p_vertex)->v_color[j] > 1)? 1 : (((*p_vertex)->v_color[j] >= 0)?*/
						(*p_vertex)->v_color[j]/* : 0)) * 1.0f*/;
				}
				break;
			default:
				return false;
			}
		}
		// get color

		if(n_texcoord_offset != -1 && n_texcoord_num) {
			switch(n_texcoord_type) {
			case GL_BYTE:
				for(int j = 0; j < n_texcoord_num; j ++) {
					(((GLbyte*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_texcoord[j / 4][j % 4] > 1)? 1 :
						(((*p_vertex)->v_texcoord[j / 4][j % 4] >= 0)?
						(*p_vertex)->v_texcoord[j / 4][j % 4] : 0)) * 127;
				}
				break;
			case GL_UNSIGNED_BYTE:
				for(int j = 0; j < n_texcoord_num; j ++) {
					(((GLubyte*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_texcoord[j / 4][j % 4] > 1)? 1 :
						(((*p_vertex)->v_texcoord[j / 4][j % 4] >= 0)?
						(*p_vertex)->v_texcoord[j / 4][j % 4] : 0)) * 255;
				}
				break;
			case GL_SHORT:
				for(int j = 0; j < n_texcoord_num; j ++) {
					(((GLshort*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_texcoord[j / 4][j % 4] > 1)? 1 :
						(((*p_vertex)->v_texcoord[j / 4][j % 4] >= 0)?
						(*p_vertex)->v_texcoord[j / 4][j % 4] : 0)) * 32767;
				}
				break;
			case GL_UNSIGNED_SHORT:
				for(int j = 0; j < n_texcoord_num; j ++) {
					(((GLushort*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_texcoord[j / 4][j % 4] > 1)? 1 :
						(((*p_vertex)->v_texcoord[j / 4][j % 4] >= 0)?
						(*p_vertex)->v_texcoord[j / 4][j % 4] : 0)) * 65535;
				}
				break;
			case GL_INT:
				for(int j = 0; j < n_texcoord_num; j ++) {
					(((GLint*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_texcoord[j / 4][j % 4] > 1)? 1 :
						(((*p_vertex)->v_texcoord[j / 4][j % 4] >= 0)?
						(*p_vertex)->v_texcoord[j / 4][j % 4] : 0)) * 2147483647;
				}
				break;
			case GL_UNSIGNED_INT:
				for(int j = 0; j < n_texcoord_num; j ++) {
					(((GLuint*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						(((*p_vertex)->v_texcoord[j / 4][j % 4] > 1)? 1 :
						(((*p_vertex)->v_texcoord[j / 4][j % 4] >= 0)?
						(*p_vertex)->v_texcoord[j / 4][j % 4] : 0)) * 4294967295UL;
				}
				break;
			case GL_FLOAT:
				for(int j = 0; j < n_texcoord_num; j ++) {
					(((GLfloat*)((int)p_vertex_data + n_color_offset))[v_component_indices[j]]) =
						/*(((*p_vertex)->v_texcoord[j / 4][j % 4] > 1)? 1 :
						(((*p_vertex)->v_texcoord[j / 4][j % 4] >= 0)?*/
						(*p_vertex)->v_texcoord[j / 4][j % 4]/* : 0)) * 1.0f*/;
				}
				break;
			default:
				return false;
			}
		}
		// get texcoords
	}}
	// get vertex data

	return true;
}

/*
 *	bool CGeometry::GetGLVertices(void *&r_p_vertex, int &r_n_length, GLenum n_vertex_format)
 *		- add vertices, vertex format specified by OpenGL enum
 *		- n_vertex_format is one of GL_V2F, GL_V3F, GL_C4UB_V2F, GL_C4UB_V3F, GL_C3F_V3F,
 *		  GL_N3F_V3F, GL_C4F_N3F_V3F, GL_T2F_V3F, GL_T4F_V4F, GL_T2F_C4UB_V3F, GL_T2F_C3F_V3F,
 *		  GL_T2F_N3F_V3F, GL_T2F_C4F_N3F_V3F, GL_T4F_C4F_N3F_V4F
 *		- note if n_vertex_format is GL_*V4F, three position coordinates only will be used,
 *		  the fourth one is discarded
 */
bool CGeometry::GetGLVertices(void *&r_p_vertex, int &r_n_length, GLenum n_vertex_format)
{
	COpenGLFormatTable::TVertexFormat t_format;
	t_format = COpenGLFormatTable::t_Vertex_Format(n_vertex_format);
	if(!t_format.n_format)
		return false;
	// find format
	
	int n_vertex_dimension;
	switch(n_vertex_format) {
	case GL_V2F:
	case GL_C4UB_V2F:
		n_vertex_dimension = 2;
		break;
	case GL_V3F:
	case GL_C4UB_V3F:
	case GL_C3F_V3F:
	case GL_N3F_V3F:
	case GL_C4F_N3F_V3F:
	case GL_T2F_V3F:
	case GL_T2F_C4UB_V3F:
	case GL_T2F_C3F_V3F:
	case GL_T2F_N3F_V3F:
	case GL_T2F_C4F_N3F_V3F:
		n_vertex_dimension = 3;
		break;
	case GL_T4F_V4F:
	case GL_T4F_C4F_N3F_V4F:
		n_vertex_dimension = 4;
		break;
	default:
		return false;
	}
	// find out how many dimensions of vertices we should write

	return GetVertices(r_p_vertex, r_n_length, t_format.n_stride,
	   t_format.n_position_offset, t_format.n_texcoord_offset, t_format.n_texcoord_num,
	   t_format.n_texcoord_type, t_format.n_normal_offset, t_format.n_color_offset,
	   t_format.n_color_format, t_format.n_color_type, n_vertex_dimension);
	// get vertices
}

/*
 *	bool CGeometry::AddPrimitives(GLenum n_prim_type, int n_vertex_offset,
 *		const int *p_vertex_index_list, int n_vertex_num, int n_material,
 *		bool b_smooth, bool b_smooth_prev)
 *		- add certain primitives to poly-list
 *		- n_vertex_offset is added to every vertex index, usualy 0 or
 *		  n_Vertex_Num() before call to AddVertices()
 *		- p_vertex_list can be 0 to use linear vertex indices in range
 *		  <n_vertex_offset, n_vertex_offset + n_vertex_num - 1>
 *		- n_prim_type can be one of GL_TRIANGLES, GL_QUADS, GL_TRIANGLE_FAN, GL_TRIANGLE_STRIP,
 *		  GL_QUAD_STRIP, GL_POLYGON
 *		- n_material is material index
 *		- set b_smooth to true if you want all polygons to have the same smoothing group
 *		- set b_smooth_prev to true if you want to smooth added polygons with previously
 *		  added polys
 */
bool CGeometry::AddPrimitives(GLenum n_prim_type, int n_vertex_offset,
	const int *p_vertex_index_list, int n_vertex_num, int n_material,
	bool b_smooth, bool b_smooth_prev)
{
	if(b_smooth_prev)
		m_n_surface_space --;
	// surface smoothing

	TPolygon t_polygon;
	t_polygon.n_surface_flags = surf_Solid;
	t_polygon.n_material = n_material;
	t_polygon.n_neighbour_edge_num = 0;
	memset(t_polygon.n_surface_map, 0, sizeof(t_polygon.n_surface_map));
	t_polygon.p_neighbour_index_list = 0;
	// polygon struct

	TRefVertex t_vertex;
	t_vertex.p_geom = this;
	// referencing vertex struct

	switch(n_prim_type) {
	case GL_POLYGON:
	case GL_TRIANGLES:
	case GL_QUADS:
		{
			int n_vertices_poly = (n_prim_type == GL_POLYGON)?
				n_vertex_num : 3 + (n_prim_type == GL_QUADS);

			for(int i = 0, n = n_vertices_poly; n <= n_vertex_num; i = n, n += n_vertices_poly) {
				CRefPolygon *p_poly;
				if(!(p_poly = new CRefPolygon))
					return false;
				for(int j = i; j < n; j ++) {
					t_vertex.p_vertex = m_vertex_list[n_vertex_offset +
						((p_vertex_index_list)? p_vertex_index_list[j] : j)];
					if(!p_poly->Add_Vertex(t_vertex))
						return false;
				}
				// create poly, add vertices

				p_poly->Calc_Normal();
				// calc normal

				t_polygon.p_poly = p_poly;
				t_polygon.n_surface = m_n_surface_space;

				m_b_poly_hash_dirty_flag = true;
				m_b_vertex_normal_smooth_group_dirty_flag = true;
				m_b_vertex_normal_smooth_all_dirty_flag = true;
				m_b_poly_neighbours_dirty_flag = true;

				m_poly_list.push_back(t_polygon);
				if(!m_poly_list.size()) {
					delete p_poly;
					return false;
				}

				if(!b_smooth)
					m_n_surface_space ++;
			}
			// add all polygons
		}
		break;

	case GL_TRIANGLE_STRIP:
		{
			n_vertex_num -= 2;
			for(int i = 0; i < n_vertex_num; i ++) {
				CRefPolygon *p_poly;
				if(!(p_poly = new CRefPolygon))
					return false;

				int p_index[3] = {(i + 1) & (~1), i | 1, i + 2};
				for(int j = 0; j < 3; j ++) {
					t_vertex.p_vertex = m_vertex_list[n_vertex_offset +
						((p_vertex_index_list)? p_vertex_index_list[p_index[j]] : p_index[j])];
					if(!p_poly->Add_Vertex(t_vertex))
						return false;
				}
				// create poly, add vertices

				p_poly->Calc_Normal();
				// calc normal

				t_polygon.p_poly = p_poly;
				t_polygon.n_surface = m_n_surface_space;

				m_b_poly_hash_dirty_flag = true;
				m_b_vertex_normal_smooth_group_dirty_flag = true;
				m_b_vertex_normal_smooth_all_dirty_flag = true;
				m_b_poly_neighbours_dirty_flag = true;

				m_poly_list.push_back(t_polygon);
				if(!m_poly_list.size()) {
					delete p_poly;
					return false;
				}

				if(!b_smooth)
					m_n_surface_space ++;
			}
			// add all polygons
		}
		break;

	case GL_QUAD_STRIP:
		{
			n_vertex_num -= 3;
			for(int i = 0; i < n_vertex_num; i += 2) {
				CRefPolygon *p_poly;
				if(!(p_poly = new CRefPolygon))
					return false;

				int p_index[4] = {i, i + 1, i + 3, i + 2};
				for(int j = 0; j < 4; j ++) {
					t_vertex.p_vertex = m_vertex_list[n_vertex_offset +
						((p_vertex_index_list)? p_vertex_index_list[p_index[j]] : p_index[j])];
					if(!p_poly->Add_Vertex(t_vertex))
						return false;
				}
				// create poly, add vertices

				p_poly->Calc_Normal();
				// calc normal

				t_polygon.p_poly = p_poly;
				t_polygon.n_surface = m_n_surface_space;

				m_b_poly_hash_dirty_flag = true;
				m_b_vertex_normal_smooth_group_dirty_flag = true;
				m_b_vertex_normal_smooth_all_dirty_flag = true;
				m_b_poly_neighbours_dirty_flag = true;

				m_poly_list.push_back(t_polygon);
				if(!m_poly_list.size()) {
					delete p_poly;
					return false;
				}

				if(!b_smooth)
					m_n_surface_space ++;
			}
			// add all polygons
		}
		break;

	default:
		return false;
	}

	if(b_smooth)
		m_n_surface_space ++;
	// if it wasn't smooth, we incremented it already

	return true;

	/*struct {
		GLenum n_type;
		int n_fixed_vertices;
		int n_reuse_vertices;
		int n_vertices_poly;
	} p_prim_list[] = {
		{GL_TRIANGLES, 0, 0, 3},
		{GL_QUADS, 0, 0, 4},
		{GL_TRIANGLE_FAN, 1, 1, 3},
		{GL_TRIANGLE_STRIP, 0, 2, 3},
		{GL_QUAD_STRIP, 0, 2, 4},
		{GL_POLYGON, 0, 0, -1} // -1 == use n_vertex_num
	};

	for(int i = 0; i < sizeof(p_prim_list) / sizeof(p_prim_list[0]); i ++) {
		if(p_prim_list[i].n_type == n_prim_type) {
			if(b_smooth_prev)
				m_n_surface_space --;
			// surface smoothing

			int n_vertices_poly = (p_prim_list[i].n_vertices_poly == -1)?
				n_vertex_num : p_prim_list[i].n_vertices_poly;
			// vertex number per primitive

			TPolygon t_polygon;
			t_polygon.n_surface_flags = surf_Solid;
			t_polygon.n_material = n_material;
			t_polygon.n_neighbour_edge_num = 0;
			memset(t_polygon.n_surface_map, 0, sizeof(t_polygon.n_surface_map));
			t_polygon.p_neighbour_index_list = 0;
			// polygon struct

			TRefVertex t_vertex;
			t_vertex.p_geom = this;
			// referencing vertex struct

			for(int j = 0/*p_prim_list[i].n_fixed_vertices + p_prim_list[i].n_reuse_vertices* /;
			   j < n_vertex_num; j += n_vertices_poly -
			   (p_prim_list[i].n_fixed_vertices + p_prim_list[i].n_reuse_vertices)) {
				CRefPolygon *p_poly;
				if(!(p_poly = new CRefPolygon))
					return false;
				for(int n = 0; n < p_prim_list[i].n_fixed_vertices; n ++) {
					t_vertex.p_vertex = m_vertex_list[n_vertex_offset +
						((p_vertex_index_list)? p_vertex_index_list[n] : n)];
					if(!p_poly->Add_Vertex(t_vertex))
						return false;
				}
				// add fixed vertices ...

				if(!((j & 1) && (n_vertices_poly & 1))) {
					for(int n = 0; n < p_prim_list[i].n_reuse_vertices; n ++) {
						t_vertex.p_vertex = m_vertex_list[n_vertex_offset +
							((p_vertex_index_list)? p_vertex_index_list[j + n] : j + n)];
						if(!p_poly->Add_Vertex(t_vertex))
							return false;
					}
				} else {
					for(int n = p_prim_list[i].n_reuse_vertices - 1; n >= 0; n --) {
						t_vertex.p_vertex = m_vertex_list[n_vertex_offset +
							((p_vertex_index_list)? p_vertex_index_list[j + n] : j + n)];
						if(!p_poly->Add_Vertex(t_vertex))
							return false;
					}
				}
				// add reused vertices (in forward order or in backward order
				// for triangles on odd steps)

				for(int n = p_prim_list[i].n_fixed_vertices + p_prim_list[i].n_reuse_vertices;
				   n < n_vertices_poly - p_prim_list[i].n_reuse_vertices; n ++) {
					t_vertex.p_vertex = m_vertex_list[n_vertex_offset +
						((p_vertex_index_list)? p_vertex_index_list[j + n] : j + n)];
					if(!p_poly->Add_Vertex(t_vertex))
						return false;
				}
				// add vertices in forward order

				for(int n = n_vertices_poly - 1;
				   n >= n_vertices_poly - p_prim_list[i].n_reuse_vertices; n --) {
					t_vertex.p_vertex = m_vertex_list[n_vertex_offset +
						((p_vertex_index_list)? p_vertex_index_list[j + n] : j + n)];
					if(!p_poly->Add_Vertex(t_vertex))
						return false;
				}
				// add vertices to be reused in reverse order

				p_poly->Calc_Normal();
				// calc normal

				t_polygon.p_poly = p_poly;
				t_polygon.n_surface = m_n_surface_space;

				m_b_poly_hash_dirty_flag = true;
				m_b_vertex_normal_smooth_group_dirty_flag = true;
				m_b_vertex_normal_smooth_all_dirty_flag = true;
				m_b_poly_neighbours_dirty_flag = true;

				m_poly_list.push_back(t_polygon);
				if(!m_poly_list.size()) {
					delete p_poly;
					return false;
				}

				if(!b_smooth)
					m_n_surface_space ++;
			}
			// add all polygons

			m_n_surface_space ++;

			return true;
		}
	}*/ // not working properly for tri-strips, don't have time to debug
	// (but it should be worth it - it will generate potentially smaller resulting code)

	return false;
}

/*
 *	void CGeometry::ClearUnusedVertexFields(TVertex *p_vertex)
 *		- fill vertex fields, Create_ functions doesn't fill with some default values
 *		  to avoid grabage in those
 *		- namely clears flag to zero, color to (0, 0, 0, 1) and
 *		  texcoords 0 .. max_Vertex_Texcoords to (0, 0, 0, 1)
 */
/*void CGeometry::ClearUnusedVertexFields(TVertex *p_vertex)
{
	p_vertex->n_flag = 0;
	p_vertex->v_color = Color4f(0, 0, 0, 1);
	for(Vector4f *p_texcoord = p_vertex->v_texcoord, *p_end = p_vertex->v_texcoord +
	   max_Vertex_Texcoords; p_texcoord < p_end;)
		*p_texcoord ++ = Vector4f(0, 0, 0, 1);
}*/

/*
 *	bool CGeometry::Create_Grid(Vector3f v_org, Vector3f v_x, Vector3f v_y,
 *		int n_material, int n_tess_x, int n_tess_y)
 *		- create (2D) grid with center in origin, size v_x  v_z (v_y is normal)
 *		  and n_tess_x  n_tess_y quads
 *		- negating one of direction vectors cause quads flips
 *		- create n_tess_x  n_tess_y quads and (n_tess_x + 1)  (n_tess_y + 1) vertices
 */
bool CGeometry::Create_Grid(Vector3f v_org, Vector3f v_x, Vector3f v_z,
	int n_material, int n_tess_x, int n_tess_y)
{
	float f_x_step = 1.0f / n_tess_x;
	float f_y_step = 1.0f / n_tess_y;

	int n_first_vertex = m_vertex_list.size();

	Vector3f v_normal = v_x.v_Cross(v_z);
	v_normal.Normalize();

	for(int j = 0; j <= n_tess_y; j ++) {
		for(int i = 0; i <= n_tess_x; i ++) {
			TVertex t_tmp;
			t_tmp.Clear_ExtraFields();
			t_tmp.v_pos = v_org + v_x * (.5f - i * f_x_step) + v_z * (-.5f + j * f_y_step);
			t_tmp.v_normal = v_normal;
			t_tmp.v_texcoord[0] = Vector4f(i * f_x_step, j * f_y_step, 0, 1);
			if(!p_AddVertex(t_tmp))
				return false;
		}
	}
	// calc vertices

	return AddGridQuads(n_first_vertex, n_tess_x, n_tess_y, false, n_material);
}

/*
 *	bool CGeometry::Create_Heightfield(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		int n_material, int n_tess_x, int n_tess_y, unsigned char *p_heightmap_grayscale)
 *		- create (2D) grid with center in origin, size v_x  v_z (v_y is normal)
 *		  and n_tess_x  n_tess_y quads
 *		- heightmap is used to displace vertices in v_y direction, it has to have
 *		  resolution of (n_tess_x + 1)  (n_tess_y + 1) pixels
 *		- negating one of direction vectors cause quads flips
 *		- create n_tess_x  n_tess_y quads and (n_tess_x + 1)  (n_tess_y + 1) vertices
 */
bool CGeometry::Create_Heightfield(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z, int n_material,
	int n_tess_x, int n_tess_y, unsigned char *p_heightmap_grayscale)
{
	float f_x_step = 1.0f / n_tess_x;
	float f_y_step = 1.0f / n_tess_y;

	int n_first_vertex = m_vertex_list.size();

	Vector3f v_normal = v_x.v_Cross(v_z);
	v_normal.Normalize();

	for(int j = 0; j <= n_tess_y; j ++) {
		for(int i = 0; i <= n_tess_x; i ++) {
			TVertex t_tmp;
			t_tmp.Clear_ExtraFields();
			t_tmp.v_pos = v_org + v_x * (.5f - i * f_x_step) + v_z * (-.5f + j * f_y_step) +
				v_y * ((float)p_heightmap_grayscale[i + j * (n_tess_x + 1)] / 256);
			t_tmp.v_normal = v_normal;
			t_tmp.v_texcoord[0] = Vector4f(i * f_x_step, j * f_y_step, 0, 1);
			if(!p_AddVertex(t_tmp))
				return false;
		}
	}
	// calc vertices

	return AddGridQuads(n_first_vertex, n_tess_x, n_tess_y, false, n_material);
}

/*
 *	bool CGeometry::Create_Box(Vector3f v_org, Vector3f v_x, Vector3f v_y,
 *		Vector3f v_z, int n_material, int n_tess_x, int n_tess_y, int n_tess_z)
 *		- create box with center in v_org, size v_x  v_y  v_z, teselated
 *		  n_tess_x  n_tess_y  n_tess_z
 *		- walls of box are separate (i.e.) vertices on edges aren't weld
 *		- create neg-x wall (n_tess_z  n_tess_y quads, (n_tess_z + 1)  (n_tess_y + 1) vertices)
 *		- create pos-x wall (n_tess_z  n_tess_y quads, (n_tess_z + 1)  (n_tess_y + 1) vertices)
 *		- create neg-y wall (n_tess_x  n_tess_z quads, (n_tess_x + 1)  (n_tess_z + 1) vertices)
 *		- create pos-y wall (n_tess_x  n_tess_z quads, (n_tess_x + 1)  (n_tess_z + 1) vertices)
 *		- create neg-z wall (n_tess_x  n_tess_y quads, (n_tess_x + 1)  (n_tess_y + 1) vertices)
 *		- create pos-z wall (n_tess_x  n_tess_y quads, (n_tess_x + 1)  (n_tess_y + 1) vertices)
 */
bool CGeometry::Create_Box(Vector3f v_org, Vector3f v_x, Vector3f v_y,
	Vector3f v_z, int n_material, int n_tess_x, int n_tess_y, int n_tess_z)
{
	if(!Create_Grid(v_org - v_x * .5f, v_z, -v_y, n_material, n_tess_z, n_tess_y) ||
	   !Create_Grid(v_org + v_x * .5f, v_z, v_y, n_material, n_tess_z, n_tess_y) ||
	   !Create_Grid(v_org - v_y * .5f, v_x, -v_z, n_material, n_tess_x, n_tess_z) ||
	   !Create_Grid(v_org + v_y * .5f, v_x, v_z, n_material, n_tess_x, n_tess_z) ||
	   !Create_Grid(v_org - v_z * .5f, v_x, v_y, n_material, n_tess_x, n_tess_y) ||
	   !Create_Grid(v_org + v_z * .5f, v_x, -v_y, n_material, n_tess_x, n_tess_y))
		return false;
	return true;
}

/*
 *	bool CGeometry::Create_Cylinder(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		int n_material, int n_cap_material, int n_tess_r, int n_tess_a, int n_tess_cap)
 *		- create cylinder, v_org is in the center of bottom cap, axis goes up along v_y
 *		- coat is tesselated n_tess_r (radially)  n_tess_a (axially), consists of
 *		  n_tess_r  n_tess_a quads and (n_tess_r + 1)  (n_tess_a + 1) vertices
 *		- cap is n_tess_r  (n_tess_cap - 1) quads (for n_tess_cap > 1) and n_tess_r
 *		  triangles (for n_tess_cap > 0) or n_tess_r - 2 triangles, placed assymetrically
 *		  for n_tess_cap == 0 (for values below zero there are no caps)
 *		- caps aren't welded to coat
 *		- for n_tess_cap > 0 cap is n_tess_r  n_tess_cap polygons (last n_tess_r are triangles)
 *		  and n_tess_r  (n_tess_cap - 1) + 1 vertices (last vertex is the center one)
 *		- for n_tess_cap == 0 cap is n_tess_r - 2 polygons and n_tess_r vertices
 *		- bottom cap is inserted first, top second
 */
bool CGeometry::Create_Cylinder(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
	int n_material, int n_cap_material, int n_tess_r, int n_tess_a, int n_tess_cap)
{
	int n_first_vertex = m_vertex_list.size();

	if(n_tess_a) {
		if(!AddCylVertices(v_org, v_x, v_y, v_z, n_tess_r, n_tess_a, false) ||
		   !AddGridQuads(n_first_vertex, n_tess_r, n_tess_a, false, n_material))
			return false;
	}
	// add coat polys

	n_first_vertex = m_vertex_list.size();
	if(!AddCapVertices(v_org, v_x, v_y, v_z, n_tess_r, n_tess_cap, true) ||
	   !AddCapPolygons(n_first_vertex, n_tess_r, n_tess_cap, false, n_cap_material))
		return false;
	n_first_vertex = m_vertex_list.size();
	if(!AddCapVertices(v_org + v_y, v_x, v_y, v_z, n_tess_r, n_tess_cap, false) ||
	   !AddCapPolygons(n_first_vertex, n_tess_r, n_tess_cap, true, n_cap_material))
		return false;
	// calc cap vertices & add cap polys

	return true;
}

/*
 *	bool CGeometry::AddCylVertices(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		int n_tess_r, int n_tess_a, bool b_flip_normals)
 *		- add n_tess_r  n_tess_a cylinder vertices, v_org is at bottom end of cone (at axis),
 *		  v_x & v_z define radius, v_y is normal, n_tess_r is radial, n_tess_a axial tesselation
 *		- can flip vertex normals
 */
bool CGeometry::AddCylVertices(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
	int n_tess_r, int n_tess_a, bool b_flip_normals)
{
	float f_r_step = 2.0f * f_pi / n_tess_r;
	float f_h_step = 1.0f / n_tess_a;

	for(int j = 0; j <= n_tess_a; j ++) {
		for(int i = 0; i <= n_tess_r; i ++) {
			float f_angle = i * f_r_step;

			TVertex t_tmp;
			t_tmp.Clear_ExtraFields();
			t_tmp.v_pos = v_org + v_x * sinf(f_angle) + v_y * (j * f_h_step) +
				v_z * cosf(f_angle);
			t_tmp.v_normal = v_x * sinf(f_angle) + v_z * cosf(f_angle);
			if(b_flip_normals)
				t_tmp.v_normal *= -1.0f;
			t_tmp.v_normal.Normalize();
			t_tmp.v_texcoord[0] = Vector4f(i * (1.0f / n_tess_r), j * f_h_step, 0, 1);
			if(!p_AddVertex(t_tmp))
				return false;
		}
	}
	// calc coat vertices

	return true;
}

/*
 *	bool CGeometry::AddGridQuads(int n_first_vertex, int n_tess_x, int n_tess_y, bool b_flip, int n_material)
 *		- add quads, corresponding to grid of (n_tess_x + 1)  (n_tess_y + 1) vertices,
 *		  beggining with n_first_vertex
 *		- can flip quads
 */
bool CGeometry::AddGridQuads(int n_first_vertex, int n_tess_x, int n_tess_y, bool b_flip, int n_material)
{
	for(int i = 0; i < n_tess_x; i ++) {
		for(int j = 0; j < n_tess_y; j ++) {
			TRefVertex t_ref_vertex[4];
			t_ref_vertex[0].p_geom = this;
			t_ref_vertex[0].p_vertex = m_vertex_list[n_first_vertex + i + j * (n_tess_x + 1)];
			t_ref_vertex[1].p_geom = this;
			t_ref_vertex[1].p_vertex = m_vertex_list[n_first_vertex + i + (j + 1) * (n_tess_x + 1)];
			t_ref_vertex[2].p_geom = this;
			t_ref_vertex[2].p_vertex = m_vertex_list[n_first_vertex + (i + 1) + (j + 1) * (n_tess_x + 1)];
			t_ref_vertex[3].p_geom = this;
			t_ref_vertex[3].p_vertex = m_vertex_list[n_first_vertex + (i + 1) + j * (n_tess_x + 1)];

			TPolygon t_poly;
			t_poly.n_surface_flags = surf_Solid;
			t_poly.n_neighbour_edge_num = 0;
			t_poly.p_neighbour_index_list = 0;
			memset(t_poly.n_surface_map, 0, sizeof(t_poly.n_surface_map));
			t_poly.n_surface = m_n_surface_space;
			t_poly.n_material = n_material;
			if(!(t_poly.p_poly = new CRefPolygon))
				return false;
			if(!t_poly.p_poly->Add_Vertex(t_ref_vertex, 4)) {
				delete t_poly.p_poly;
				return false;
			}

			if(b_flip)
				t_poly.p_poly->Reverb();
			t_poly.p_poly->Calc_Normal();

			m_b_poly_hash_dirty_flag = true;
			m_b_vertex_normal_smooth_group_dirty_flag = true;
			m_b_vertex_normal_smooth_all_dirty_flag = true;
			m_b_poly_neighbours_dirty_flag = true;

			m_poly_list.push_back(t_poly);
			if(!m_poly_list.size()) {
				delete t_poly.p_poly;
				return false;
			}
		}
	}
	// calc coat polygons

	m_n_surface_space ++;

	return true;
}

/*
 *	bool CGeometry::AddCapVertices(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		int n_tess_r, int n_tess_cap, bool b_flip_normals)
 *		- add cap vertices, v_org is center of the cap, v_x and v_z define radius vectors,
 *		  v_y is normal vector
 *		- cap is radially tesselated n_tess_r
 *		- for n_tess_cap >= 1, cap is n_tess_cap centric rings of n_tess_r
 *		  vertices and a single vertex in centre
 *		- for n_tess_cap == 0, cap is n_tess_r vertices, no center vertex
 *		- for n_tess_cap < 0 it do nothing
 */
bool CGeometry::AddCapVertices(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
	int n_tess_r, int n_tess_cap, bool b_flip_normals)
{
	float f_r_step = 2.0f * f_pi / n_tess_r;
	float f_cap_step = 1.0f / n_tess_cap;

	for(int j = 0; j <= n_tess_cap; j ++) {
		for(int i = 0; i <= ((j == n_tess_cap && n_tess_cap)? 0 : n_tess_r); i ++) {
			float f_angle = i * f_r_step;
			// last vertex round has single vertex only (prev ones has n_tess_r vertices)

			TVertex t_tmp;
			t_tmp.Clear_ExtraFields();
			t_tmp.v_pos = (v_x * sinf(f_angle) + v_z * cosf(f_angle)) *
				(1 - ((n_tess_cap)? j * f_cap_step : 0)) + v_org;
			t_tmp.v_normal = v_y;
			if(b_flip_normals)
				t_tmp.v_normal *= -1.0f;
			t_tmp.v_normal.Normalize();
 			t_tmp.v_texcoord[0] = Vector4f(sinf(f_angle), cosf(f_angle), 0, 1) *
				(1 - ((n_tess_cap)? j * f_cap_step : 0));
			if(!p_AddVertex(t_tmp))
				return false;
		}
	}
	// calc cap vertices

	return true;
}

/*
 *	bool CGeometry::AddCapPolygons(int n_first_vertex, int n_tess_r, int n_tess_cap,
 *		bool b_flip, int n_material)
 *		- add cap polygons, n_first_vertex is first vertex to be assigned
 *		- cap is radially tesselated n_tess_r
 *		- for n_tess_cap == 1, cap is triangle fan
 *		- for n_tess_cap > 1, cap is n_tess_cap centric rings of quads, surrounding triangle fan
 *		  (first quads are the outside ones, last are the ones of triangle fan)
 *		- for n_tess_cap == 0, cap is n_tess_r - 2 triangles, placed assymetrically
 *		- for n_tess_cap < 0 it doesn't create any polygons
 *		- can flip polygons
 */
bool CGeometry::AddCapPolygons(int n_first_vertex, int n_tess_r, int n_tess_cap, bool b_flip, int n_material)
{
	if(n_tess_cap < 0)
		return true;

	if(!AddGridQuads(n_first_vertex, n_tess_r, n_tess_cap - 1, !b_flip, n_material))
		return false;

	for(int i = ((n_tess_cap)? 0 : 1); i < n_tess_r - ((n_tess_cap)? 0 : 1); i ++) {
		TRefVertex t_ref_vertex[3];
		t_ref_vertex[0].p_geom = this;
		t_ref_vertex[1].p_geom = this;
		t_ref_vertex[2].p_geom = this;
		if(n_tess_cap) {
			t_ref_vertex[0].p_vertex = m_vertex_list[n_first_vertex + i + (n_tess_cap - 1) * (n_tess_r + 1)];
			t_ref_vertex[1].p_vertex = m_vertex_list[n_first_vertex + (i + 1) + (n_tess_cap - 1) * (n_tess_r + 1)];
			t_ref_vertex[2].p_vertex = m_vertex_list[n_first_vertex + (n_tess_r + 1) * n_tess_cap];
		} else {
			t_ref_vertex[0].p_vertex = m_vertex_list[n_first_vertex + i];
			t_ref_vertex[1].p_vertex = m_vertex_list[n_first_vertex + (i + 1)];
			t_ref_vertex[2].p_vertex = m_vertex_list[n_first_vertex];
		}

		TPolygon t_poly;
		t_poly.n_surface_flags = surf_Solid;
		t_poly.n_neighbour_edge_num = 0;
		t_poly.p_neighbour_index_list = 0;
		memset(t_poly.n_surface_map, 0, sizeof(t_poly.n_surface_map));
		t_poly.n_surface = m_n_surface_space;
		t_poly.n_material = n_material;
		if(!(t_poly.p_poly = new CRefPolygon))
			return false;
		if(!t_poly.p_poly->Add_Vertex(t_ref_vertex, 3)) {
			delete t_poly.p_poly;
			return false;
		}
		if(b_flip)
			t_poly.p_poly->Reverb();
		t_poly.p_poly->Calc_Normal();

		m_b_poly_hash_dirty_flag = true;
		m_b_vertex_normal_smooth_group_dirty_flag = true;
		m_b_vertex_normal_smooth_all_dirty_flag = true;
		m_b_poly_neighbours_dirty_flag = true;

		m_poly_list.push_back(t_poly);
		if(!m_poly_list.size()) {
			delete t_poly.p_poly;
			return false;
		}
	}
	m_n_surface_space ++;

	return true;
}

/*
 *	bool CGeometry::Create_Cone(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		int n_material, int n_cap_material, int n_tess_r, int n_tess_a, int n_tess_cap)
 *		- create cone, v_org is in the center of the bottom cap, axis goes up along v_y
 *		- coat is tesselated n_tess_r (radially)  n_tess_a (axially), consists of
 *		  n_tess_r  n_tess_a quads and (n_tess_r + 1)  (n_tess_a + 1) vertices
 *		  (top-most n_tess_r + 1 vertices lie all on cone axis in same position)
 *		- cap is n_tess_r  (n_tess_cap - 1) quads (for n_tess_cap > 1) and n_tess_r
 *		  triangles (for n_tess_cap > 0) or n_tess_r - 2 triangles, placed assymetrically
 *		  for n_tess_cap == 0 (for values below zero there are no caps)
 *		- cap isn't welded to coat
 *		- for n_tess_cap > 0 cap is n_tess_r  n_tess_cap polygons (last n_tess_r are triangles)
 *		  and n_tess_r  (n_tess_cap - 1) + 1 vertices (last vertex is the center one)
 *		- for n_tess_cap == 0 cap is n_tess_r - 2 polygons and n_tess_r vertices
 */
bool CGeometry::Create_Cone(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
	int n_material, int n_cap_material, int n_tess_r, int n_tess_a, int n_tess_cap)
{
	float f_r_step = 2.0f * f_pi / n_tess_r;
	float f_h_step = 1.0f / n_tess_a;
	float f_cap_step = 1.0f / n_tess_cap;

	int n_first_vertex = m_vertex_list.size();

	if(n_tess_a) {
		for(int j = 0; j <= n_tess_a; j ++) {
			for(int i = 0; i <= n_tess_r; i ++) {
				float f_angle = i * f_r_step;

				TVertex t_tmp;
				t_tmp.Clear_ExtraFields();
				t_tmp.v_pos = v_org + (v_x * sinf(f_angle) + v_z * cosf(f_angle)) *
						(1 - j * f_h_step) + v_y * j * f_h_step;
				if(j < n_tess_a) {
					Vector3f v_tangent = v_x * sinf(f_angle + f_pi / 2) + v_z * cosf(f_angle + f_pi / 2);
					Vector3f v_edge = (v_org + v_y) - t_tmp.v_pos;
					t_tmp.v_normal = v_edge.v_Cross(v_tangent);
				} else
					t_tmp.v_normal = v_y;
				t_tmp.v_normal.Normalize();
				t_tmp.v_texcoord[0] = Vector4f(i * (1.0f / n_tess_r), j * f_h_step, 0, 1);
				if(!p_AddVertex(t_tmp))
					return false;
			}
		}
		// calc coat vertices

		if(!AddGridQuads(n_first_vertex, n_tess_r, n_tess_a, false, n_material))
			return false;
	}
	// add coat polys

	n_first_vertex = m_vertex_list.size();
	if(!AddCapVertices(v_org, v_x, v_y, v_z, n_tess_r, n_tess_cap, true) ||
	   !AddCapPolygons(n_first_vertex, n_tess_r, n_tess_cap, false, n_cap_material))
		return false;
	// calc cap vertices & add cap polys

	return true;
}

/*
 *	bool CGeometry::Create_Tube(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		float f_inner_r, int n_material, int n_cap_material, int n_tess_r, int n_tess_a)
 *		- create tube, v_org is on axis at the bottom end of tube, axis goes up along v_y
 *		- f_inner_r is relative radius of inner wall; if f_inner_r > 1, polygons flip inside-out
 *		- don't call with f_inner_r == 1, because it will create degenerate cap polygons
 *		  (rather call twice Create_Cylinder with cap tesselation -1 (= no caps)
 *		  and second time with twisted vector basis so polygons flip)
 *		- outer coat is inserted first, then inner coat. coats both consist of n_tess_r  n_tess_a
 *		  quads and (n_tess_r + 1)  (n_tess_a + 1) vertices
 *		- each cap consists of n_tess_r quads (n_tess_r + 1)  2 vertices,
 *		  bottom cap is inserted first
 */
bool CGeometry::Create_Tube(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
	float f_inner_r, int n_material, int n_cap_material, int n_tess_r, int n_tess_a)
{
	if(n_tess_a) {
		int n_first_vertex = m_vertex_list.size();
		if(!AddCylVertices(v_org, v_x, v_y, v_z, n_tess_r, n_tess_a, false) ||
		   !AddGridQuads(n_first_vertex, n_tess_r, n_tess_a, false, n_material))
			return false;
		n_first_vertex = m_vertex_list.size();
		if(!AddCylVertices(v_org, v_x * f_inner_r, v_y, v_z * f_inner_r, n_tess_r, n_tess_a, true) ||
		   !AddGridQuads(n_first_vertex, n_tess_r, n_tess_a, true, n_material))
			return false;
	}
	// inner & outer coat

	float f_r_step = 2.0f * f_pi / n_tess_r;

	for(float f_cap = 0; f_cap < 1.5f; f_cap += 1) {
		int n_first_vertex = m_vertex_list.size();
		int n = 0;
		for(float f_radius = 1.0f; n < 2; f_radius = f_inner_r, n ++) {
			for(int i = 0; i <= n_tess_r; i ++) {
				float f_angle = i * f_r_step;

				TVertex t_tmp;
				t_tmp.Clear_ExtraFields();
				t_tmp.v_pos = v_org + v_x * sinf(f_angle) * f_radius +
					v_z * cosf(f_angle) * f_radius + v_y * f_cap;
				t_tmp.v_normal = v_y;
				if((f_inner_r <= 1 && f_cap < .5f) || (f_inner_r > 1 && f_cap >= .5f))
					t_tmp.v_normal *= -1;
				t_tmp.v_normal.Normalize();
				t_tmp.v_texcoord[0] = Vector4f(sinf(f_angle), cosf(f_angle), 0, 1) * f_radius;
				if(!p_AddVertex(t_tmp))
					return false;
			}
		}
		// calc cap vertices

		if(!AddGridQuads(n_first_vertex, n_tess_r, 1, (f_cap < .5f), n_cap_material))
			return false;
		// triangulate cap
	}

	return true;
}

/*
 *	bool CGeometry::Create_Sphere(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		int n_material, int n_tess_r, int n_tess_a)
 *		- create sphere, v_org is center, v_x trough v_z define radius
 *		- sphere is tesselated n_tess_r radially (axis goes along v_y) and n_tess_a axialy
 *		- sphere consist of grid of n_tess_a  n_tess_r quads (top-most and bottom-most
 *		  have degenerated edges so texture shrinks correctly)
 *		  and (n_tess_a + 1)  (n_tess_r + 1) vertices
 */
bool CGeometry::Create_Sphere(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z, int n_material, int n_tess_r, int n_tess_a)
{
	float f_r_step = 2.0f * f_pi / n_tess_r;
	float f_h_step = f_pi / n_tess_a;

	int n_first_vertex = m_vertex_list.size();
	for(int i = 0; i <= n_tess_r; i ++) {
		float f_angle_r = i * f_r_step;
		Vector3f v_r = v_x * sinf(f_angle_r) + v_z * cosf(f_angle_r);
		for(int j = 0; j <= n_tess_a; j ++) {
			float f_angle = j * f_h_step;

			TVertex t_tmp;
			t_tmp.Clear_ExtraFields();
			t_tmp.v_pos = v_org + v_r * sinf(f_angle) + v_y * cosf(f_angle);
			t_tmp.v_normal = t_tmp.v_pos - v_org;
			t_tmp.v_normal.Normalize();
			t_tmp.v_texcoord[0] = Vector4f(i * f_r_step, j * f_h_step, 0, 1);
			if(!p_AddVertex(t_tmp))
				return false;
		}
	}

	if(!AddGridQuads(n_first_vertex, n_tess_a, n_tess_r, false, n_material))
		return false;

	return true;
}

/*
 *	bool CGeometry::Create_Hemisphere(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		int n_material, int n_tess_r, int n_tess_a)
 *		- create hemi-sphere, v_org is center, v_x trough v_z define radius
 *		- sphere is tesselated n_tess_r radially (axis goes along v_y) and n_tess_a axialy
 *		- sphere consist of grid of n_tess_a * n_tess_r quads (top-most and bottom-most
 *		  have degenerated edges so texture shrinks correctly)
 *		  and (n_tess_a + 1) * (n_tess_r + 1) vertices
 *		- cap is n_tess_r  (n_tess_cap - 1) quads (for n_tess_cap > 1) and n_tess_r
 *		  triangles (for n_tess_cap > 0) or n_tess_r - 2 triangles, placed assymetrically
 *		  for n_tess_cap == 0 (for values below zero there are no caps)
 *		- cap isn't welded to coat
 *		- for n_tess_cap > 0 cap is n_tess_r  n_tess_cap polygons (last n_tess_r are triangles)
 *		  and n_tess_r  (n_tess_cap - 1) + 1 vertices (last vertex is the center one)
 *		- for n_tess_cap == 0 cap is n_tess_r - 2 polygons and n_tess_r vertices
 */
bool CGeometry::Create_Hemisphere(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z, int n_material, int n_cap_material, int n_tess_r, int n_tess_a, int n_tess_cap)
{
	float f_r_step = 2.0f * f_pi / n_tess_r;
	float f_h_step = .5f * f_pi / n_tess_a;

	int n_first_vertex = m_vertex_list.size();
	for(int i = 0; i <= n_tess_r; i ++) {
		float f_angle_r = i * f_r_step;
		Vector3f v_r = v_x * sinf(f_angle_r) + v_z * cosf(f_angle_r);
		for(int j = 0; j <= n_tess_a; j ++) {
			float f_angle = j * f_h_step;

			TVertex t_tmp;
			t_tmp.Clear_ExtraFields();
			t_tmp.v_pos = v_org + v_r * sinf(f_angle) + v_y * cosf(f_angle);
			t_tmp.v_normal = t_tmp.v_pos - v_org;
			t_tmp.v_normal.Normalize();
			t_tmp.v_texcoord[0] = Vector4f(i * f_r_step, j * f_h_step, 0, 1);
			if(!p_AddVertex(t_tmp))
				return false;
		}
	}

	if(!AddGridQuads(n_first_vertex, n_tess_a, n_tess_r, false, n_material))
		return false;
	// hemisphere

	n_first_vertex = m_vertex_list.size();
	if(!AddCapVertices(v_org, v_x, -v_y, v_z, n_tess_r, n_tess_cap, false) ||
	   !AddCapPolygons(n_first_vertex, n_tess_r, n_tess_cap, false, n_cap_material))
		return false;
	// cap

	return true;
}

/*
 *	bool CGeometry::Create_Torus(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		int n_material, float f_inner_r, int n_tess_r, int n_tess_sub_r)
 *		- create torus, v_org is center, v_x trough v_z define radius (of major circle)
 *		  f_inner_r is relative radius, yielding cross-section size; if f_inner_r < 1 surface flips
 *		- torus is tesselated n_tess_r radially (axis goes along v_y) and n_tess_sub_r in cross-section
 *		- consists of grid of n_tess_sub_r  n_tess_r quads
 *		  and (n_tess_sub_r + 1)  (n_tess_r + 1) vertices
 */
bool CGeometry::Create_Torus(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z, int n_material, float f_inner_r, int n_tess_r, int n_tess_sub_r)
{
	float f_r_step = 2.0f * f_pi / n_tess_r;
	float f_sub_r_step = 2.0f * f_pi / n_tess_sub_r;

	int n_first_vertex = m_vertex_list.size();
	for(int i = 0; i <= n_tess_r; i ++) {
		float f_angle_r = i * f_r_step;
		Vector3f v_r = v_x * sinf(f_angle_r) + v_z * cosf(f_angle_r);
		Vector3f v_local_org = v_r + v_org;
		Vector3f v_local_x = v_r * f_inner_r;
		Vector3f v_local_z = v_y * f_inner_r;
		Vector3f v_local_y = (v_x * sinf(.5f * f_pi + f_angle_r) +
			v_z * cosf(.5f * f_pi + f_angle_r)) * f_inner_r;

		for(int j = 0; j <= n_tess_sub_r; j ++) {
			float f_angle = j * f_sub_r_step;

			TVertex t_tmp;
			t_tmp.Clear_ExtraFields();
			t_tmp.v_pos = (v_local_x * sinf(f_angle) + v_local_z * cosf(f_angle)) + v_local_org;
			t_tmp.v_normal = t_tmp.v_pos - v_local_org;
			t_tmp.v_normal.Normalize();
 			t_tmp.v_texcoord[0] = Vector4f((float)i / (float)n_tess_r,
				(float)j / (float)n_tess_sub_r, 0, 1);
			if(!p_AddVertex(t_tmp))
				return false;
		}
	}
	// calc vertices

	if(!AddGridQuads(n_first_vertex, n_tess_sub_r, n_tess_r, false, n_material))
		return false;
	// triangulate

	return true;
}

/*
 *	bool CGeometry::Create_Octagon(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		int n_material)
 *		- create octagon, v_org is center, v_x trough v_z define "radius"
 *		- first for vertices are on xz-plane, next one is bottom one and last one is top one
 *		- contain of eight triangles, first four ones are bottom ones
 */
bool CGeometry::Create_Octagon(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z, int n_material)
{
	TVertex t_tmp;
	t_tmp.Clear_ExtraFields();

	int n_first_vertex = m_vertex_list.size();

	for(int i = 0; i < 4; i ++) {
		int n_permut[4] = {0, 1, 3, 2};
		float f_x = (n_permut[i] & 1) * 2 - 1;
		float f_y = (n_permut[i] >> 1) * 2 - 1;
		t_tmp.v_pos = v_org + (v_x * f_x + v_z * f_y);
		t_tmp.v_normal = Vector3f(f_x, 0, f_y);
		t_tmp.v_normal.Normalize();
		t_tmp.v_texcoord[0] = Vector4f((f_x + 1) * .5f, (f_y + 1) * .5f, 0, 1);
		if(!p_AddVertex(t_tmp))
			return false;
	}
	// 4 vertices in xz-plane

	for(float f_dir = -1.0f; f_dir < 1.5f; f_dir += 2) {
		t_tmp.v_pos = v_org + v_y * f_dir;
		t_tmp.v_normal = Vector3f(0, f_dir, 0);
		t_tmp.v_texcoord[0] = Vector4f(.5f, .5f, 0, 1);
		if(!p_AddVertex(t_tmp))
			return false;
	}
	// top and bottom vertex

	for(int i = 0; i < 8; i ++) {
		TRefVertex t_ref_vertex[3];
		t_ref_vertex[0].p_geom = this;
		t_ref_vertex[1].p_geom = this;
		t_ref_vertex[2].p_geom = this;
		t_ref_vertex[0].p_vertex = m_vertex_list[n_first_vertex + i % 4];
		t_ref_vertex[1].p_vertex = m_vertex_list[n_first_vertex + (i + 1) % 4];
		t_ref_vertex[2].p_vertex = m_vertex_list[n_first_vertex + 4 + i / 4];

		TPolygon t_poly;
		t_poly.n_surface_flags = surf_Solid;
		t_poly.n_neighbour_edge_num = 0;
		t_poly.p_neighbour_index_list = 0;
		memset(t_poly.n_surface_map, 0, sizeof(t_poly.n_surface_map));
		t_poly.n_surface = m_n_surface_space;
		t_poly.n_material = n_material;
		if(!(t_poly.p_poly = new CRefPolygon))
			return false;
		if(!t_poly.p_poly->Add_Vertex(t_ref_vertex, 3)) {
			delete t_poly.p_poly;
			return false;
		}
		if(i < 4)
			t_poly.p_poly->Reverb();
		t_poly.p_poly->Calc_Normal();

		m_b_poly_hash_dirty_flag = true;
		m_b_vertex_normal_smooth_group_dirty_flag = true;
		m_b_vertex_normal_smooth_all_dirty_flag = true;
		m_b_poly_neighbours_dirty_flag = true;

		m_poly_list.push_back(t_poly);
		if(!m_poly_list.size()) {
			delete t_poly.p_poly;
			return false;
		}
	}

	m_n_surface_space ++;

	return true;
}

/*
 *	bool CGeometry::Create_Tetragon(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z,
 *		int n_material)
 *		- create tetragon, v_org is center, v_x trough v_z define "radius"
 *		- consist of four vertices and four triangles
 *		- texture coordinates are taken from normal
 */
bool CGeometry::Create_Tetragon(Vector3f v_org, Vector3f v_x, Vector3f v_y, Vector3f v_z, int n_material)
{
	TVertex t_tmp;
	t_tmp.Clear_ExtraFields();

	int n_first_vertex = m_vertex_list.size();

	for(int i = 0; i < 4; i ++) {
		float f_x = (i & 1) * 2 - 1;
		float f_y = ((i & 1) != (i >> 1)) * 2 - 1;
		float f_z = (i >> 1) * 2 - 1;
		t_tmp.v_pos = v_org + (v_x * f_x + v_y * f_y + v_z * f_z);
		t_tmp.v_normal = t_tmp.v_pos - v_org;
		t_tmp.v_normal.Normalize();
		t_tmp.v_texcoord[0] = Vector4f(t_tmp.v_normal.x, t_tmp.v_normal.y, t_tmp.v_normal.z, 1);
		if(!p_AddVertex(t_tmp))
			return false;
	}
	// 4 vertices in xz-plane

	for(int i = 0; i < 4; i ++) {
		TRefVertex t_ref_vertex[3];
		t_ref_vertex[0].p_geom = this;
		t_ref_vertex[1].p_geom = this;
		t_ref_vertex[2].p_geom = this;
		int a = i & 1;
		int b = i >> 1;
		t_ref_vertex[0].p_vertex = m_vertex_list[n_first_vertex + !a + 2 * (i != 0)];
		t_ref_vertex[1].p_vertex = m_vertex_list[n_first_vertex + (a & b)];
		t_ref_vertex[2].p_vertex = m_vertex_list[n_first_vertex + (!b | a) + 2 * (!a | b)];

		TPolygon t_poly;
		t_poly.n_surface_flags = surf_Solid;
		t_poly.n_neighbour_edge_num = 0;
		t_poly.p_neighbour_index_list = 0;
		memset(t_poly.n_surface_map, 0, sizeof(t_poly.n_surface_map));
		t_poly.n_surface = m_n_surface_space;
		t_poly.n_material = n_material;
		if(!(t_poly.p_poly = new CRefPolygon))
			return false;
		if(!t_poly.p_poly->Add_Vertex(t_ref_vertex, 3)) {
			delete t_poly.p_poly;
			return false;
		}
		t_poly.p_poly->Calc_Normal();

		m_b_poly_hash_dirty_flag = true;
		m_b_vertex_normal_smooth_group_dirty_flag = true;
		m_b_vertex_normal_smooth_all_dirty_flag = true;
		m_b_poly_neighbours_dirty_flag = true;

		m_poly_list.push_back(t_poly);
		if(!m_poly_list.size()) {
			delete t_poly.p_poly;
			return false;
		}
	}
	// calc tris

	m_n_surface_space ++;

	return true;
}

/*
 *	int CGeometry::n_HashFunction(Vector3f v_pos, int n_table_size, float f_cell_size)
 *		- vertex hash function. when used with vertex hash table, generated with
 *		  n_table_size and f_cell_size, you get index to cell where you find all vertices
 *		  within some distance to v_pos (and also possibly some other vertices - depends
 *		  on table size)
 */
int CGeometry::n_HashFunction(Vector3f v_pos, int n_table_size, float f_cell_size)
{
	int n_index = 0;
	for(int i = 2; i >= 0; i --) {
		n_index *= n_table_size;
		int n_axe_index = ((int)(v_pos[i] / f_cell_size)) % n_table_size;
		while(n_axe_index < 0)
			n_axe_index += n_table_size;
		n_index += n_axe_index;
	}

	return n_index;
}

/*
 *	std::vector<TVertex*> **CGeometry::p_HashVertices(int n_table_size,
 *		float f_cell_size, float f_cell_overlap)
 *		- create vertex hash-table. n_table_size is size in one dimension, table is three-dimensional
 *		  (but one-dimensional array), f_cell_size is size of cluster, vertices are merged into.
 *		  reasonable size would be 1 / n_table_size * size of your geometry, but it might be less
 *		  f_cell_overlap is overlap of cells (see, if you're searching for vertices with the same
 *		  position, you can get situation when one vertex, very close (under epsilon) to another
 *		  falls into another cell so we wouldn't consider it same position as we should. so we
 *		  introduce cell overlap to avoid such a situations by adding vertices within overlap
 *		  into both neighbour cells)
 *		- cell overlap can be zero, but it should be some small value like one tenth of f_cell_size.
 *		  increasing it just escalates vertex redundance and thus table size. however, escalating
 *		  it over f_cell_size will cause table to contain incomplete information
 *		- when n_table_size is 32, table takes 32 * 32 * 32 * sizeof(TVertex*) = 128kB of memory
 *		  + size of vertex pointer arrays, which depends on vertex count. it seems reasonable
 *		  to me (+ we never keep such a table, since any vertex manipulation invalidates
 *		  it and it's quite fast to re-build)
 *		- of course - all depends on vertex distribution accross space and world size, etc ...
 *		- when you call n_HashFunction, you get index into this table, where you can find all
 *		  vertices within distance of f_cell_overlap to f_cell_size * sqrt(3) units (and also
 *		  possibly some other vertices - depends on table size) to positon you passed to
 *		  n_HashFunction
 */
std::vector<TVertex*> **CGeometry::p_HashVertices(int n_table_size,
	float f_cell_size, float f_cell_overlap) const
{
	std::vector<TVertex*> **p_table;
	int n_table_size3 = n_table_size * n_table_size * n_table_size;
	if(!(p_table = new std::vector<TVertex*>*[n_table_size3]))
		return 0;
	for(std::vector<TVertex*> **p_hash = p_table, **p_end = p_table + n_table_size3; p_hash < p_end;) {
		if(!(*p_hash ++ = new std::vector<TVertex*>)) {
			for(p_hash --; p_hash >= p_table;)
				delete[] *(p_hash --);
			delete[] p_table;
			return 0;
		}
	}
	// alloc table

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end(); p_vertex ++) {
		int n_index[3][2];
		for(int j = 0; j < 2; j ++) {
			for(int i = 0; i < 3; i ++) {
				n_index[i][j] = ((int)(((*p_vertex)->v_pos[i] +
					j * f_cell_overlap) / f_cell_size)) % n_table_size;
				while(n_index[i][j] < 0)
					n_index[i][j] += n_table_size;
			}
		}
		// calc indices ...

		for(int x = 0; x < 2; x ++) {
			for(int y = 0; y < 2; y ++) {
				for(int z = 0; z < 2; z ++) {
					int n_cell = n_index[0][x] + (n_index[1][y] +
						n_index[2][z] * n_table_size) * n_table_size;
					// calc cell index

					p_table[n_cell]->push_back((*p_vertex));
					if(!p_table[n_cell]->size()) {
						for(std::vector<TVertex*> **p_hash = p_table, **p_end =
						   p_table + n_table_size3; p_hash < p_end;)
						    delete *(p_hash ++);
						delete[] p_table;
						return 0;
					}
					// add reference into cell

					if(n_index[2][0] == n_index[2][1])
						break;
				}
				if(n_index[1][0] == n_index[1][1])
					break;
			}
			if(n_index[0][0] == n_index[0][1])
				break;
		}
		// add references to a vertex to all cells it hit
	}

	return p_table;
}

/*
 *	int CGeometry::n_HashFunction(const TVertex *p_vertex_address, int n_table_size)
 *		- function for hashing vertices
 *		- index into hash table is generated upon vertex address
 */
int CGeometry::n_HashFunction(const TVertex *p_vertex_address, int n_table_size)
{
	return ((int)p_vertex_address / sizeof(TVertex)) % n_table_size;
}

/*
 *	std::vector<int> **CGeometry::p_HashPolygons(int n_table_size) const
 *		- create table of poylgon indices so when you call n_HashFunction() for any vertex,
 *		  returned value is index into table where can be found indices of all poylgons,
 *		  referencing that vertex (and maybe also some other polygons - depends on table size)
 *		- return 0 if there's not enough memory
 */
std::vector<int> **CGeometry::p_HashPolygons(int n_table_size) const
{
	std::vector<int> **p_table;
	if(!(p_table = new std::vector<int>*[n_table_size]))
		return 0;
	for(std::vector<int> **p_hash = p_table, **p_end = p_table + n_table_size; p_hash < p_end;) {
		if(!(*p_hash ++ = new std::vector<int>)) {
			for(p_hash --; p_hash >= p_table;)
				delete *(p_hash --);
			delete[] p_table;
			return 0;
		}
	}
	{int n_index = 0;
	for(std::vector<TPolygon>::const_iterator p_poly = m_poly_list.begin();
	   p_poly < m_poly_list.end(); p_poly ++, n_index ++) {
		for(int j = 0; j < (*p_poly).p_poly->n_Vertex_Num(); j ++) {
			int n_hash = n_HashFunction((*p_poly).p_poly->r_p_Vertex(j), n_table_size);
			if(p_table[n_hash]->size() && p_table[n_hash]->back() == n_index)
				continue; // in case we already put it there ...
			p_table[n_hash]->push_back(n_index);
			if(!p_table[n_hash]->size()) {
				for(std::vector<int> **p_hash = p_table, **p_end = p_table + n_table_size;
				   p_hash < p_end;)
					delete *(p_hash ++);
				delete[] p_table;
				return 0;
			}
		}
	}}

	return p_table;
}

/*
 *	std::vector<std::vector<TVertex*>*> *CGeometry::p_VertexMergeTable(bool b_compare_pos_only,
 *		float f_epsilon_ex) const
 *		- create table of groups of vertices. each group contain vertices, considered to be equal
 *		- you can pass pointer to vertex hash table if you got one. otherwise enter 0 and
 *		  supply valid parameters (default ones will do well); cell overlap will be .01 * cell size
 *		- thanks to hash table, complexity is somewhat worse o(n), instead of classical o(n.log(n))
 *		  (for example mesh with 5000 tris takes 2.346 secs using BRUTE_FORCE_WAY. on the other
 *		  hand with well chosen hash tables it's 0.052 sec -> over 45 times faster (release mode;
 *		  in debug mode, it was 10.3 secs and 0.58 secs = >17 times faster))
 */
std::vector<std::vector<TVertex*>*> *CGeometry::p_VertexMergeTable(bool b_compare_position_only,
	float f_epsilon_ex)
{
#ifdef BRUTE_FORCE_WAY
	std::vector<std::vector<TVertex*>*> *p_table;
	if(!(p_table = new std::vector<std::vector<TVertex*>*>))
		return 0;

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end(); p_vertex ++) {
		int n_found = -1;
		for(std::vector<std::vector<TVertex*>*>::const_iterator p_list = p_table->begin();
		   p_list < p_table->end(); p_list ++) {
			TVertex *p_other = (**p_list)[0];
			if((p_other->v_pos - (*p_vertex)->v_pos).f_Length() < f_epsilon_ex &&
			   (b_compare_position_only || *p_other == **p_vertex)) {
				n_found = p_list - p_table->begin();
				break;
			}
		}
		// search for similar vertex in table

		if(n_found > -1) {
			(*p_table)[n_found]->push_back(*p_vertex);
			if(!(*p_table)[n_found]->size()) {
				for(std::vector<std::vector<TVertex*>*>::iterator p_list =
				   p_table->begin(); p_list < p_table->end();)
					delete *(p_list ++);
				delete p_table;
				return false;
			}
			// if found, just put this one into the same group
		} else {
			std::vector<TVertex*> *p_group;
			if(!(p_group = new std::vector<TVertex*>)) {
				for(std::vector<std::vector<TVertex*>*>::iterator p_list =
				   p_table->begin(); p_list < p_table->end();)
					delete *(p_list ++);
				delete p_table;
				return false;
			}
			p_group->push_back(*p_vertex);
			p_table->push_back(p_group);
			if(!p_group->size() || !p_table->size()) {
				if(!p_table->size())
					delete p_group;
				for(std::vector<std::vector<TVertex*>*>::iterator p_list =
				   p_table->begin(); p_list < p_table->end();)
					delete *(p_list ++);
				delete p_table;
				return false;
			}
			// otherwise, create a new group with single vertex
		}
	}

	return p_table;
#else //BRUTE_FORCE_WAY
	std::vector<std::vector<TVertex*>*> *p_table;
	if(!(p_table = new std::vector<std::vector<TVertex*>*>))
		return 0;

	if(!UpdateVertexHash())
		return false;
	// update polygon hash table if necessary

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = 0;
	// clear vertex flags

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end(); p_vertex ++) {
		int n_found = -1;
		if((*p_vertex)->n_flag)
			continue;
		// we've already added this one

		std::vector<TVertex*> *p_group;

		p_group = new std::vector<TVertex*>;
		if(p_group)
			p_table->push_back(p_group);
		if(!p_group || !p_table->size()) {
			if(!p_table->size())
				delete p_group;
			for(std::vector<std::vector<TVertex*>*>::iterator p_list =
			   p_table->begin(); p_list < p_table->end();)
				delete *(p_list ++);
			delete p_table;
			return 0;
		}
		// create a new vertex group & add it to group table

		std::vector<TVertex*> *p_other_vertices =
			m_p_vertex_hash_table[n_HashFunction((*p_vertex)->v_pos,
			m_n_vertex_hash_table_size, m_f_vertex_hash_table_cell_size)];
		// other vertices include current vertex as well so we don't need to explicitly add it

		for(std::vector<TVertex*>::const_iterator p_other_vertex = p_other_vertices->begin();
		   p_other_vertex < p_other_vertices->end(); p_other_vertex ++) {
			if(((*p_other_vertex)->v_pos - (*p_vertex)->v_pos).f_Length() < f_epsilon_ex) {
				if(b_compare_position_only ||
				   (*p_other_vertex)->b_Compare(**p_vertex, f_epsilon_ex)) {
					(*p_other_vertex)->n_flag = 1;
					// set flags so this vertex is not inserted again

					p_group->push_back(*p_other_vertex);
					if(!p_group->size()) {
						for(std::vector<std::vector<TVertex*>*>::iterator p_list =
						   p_table->begin(); p_list < p_table->end();)
							delete *(p_list ++);
						delete p_table;
					}
					// add to table
				}
			}
		}
		// search for similar vertices in table

		if(!p_group->size()) {
			p_table->erase(p_table->end() - 1);
			delete p_group;
		}
		// when there's vertex with NAN / INF coordinate, it can't be compared with the other
		// ones, neither with itself so it's not included in table
	}

	return p_table;
#endif //BRUTE_FORCE_WAY
}

/*
 *	bool CGeometry::OptimizeVertices(float f_epsilon_ex)
 *		- optimize mesh vertices, ie. merge the ones with the same position, normal & texcoords
 *		- you should call it after you're done, since it may weld vertices of different objects
 *		  and you could have trouble with further modelling
 *		- you can pass pointer to vertex hash table if you got one. otherwise enter 0 and
 *		  supply valid parameters (default ones will do well); cell overlap will be .01 * cell size
 */
bool CGeometry::OptimizeVertices(float f_epsilon_ex)
{
	if(!UpdatePolygonHash())
		return false;
	// update polygon hash table if necessary

	std::vector<std::vector<TVertex*>*> *p_merge_table;
	if(!(p_merge_table = p_VertexMergeTable(false, f_epsilon_ex)))
		return false;
	// create merge table ...

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = true;
	// clear vertex flags

	for(std::vector<std::vector<TVertex*>*>::const_iterator p_list = p_merge_table->begin();
	   p_list < p_merge_table->end(); p_list ++) {
		TVertex *p_new_vertex = (**p_list)[0];
		for(std::vector<TVertex*>::const_iterator p_deleted_vertex = (*p_list)->begin();
		   p_deleted_vertex < (*p_list)->end(); p_deleted_vertex ++) {
			(*p_deleted_vertex)->n_flag = false;
			std::vector<int> *p_hash = m_p_poly_hash_table[
				n_HashFunction((*p_deleted_vertex), m_n_poly_hash_table_size)];
			for(std::vector<int>::const_iterator p_poly_index = p_hash->begin();
			   p_poly_index < p_hash->end();) {
				CRefPolygon *p_poly = m_poly_list[*(p_poly_index ++)].p_poly;
				for(int i = 0; i < p_poly->n_Vertex_Num(); i ++) {
					if(p_poly->r_p_Vertex(i) == *p_deleted_vertex) {
						p_poly->r_p_Vertex(i) = p_new_vertex;
						m_b_poly_hash_dirty_flag = true;
					}
				}
			}
		}
		// mark vertices, replaced by other ones
	}
	// replace references to redundant vertices and mark used vertices

	for(std::vector<std::vector<TVertex*>*>::iterator p_list = p_merge_table->begin();
	   p_list < p_merge_table->end(); p_list ++)
		delete *(p_list ++);
	delete p_merge_table;
	// don't need merge table anymore

	if(!UpdatePolygonHash()) // right here
		return false;

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end(); p_vertex ++) {
		(*p_vertex)->n_flag = false;
		std::vector<int> *p_hash = m_p_poly_hash_table[
			n_HashFunction(*p_vertex, m_n_poly_hash_table_size)];
		for(std::vector<int>::const_iterator p_poly_index = p_hash->begin();
		   p_poly_index < p_hash->end();) {
			CRefPolygon *p_poly = m_poly_list[*(p_poly_index ++)].p_poly;
			for(int l = 0; l < p_poly->n_Vertex_Num(); l ++) {
				if(p_poly->r_p_Vertex(l) == *p_vertex) {
					(*p_vertex)->n_flag = true;
					break;
				}
			}
		}
		// mark vertices that are unused at all (i.e. not only those we replaced)
	}

	for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end(); p_vertex ++) {
		if(!(*p_vertex)->n_flag) {
			m_vertex_list.erase(&*(p_vertex --));
			// fixme - should i set polygon dirty flag as well? i see no reason now, because there was just pointer change, but no data really changed

			m_b_bounding_box_dirty_flag = true;
			m_b_poly_hash_dirty_flag = true;
			m_b_vertex_hash_dirty_flag = true;
		}
	}
	// delete unused vertices

	return true;
}

/*
 *	TVertex CGeometry::t_DerivateInterpolants(const TVertex *p_vertex,
 *		const Vector3f &r_v_direction) const
 *		- needed to extrapolate vertex coordinates in ContractEdge_UpdateNeighbours
 *		- returns vertex, containing delta-position (== r_v_direction), delta-normal,
 *		  delta-color and delta-texcoords in a given direction, n_flag is set to true
 *		- polygon hash must be available, this function doesn't refresh it
 *		- in case there is no polygon referencing given vertex, all coordinates of result
 *		  vertex are set to 0 and n_flag to false
 */
TVertex CGeometry::t_DerivateInterpolants(const TVertex *p_vertex,
	const Vector3f &r_v_direction) const
{
	const std::vector<int> *p_hash = m_p_poly_hash_table[n_HashFunction(p_vertex,
		m_n_poly_hash_table_size)];
	for(std::vector<int>::const_iterator p_poly_index = p_hash->begin();
	   p_poly_index < p_hash->end(); p_poly_index ++) {
		CRefPolygon *p_poly = m_poly_list[*p_poly_index].p_poly;
		for(int i = 0; i < p_poly->n_Vertex_Num(); i ++) {
			if(p_poly->r_p_Vertex(i) == p_vertex) {
				for(int a = 0; a < p_poly->n_Vertex_Num(); a ++) {
					if(a == i)
						continue;
					for(int b = 0; b < p_poly->n_Vertex_Num(); b ++) {
						if(b == i || b == a)
							continue;
						// try if edges i-a and i-b are not colinear

						Vector3f v_edge_0 = p_poly->t_Vertex(a).v_pos - p_poly->t_Vertex(i).v_pos;
						float f_length_0 = v_edge_0.f_Length();
						v_edge_0 *= 1 / f_length_0;
						Vector3f v_edge_1 = p_poly->t_Vertex(b).v_pos - p_poly->t_Vertex(i).v_pos;
						float f_length_1 = v_edge_1.f_Length();
						v_edge_1 *= 1 / f_length_1;

						if((float)fabs(v_edge_0.f_Dot(v_edge_1)) > 1.0f - f_epsilon)
							continue;
						// they are colinear, we can't use them

						float f_dot_0 = v_edge_0.f_Dot(r_v_direction);
						float f_dot_1 = v_edge_1.f_Dot(r_v_direction);

						TVertex t_vertex;
						t_vertex.n_flag = true;
						t_vertex.v_color = (p_poly->t_Vertex(a).v_color -
							p_poly->t_Vertex(i).v_color) * f_dot_0 +
							(p_poly->t_Vertex(b).v_color -
							p_poly->t_Vertex(i).v_color) * f_dot_1;
						t_vertex.v_normal = (p_poly->t_Vertex(a).v_normal -
							p_poly->t_Vertex(i).v_normal) * f_dot_0 +
							(p_poly->t_Vertex(b).v_normal -
							p_poly->t_Vertex(i).v_normal) * f_dot_1;
						t_vertex.v_pos = v_edge_0 * f_length_0 * f_dot_0 +
							v_edge_1 * f_length_1 * f_dot_1;
						for(int j = 0; j < max_Vertex_Texcoords; j ++) {
							t_vertex.v_texcoord[j] = (p_poly->t_Vertex(a).v_texcoord[j] -
								p_poly->t_Vertex(i).v_texcoord[j]) * f_dot_0 +
								(p_poly->t_Vertex(b).v_texcoord[j] -
								p_poly->t_Vertex(i).v_texcoord[j]) * f_dot_1;
						}
						// project coordinate vectors onto r_v_direction vector

						return t_vertex;
					}
				}
			}
		}
	}

	TVertex t_vertex;
	t_vertex.n_flag = false;
	t_vertex.v_color = Vector4f(0, 0, 0, 0);
	t_vertex.v_normal = Vector3f(0, 0, 0);
	t_vertex.v_pos = Vector3f(0, 0, 0);
	for(int i = 0; i < max_Vertex_Texcoords; i ++)
		t_vertex.v_texcoord[i] = Vector4f(0, 0, 0, 0);
	// fill with zeros

	return t_vertex;
}

/*
 *	void CGeometry::ContractEdge_UpdateNeighbours(TPolygon &r_t_polygon,
 *		int n_edge, float f_position, const std::vector<std::vector<TVertex*>*> *p_merge_table)
 *		- contract polygon edge of r_t_polygon and of all it's neighbours, update polygon
 *		  neighbour lists (if m_b_polygon_neighbours_dirty_flag == false).
 *		- final edge vertex will be between first edge vertex (as sorted in r_t_polygon,
 *		  (f_position == .0), and second edge vertex (f_position == 1.0)
 *		- each vertex's flag has to contain index of merge list, containing it
 *		- polygon hash must be available, this calls t_DerivateInterpolants function
 *		  which needs it, but doesn't refresh it (it will not be damaged so much we couldn't
 *		  reuse it so we can save plenty of time, we'd spend refreshing it)
 */
void CGeometry::ContractEdge_UpdateNeighbours(TPolygon &r_t_polygon,
	int n_edge, float f_position, const std::vector<std::vector<TVertex*>*> *p_merge_table)
{
	TVertex t_new_vertex;
	if(f_position == .0f)
		t_new_vertex = r_t_polygon.p_poly->t_Vertex(n_edge);
	else if(f_position == 1.0f) {
		t_new_vertex = r_t_polygon.p_poly->t_Vertex((n_edge + 1) %
			r_t_polygon.p_poly->n_Vertex_Num());
	} else {
		t_new_vertex = r_t_polygon.p_poly->t_Vertex(n_edge).Lerp(f_position,
			r_t_polygon.p_poly->t_Vertex((n_edge + 1) % r_t_polygon.p_poly->n_Vertex_Num()));
	}
	// calculate position of a new vertex

	for(int i = 0, n = n_edge; i < 2; i ++) {
		std::vector<TVertex*> *p_list =
			(*p_merge_table)[r_t_polygon.p_poly->r_p_Vertex(n)->n_flag];
		for(std::vector<TVertex*>::iterator p_vertex = p_list->begin();
		   p_vertex < p_list->end(); p_vertex ++) {
			TVertex t_delta_vertex = t_DerivateInterpolants(*p_vertex,
				t_new_vertex.v_pos - (*p_vertex)->v_pos);
			(*p_vertex)->v_pos = t_new_vertex.v_pos;
			//(*p_vertex)->v_pos += t_delta_vertex.v_pos; // fixme - must have the same effect as the line above
			(*p_vertex)->v_color += t_delta_vertex.v_color;
			(*p_vertex)->v_normal += t_delta_vertex.v_normal;
			(*p_vertex)->v_normal.Normalize();
			if((*p_vertex)->v_normal.f_Length() < f_epsilon)
				(*p_vertex)->v_normal = r_t_polygon.p_poly->t_Normal().v_normal;
			for(Vector4f *p_texcoord = (*p_vertex)->v_texcoord, *p_end = (*p_vertex)->v_texcoord +
			   max_Vertex_Texcoords, *p_src_coord = t_delta_vertex.v_texcoord; p_texcoord < p_end;)
				*(p_texcoord ++) += *(p_src_coord ++);
		}
		n = (n_edge + 1) % r_t_polygon.p_poly->n_Vertex_Num();
	}
	// calculate new coordinates for all the edge vertices

	for(std::vector<int>::const_iterator p_neigh_index =
	   r_t_polygon.p_neighbour_index_list[n_edge]->begin(); p_neigh_index <
	   r_t_polygon.p_neighbour_index_list[n_edge]->end(); p_neigh_index ++) {
		TPolygon *p_polygon = &m_poly_list[*p_neigh_index];
		int n_edge2 = 0;
		for(std::vector<int> **p_neigh2_list = p_polygon->p_neighbour_index_list,
		   **p_neigh2_end = p_polygon->p_neighbour_index_list + p_polygon->n_neighbour_edge_num;
		   p_neigh2_list < p_neigh2_end; p_neigh2_list ++, n_edge2 ++) {
			for(std::vector<int>::const_iterator p_neigh2_index = (*p_neigh2_list)->begin();
			   p_neigh2_index < (*p_neigh2_list)->end(); p_neigh2_index ++) {
				if(&m_poly_list[*p_neigh2_index] == &r_t_polygon) // todo - see if this works right (i guess it doesn't)
					break;
			}
		}
		// find on which edge r_t_polygon lies

		_ASSERTE(n_edge < p_polygon->n_neighbour_edge_num); // we found it, right?

		delete p_polygon->p_neighbour_index_list[n_edge2];
		for(int i = n_edge2; i < p_polygon->n_neighbour_edge_num - 1; i ++)
			p_polygon->p_neighbour_index_list[i] = p_polygon->p_neighbour_index_list[i + 1];
		p_polygon->n_neighbour_edge_num --;
		// remove neighbour list record

		p_polygon->p_poly->Delete_Vertices((n_edge2 + 1) % p_polygon->p_poly->n_Vertex_Num(), 1);
		// remove vertex // todo - do i have to remove the second one ? (would save inc, mod and function call)

		if(p_polygon->p_poly->n_Vertex_Num() < 3)
			DeletePolygons(*p_neigh_index);
		// in case we contracted edge, shared by triangle, delete it (the former triangle)
	}
	// remove single vertex from each neighbour's edge (the shared edge) and update neighbour list

	delete r_t_polygon.p_neighbour_index_list[n_edge];
	for(int i = n_edge; i < r_t_polygon.n_neighbour_edge_num - 1; i ++)
		r_t_polygon.p_neighbour_index_list[i] = r_t_polygon.p_neighbour_index_list[i + 1];
	r_t_polygon.n_neighbour_edge_num --;
	// remove neighbour list record

	r_t_polygon.p_poly->Delete_Vertices((n_edge + 1) % r_t_polygon.p_poly->n_Vertex_Num(), 1);
	// remove vertex // todo - do i have to remove the second one ? (would save inc, mod and function call)

	if(r_t_polygon.p_poly->n_Vertex_Num() < 3)
		DeletePolygons(&r_t_polygon - &m_poly_list[0]);
	// in case we contracted edge, shared by triangle, delete it (the former triangle)
}

bool CGeometry::OptimizeTris(const int *p_polygon_indices, int n_polygon_num, float f_epsilon_ex)
{
	if(!UpdatePolygonNeighbours())
		return false;
	// need some info on which vertex is neighbour of which one

	std::vector<std::vector<TVertex*>*> *p_merge_table;
	if(!(p_merge_table = p_VertexMergeTable(false, f_epsilon_ex)))
		return false;
	// create merge table ...

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = -1;
	// clear vertex flags

	{int n_index = 0;
	for(std::vector<std::vector<TVertex*>*>::const_iterator p_list = p_merge_table->begin();
	   p_list < p_merge_table->end(); p_list ++, n_index ++) {
		for(std::vector<TVertex*>::iterator p_vertex = (*p_list)->begin();
		   p_vertex < (*p_list)->end();)
			(*(p_vertex ++))->n_flag = n_index;
		// by the way, set vertex flags ...
	}}
	// set vertex flags to indices of merge table lists (when contracting edges, we're going to need to move all vertices)

	bool b_contracted;
	do {
		b_contracted = false;
		{int n_index = 0;
		for(std::vector<TPolygon>::iterator p_polygon = (p_polygon_indices)?
		   m_poly_list.begin() + (n_index = (*p_polygon_indices)) : m_poly_list.begin();
		   n_polygon_num > 0; n_polygon_num --, (p_polygon_indices)? p_polygon = m_poly_list.begin() +
		   (n_index = (*(++ p_polygon_indices))) : (n_index ++, p_polygon ++)) { // todo - see what g++ thinks of such a construct
			for(std::vector<int> **p_neigh_list = (*p_polygon).p_neighbour_index_list,
			   **p_neigh_end = (*p_polygon).p_neighbour_index_list +
			   (*p_polygon).n_neighbour_edge_num; p_neigh_list < p_neigh_end; p_neigh_list ++) {
				bool b_vertex_movable[2] = {true, true};
				// flags, telling us wheter it's safe to move edge vertices

				int n_edge_index = p_neigh_list - (*p_polygon).p_neighbour_index_list;
				Vector3f v_edge_org = (*p_polygon).p_poly->t_Vertex(n_edge_index).v_pos;
				Vector3f v_edge = v_edge_org -
					(*p_polygon).p_poly->t_Vertex((n_edge_index + 1) %
					(*p_polygon).n_neighbour_edge_num).v_pos;
				// get edge vectors

				TPolygon *p_cur_polygon = &*p_polygon;
				for(std::vector<int>::const_iterator p_neigh_index = (*p_neigh_list)->begin() - 1;
				   p_neigh_index < (*p_neigh_list)->end();
				   p_cur_polygon = &m_poly_list[*(p_neigh_index ++)]) {
					// loop trough every polygon, sharing edge (at least one)

					int n_cur_edge;
					int n_swap_vertices; // wheter vertices in the other poly comes in the other
					if(p_cur_polygon == &*p_polygon) {
						n_cur_edge = n_edge_index;
						n_swap_vertices = 0;
					} else {
						for(std::vector<int> **p_neigh_list2 =
						   p_cur_polygon->p_neighbour_index_list, **p_neigh_end2 =
						   p_cur_polygon->p_neighbour_index_list +
						   p_cur_polygon->n_neighbour_edge_num; p_neigh_list2 < p_neigh_end2;
						   p_neigh_list2 ++) {
							for(std::vector<int>::const_iterator p_neigh_index2 =
							   (*p_neigh_list2)->begin(); p_neigh_index2 <
							   (*p_neigh_list2)->end(); p_neigh_index2 ++) {
								if(*p_neigh_index2 == n_index) {
									n_cur_edge = p_neigh_list2 -
										p_cur_polygon->p_neighbour_index_list;
									p_neigh_index2 = (*p_neigh_list2)->end(); // break loop above
									break;
								}
							}
						}
						n_swap_vertices = (p_cur_polygon->p_poly->t_Vertex(n_cur_edge).n_flag !=
							(*p_polygon).p_poly->t_Vertex(p_neigh_list -
							(*p_polygon).p_neighbour_index_list).n_flag);
						// see if edge is oriented as in p_polygon
					}
					for(int i = 0, n_off = -1 + p_cur_polygon->n_neighbour_edge_num;
					   i < 2; i ++, n_off = 1) {
						const std::vector<int> *p_neigh_list2 =
							p_cur_polygon->p_neighbour_index_list[(n_cur_edge + n_off) %
							p_cur_polygon->n_neighbour_edge_num];
						// first see prev edge (set b_vertex_movable[0 ^ n_swap_vertices]),
						// then the second one (set b_vertex_movable[1 ^ n_swap_vertices])

						if(!p_neigh_list2->size()) {
							b_vertex_movable[i ^ n_swap_vertices] = false;
							continue;
						}
						// there is no neighbour at this edge, we'd change the shape boundary
						// (we can contract boundary edges, but not the ones leading from inside
						// to boundary)

						for(std::vector<int>::const_iterator p_neigh_index3 = p_neigh_list2->begin();
						   p_neigh_index3 < p_neigh_list2->end(); p_neigh_index3 ++) {
							Vector3f v_normal = m_poly_list[*p_neigh_index3].p_poly->t_Normal().v_normal;
							Vector3f v_edge2 = v_edge.v_Orthogonal(v_normal);
							if((v_edge - v_edge2).f_Length() > f_epsilon) {
								b_vertex_movable[i ^ n_swap_vertices] = false;
								break;
							}
						}
						// see if all neighbours are colinear with the edge (calculates edge,
						// prependicular to normal and meassure distance between it and the
						// original edge (approximately equals to surface lift))
					}
				}

				float f_position;
				if(b_vertex_movable[0] && b_vertex_movable[1])
					f_position = .5f;
				else if(b_vertex_movable[0])
					f_position = 0;
				else if(b_vertex_movable[1])
					f_position = 1;
				else
					continue;
				// calculate position of contracted vertex (or continue searching in case it's not
				// possible to move those vertices)

				ContractEdge_UpdateNeighbours(*p_polygon, n_edge_index, f_position, p_merge_table);
				// contract the edge

				break;
				// p_polygon could've been deleted // todo - handle it in a bit friendlier way
				// so algorithm can continue with next edge in case there's more than two of them
			}
		}}
		// search for edges to optimize
	} while(b_contracted);
	// loop while it can find any good edges to contract

	for(std::vector<std::vector<TVertex*>*>::iterator p_list = p_merge_table->begin();
	   p_list < p_merge_table->end(); p_list ++)
		delete *(p_list ++);
	delete p_merge_table;
	// don't need merge table anymore

	return true;
}

bool CGeometry::Alloc_PolygonNeighbourLists()
{
	for(std::vector<TPolygon>::iterator p_poly = m_poly_list.begin(); p_poly < m_poly_list.end();) {
		if((*p_poly).p_neighbour_index_list) {
			if((*p_poly).n_neighbour_edge_num != (*p_poly).p_poly->n_Vertex_Num()) {
				for(std::vector<int> **p_neighbour_index_list = (*p_poly).p_neighbour_index_list,
				   **p_end = (*p_poly).p_neighbour_index_list + (*p_poly).n_neighbour_edge_num;
				   p_neighbour_index_list < p_end;)
					delete *p_neighbour_index_list ++;
				delete[] (*p_poly).p_neighbour_index_list;
				(*p_poly).n_neighbour_edge_num = 0;
				(*p_poly).p_neighbour_index_list = 0;
			} else {
				for(std::vector<int> **p_neighbour_index_list = (*p_poly).p_neighbour_index_list,
				   **p_end = (*p_poly).p_neighbour_index_list + (*p_poly).n_neighbour_edge_num;
				   p_neighbour_index_list < p_end;)
					(*p_neighbour_index_list ++)->clear();
			}
		}
		if(!(*p_poly).p_neighbour_index_list) {
			(*p_poly).n_neighbour_edge_num = (*p_poly).p_poly->n_Vertex_Num();
			if(!((*p_poly).p_neighbour_index_list =
			   new std::vector<int>*[(*p_poly).n_neighbour_edge_num])) {
				(*p_poly).n_neighbour_edge_num = 0;
				return false;
			}
			for(std::vector<int> **p_neighbour_index_list = (*p_poly).p_neighbour_index_list,
			   **p_end = (*p_poly).p_neighbour_index_list + (*p_poly).n_neighbour_edge_num;
			   p_neighbour_index_list < p_end; p_neighbour_index_list ++) {
				if(!(*p_neighbour_index_list = new std::vector<int>)) {
					for(p_neighbour_index_list --; p_neighbour_index_list >=
					   (*p_poly).p_neighbour_index_list;)
						delete *p_neighbour_index_list --;
					delete[] (*p_poly).p_neighbour_index_list;
					(*p_poly).p_neighbour_index_list = 0;
					(*p_poly).n_neighbour_edge_num = 0;
					return false;
				}
			}
		}
		// alloc neighbour lists
	}

	return true;
}

struct TVertexLocation {
	int n_polygon_index;
	int n_vertex_index;

	int operator == (const TVertexLocation &r_loc) const
	{
		return n_polygon_index == r_loc.n_polygon_index &&
			n_vertex_index == r_loc.n_vertex_index;
	}
};

/*
 *	bool CGeometry::FindNeighbours()
 *		- find polygon neighbours
 *		- can return false in case there's not enough memory
 *		  for either neighbour-index tables or for one of look-up tables
 *		- polygon edge i is between vertex i and the next one
 */
bool CGeometry::FindNeighbours()
{
	if(!Alloc_PolygonNeighbourLists())
		return false;
	// alloc polygon neighbour lists so we can store our neighbour information

	std::vector<std::vector<TVertex*>*> *p_merge_table;
	if(!(p_merge_table = p_VertexMergeTable(true, f_epsilon)))
		return false;
	// create merge table, compare position only ...

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = -1;
	// clear vertex flags (there might be NAN / INF vertices that aren't in merge table)

	std::vector<std::vector<TVertexLocation>*> vertex_mapping_table;
	{int n_index = 0;
	for(std::vector<std::vector<TVertex*>*>::const_iterator p_merge_list = p_merge_table->begin();
	   p_merge_list < p_merge_table->end(); p_merge_list ++, n_index ++) {
		std::vector<TVertexLocation> *p_list;
		if(p_list = new std::vector<TVertexLocation>)
			vertex_mapping_table.push_back(p_list);
		if(!p_list || !vertex_mapping_table.size()) {
			if(p_list)
				delete p_list;
			for(std::vector<std::vector<TVertex*>*>::iterator p_list = p_merge_table->begin();
			   p_list < p_merge_table->end(); p_list ++)
				delete *(p_list ++);
			delete p_merge_table;
			for(std::vector<std::vector<TVertexLocation>*>::iterator p_loc =
			   vertex_mapping_table.begin(); p_loc < vertex_mapping_table.end();)
				delete *(p_loc ++);
			vertex_mapping_table.clear();
			return false;
		}

		for(std::vector<TVertex*>::iterator p_vertex = (*p_merge_list)->begin();
		   p_vertex < (*p_merge_list)->end();)
			(*(p_vertex ++))->n_flag = n_index;
		// by the way, set vertex flags ...
	}}
	// alloc table (size is equal to vertex-merge table) ...

	for(std::vector<TPolygon>::iterator p_poly = m_poly_list.begin(); p_poly < m_poly_list.end();) {
		_ASSERTE((*p_poly).p_neighbour_index_list);
		CRefPolygon *p_polygon = (*(p_poly ++)).p_poly;
		for(int i = 0; i < p_polygon->n_Vertex_Num(); i ++) {
			TVertexLocation t_loc;

			t_loc.n_polygon_index = p_poly - m_poly_list.begin();
			t_loc.n_vertex_index = i;

			int n_vertex_group = p_polygon->t_Vertex(i).n_flag;

			/*bool b_found = false;
			for(int k = 0; k < vertex_mapping_table[n_vertex_group]->size(); k ++) {
				if((*vertex_mapping_table[n_vertex_group])[k] == t_loc) {
					b_found = true;
					break;
				}
			}
			if(!b_found) {*/
				vertex_mapping_table[n_vertex_group]->push_back(t_loc);
				if(!vertex_mapping_table[n_vertex_group]->size()) {
					for(std::vector<std::vector<TVertex*>*>::iterator p_list =
					   p_merge_table->begin(); p_list < p_merge_table->end(); p_list ++)
						delete *(p_list ++);
					delete p_merge_table;
					for(std::vector<std::vector<TVertexLocation>*>::iterator p_loc =
					   vertex_mapping_table.begin(); p_loc < vertex_mapping_table.end();)
						delete *(p_loc ++);
					vertex_mapping_table.clear();
					return false;
				}
			/*} else
				; // f_ixme - does it ever happen, actually ? nope*/
		}
	}
	// for each vertex of each poly insert record into table, indexed by vertex group

	{int n_index = 0;
	for(std::vector<TPolygon>::iterator p_poly = m_poly_list.begin(); p_poly < m_poly_list.end();
	   p_poly ++, n_index ++) {
		for(int i = 0, n_prev_i = (*p_poly).p_poly->n_Vertex_Num() - 1;
		   i < (*p_poly).p_poly->n_Vertex_Num(); n_prev_i = i ++) {
			std::vector<TVertexLocation> *p_vertex_group =
				vertex_mapping_table[(*p_poly).p_poly->t_Vertex(i).n_flag];
			for(std::vector<TVertexLocation>::const_iterator p_location = p_vertex_group->begin();
			   p_location < p_vertex_group->end(); p_location ++) {
				if((*p_location).n_polygon_index == n_index) // fixme - is it == n_index or <= n_index ?
					continue;
				// don't want to add references to self and to polygons we've already processed

				CRefPolygon *p_poly2 = m_poly_list[(*p_location).n_polygon_index].p_poly;
				// p_poly2 has (*p_location).n_vertex_index-th vertex
				// on equal position with poly (*p_poly)

				int n_prev = (*p_location).n_vertex_index - 1;
				if(n_prev < 0)
					n_prev += p_poly2->n_Vertex_Num();
				int n_next = ((*p_location).n_vertex_index + 1) % p_poly2->n_Vertex_Num();
				_ASSERTE(p_poly2->t_Vertex((*p_location).n_vertex_index).n_flag ==
					(*p_poly).p_poly->t_Vertex(i).n_flag);
				if(p_poly2->t_Vertex(n_prev).n_flag == (*p_poly).p_poly->t_Vertex(n_prev_i).n_flag ||
				   p_poly2->t_Vertex(n_next).n_flag == (*p_poly).p_poly->t_Vertex(n_prev_i).n_flag) {
					(*p_poly).p_neighbour_index_list[n_prev_i]->push_back(
						(*p_location).n_polygon_index);
					if(!(*p_poly).p_neighbour_index_list[n_prev_i]->size()) {
						for(std::vector<std::vector<TVertex*>*>::iterator p_list =
						   p_merge_table->begin(); p_list < p_merge_table->end(); p_list ++)
							delete *(p_list ++);
						delete p_merge_table;
						for(std::vector<std::vector<TVertexLocation>*>::iterator p_loc =
						   vertex_mapping_table.begin(); p_loc < vertex_mapping_table.end();)
							delete *(p_loc ++);
						vertex_mapping_table.clear();
					}
				}
				// insert index of neighbour poly (in case we found neighbours)
			}
		}
	}}
	// build neighbour tables

	for(std::vector<std::vector<TVertex*>*>::iterator p_list =
	   p_merge_table->begin(); p_list < p_merge_table->end(); p_list ++)
		delete *(p_list ++);
	delete p_merge_table;
	for(std::vector<std::vector<TVertexLocation>*>::iterator p_loc =
	   vertex_mapping_table.begin(); p_loc < vertex_mapping_table.end();)
		delete *(p_loc ++);
	vertex_mapping_table.clear();
	// neither vertex mapping table
	// don't need merge table anymore

	m_b_poly_neighbours_dirty_flag = false;

	return true;
}

/*
 *	bool CGeometry::Break_SubMeshVertices(const int *p_polygon_indices, int n_polygon_num)
 *		- break vertices so selected polygons are no longer
 *		  vertex-connected to the rest of the mesh
 */
bool CGeometry::Break_SubMeshVertices(const int *p_polygon_indices, int n_polygon_num)
{
	if(!UpdatePolygonHash() || !UpdateVertexHash())
		return false;

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = false; // set "unused"
	// clear vertex flags

	{int n_index = 0;
	for(std::vector<TPolygon>::const_iterator p_polygon = (p_polygon_indices)?
	   m_poly_list.begin() + (n_index = (*p_polygon_indices)) : m_poly_list.begin();
	   n_polygon_num > 0; n_polygon_num --, (p_polygon_indices)? p_polygon = m_poly_list.begin() +
	   (n_index = (*(++ p_polygon_indices))) : (n_index ++, p_polygon ++)) { // todo - see what g++ thinks of such a construct
		CRefPolygon *p_poly = (*p_polygon).p_poly;
		for(int i = 0; i < p_poly->n_Vertex_Num(); i ++) {
			TVertex *p_vertex = p_poly->r_p_Vertex(i);
			if(p_vertex->n_flag)
				continue;
			// this vertex was marked as used already

			std::vector<int> *p_hash_table = m_p_poly_hash_table[n_HashFunction(
				p_poly->r_p_Vertex(i), m_n_poly_hash_table_size)];
			bool b_found_isolated_poly = false;
			for(std::vector<int>::const_iterator p_hash = p_hash_table->begin();
			   p_hash < p_hash_table->end(); p_hash ++) {
				if((*p_hash) == n_index)
					continue;
				// this polygon definitely will reference it's vertex

				CRefPolygon *p_potential_poly = m_poly_list[(*p_hash)].p_poly;
				bool b_has_vertex = false;
				for(int j = 0; j < p_potential_poly->n_Vertex_Num(); j ++) {
					if(p_potential_poly->r_p_Vertex(j) == p_vertex) {
						b_has_vertex = true;
						break;
					}
				}
				if(b_has_vertex) {
					b_found_isolated_poly = true;
					// we might have found it

					for(int j = 0; j < n_polygon_num; j ++) {
						if(((p_polygon_indices)? p_polygon_indices[j] : j) == (*p_hash)) {
							b_found_isolated_poly = false;
							break;
							// nope, it's in the list
						}
					}
				}
			}
			// try to find polygon, referencing current vertex,
			// that is outside of selected poly list

			if(b_found_isolated_poly)
				p_poly->r_p_Vertex(i)->n_flag = true;
			// this vertex is used by poly not in list
		}
	}}
	// mark vertices used in the rest of the mesh

	std::vector<TVertex*> duplicated_vertex_list;
	// vertices are stored in pairs, even (0th, 2nd, 4th, ...) is
	// original vertex and odd (1th, 3th, 5th, ...) is a new duplicate
	// this is bottleneck in this function because i'm performing linear
	// search trough this array. there could have been added next hash table
	// at the cost of some extra memory comsumption, of course

	for(std::vector<TPolygon>::const_iterator p_polygon = (p_polygon_indices)?
	   m_poly_list.begin() + (*p_polygon_indices) : m_poly_list.begin(); n_polygon_num > 0;
	   n_polygon_num --, (p_polygon_indices)? p_polygon = m_poly_list.begin() +
	   (*(++ p_polygon_indices)) : (p_polygon ++)) { // todo - see what g++ thinks of such a construct
		CRefPolygon *p_poly = (*p_polygon).p_poly;
		for(int i = 0; i < p_poly->n_Vertex_Num(); i ++) {
			TVertex *p_vertex = p_poly->r_p_Vertex(i);

			if(p_vertex->n_flag != false)
				continue;
			// vertices, referenced by selected polygons only don't have to be duplicated

			TVertex *p_duplicate_vertex = 0;

			for(int j = 0; j < duplicated_vertex_list.size(); j += 2) {
				if(duplicated_vertex_list[j] == p_vertex) {
					p_duplicate_vertex = duplicated_vertex_list[j + 1];
					break;
				}
			}
			// this vertex was duplicated already
			
			if(!p_duplicate_vertex) {
				m_b_poly_hash_dirty_flag = true;
				if(!(p_duplicate_vertex = p_AddVertex(*p_vertex)))
					return false;
				duplicated_vertex_list.push_back(p_vertex);
				duplicated_vertex_list.push_back(p_duplicate_vertex);
				if(!duplicated_vertex_list.size() || duplicated_vertex_list.size() % 2)
					return false;
			}
			// we didn't find it -> duplicate

			_ASSERTE(p_duplicate_vertex);
			p_poly->r_p_Vertex(i) = p_duplicate_vertex;
		}
	}

	return true;
}


/*
 *	bool CGeometry::Remove_DegenerateEdges(float f_edge_epsilon_ex, bool b_weld_vertices)
 *		- remove all edges, shorter than f_edge_epsilon_ex, if b_weld_vertices is true, move vertices
 *		  within f_edge_epsilon_ex to a single position (position of one of vertices, not average)
 */
bool CGeometry::Remove_DegenerateEdges(float f_edge_epsilon_ex, bool b_weld_vertices)
{
	std::vector<std::vector<TVertex*>*> *p_merge_table;
	if(!(p_merge_table = p_VertexMergeTable(true, f_edge_epsilon_ex)))
		return false;
	// create merge table ...

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = -1;
	// clear vertex flags

	{int n_index = 0;
	for(std::vector<std::vector<TVertex*>*>::const_iterator p_list = p_merge_table->begin();
	   p_list < p_merge_table->end(); p_list ++, n_index ++) {
		_ASSERTE((*p_list)->size());
		Vector3f v_weld_pos = (*(*p_list)->begin())->v_pos;
		for(std::vector<TVertex*>::iterator p_vertex = (*p_list)->begin();
		   p_vertex < (*p_list)->end(); p_vertex ++) {
			(*p_vertex)->n_flag = n_index;
			if(b_weld_vertices) {
				(*p_vertex)->v_pos = v_weld_pos;
				m_b_vertex_hash_dirty_flag = true; // moved vertices
				//m_b_bounding_box_dirty_flag = true; // but to positions of other ones
			}
		}
		delete *(p_list ++);
	}}
	// set vertex flags to positions, weld if required

	delete p_merge_table;
	// don't need merge table anymore

	for(std::vector<TPolygon>::iterator p_polygon = m_poly_list.begin();
	   p_polygon < m_poly_list.end(); p_polygon ++) {
		CRefPolygon *p_poly = (*p_polygon).p_poly;
		for(int i = 0, n_prev_i = p_poly->n_Vertex_Num() - 1;
		   i < p_poly->n_Vertex_Num(); n_prev_i = i ++) {
			if(p_poly->r_p_Vertex(i)->n_flag == -1 ||
			   p_poly->r_p_Vertex(i)->n_flag == p_poly->r_p_Vertex(n_prev_i)->n_flag) {
				p_poly->Delete_Vertices(i --, 1);
				m_b_poly_hash_dirty_flag = true; // changed vertex mapping
				m_b_poly_neighbours_dirty_flag = true; // removed edges
			}
		}
	}
	// erase degenerate edges

	return true;
}

/*
 *	struct TEdgeLine
 *		- structure, keeping list of colinear edges
 *		- used in CGeometry::Remove_TJunctions()
 */
struct TEdgeLine {
	struct TEdge {
		TVertex *p_vertex[2];
		CRefPolygon *p_polygon;
		int n_vertex_index; // index of first vertex in polygon (second is the next one)
		// note index changes when you add vertices, causing t-junctions
	};

	struct TEdgeVertex {
		TVertex *p_vertex;
		float f_position; // position along edge
		std::vector<TEdgeVertex> next_vertex_list; // next vertices lying on the same position

		TEdgeVertex(TVertex *_p_vertex, float _f_position)
			:p_vertex(_p_vertex), f_position(_f_position), next_vertex_list()
		{}
	};

protected:
	std::vector<TEdgeVertex> vertex_list; // list of all vertices, lying on this edge
	std::vector<TEdge*> edge_list; // list of all edges from all polygons
	Vector3f v_begin, v_axis; // normalized axis of the edge (needs to be normalized
	// so we can use epsilons - otherwise we'd have to scale them)

public:
	/*
	 *	TEdgeLine::TEdgeLine(TEdge *p_edge)
	 *		- default constructor, parameter is first edge to be filled in
	 *		- if constructor succeeds, n_Edge_Num() must return 1 and n_DifferentVertex_Num() 2
	 *		- do not insert degenerate edge, otherwise edge direction could be zero vector
	 *		  which would result in all other edges being added here and connected together
	 *		- p_edge will be deleted by TEdgeLine destructor
	 */
	TEdgeLine(TEdge *p_edge)
	{
		v_begin = p_edge->p_vertex[0]->v_pos;
		v_axis = p_edge->p_vertex[1]->v_pos - p_edge->p_vertex[0]->v_pos;
		v_axis.Normalize();

		edge_list.push_back(p_edge);
		// insert edge

		for(int i = 0; i < 2; i ++) {
			vertex_list.front().next_vertex_list.push_back(
				TEdgeVertex(p_edge->p_vertex[i], f_Vertex_Pos(p_edge->p_vertex[i]->v_pos)));
		}
		// insert vertices
	}

	/*
	 *	TEdgeLine::~TEdgeLine()
	 *		- default destructor
	 */
	~TEdgeLine()
	{
		for(std::vector<TEdge*>::iterator p_edge = edge_list.begin(); p_edge < edge_list.end();)
			delete *p_edge ++;
	}

	/*
	 *	int TEdgeLine::n_Edge_Num() const
	 *		- returns number of edges (there must be min 1)
	 */
	int n_Edge_Num() const { return edge_list.size(); }

	/*
	 *	int TEdgeLine::n_DifferentVertex_Num() const
	 *		- returns number of different vertices (must be at least 2)
	 */
	int n_DifferentVertex_Num() const { return vertex_list.size(); }

	/*
	 *	inline float TEdgeLine::f_Vertex_Pos(const Vector3f &r_v_pos) const
	 *		- returns position of vertex along the edge
	 */
	inline float f_Vertex_Pos(const Vector3f &r_v_pos) const
	{
		return v_axis.f_Dot(r_v_pos - v_begin);
	}

	/*
	 *	inline Vector3f TEdgeLine::v_Edge_Vertex(float f_position) const
	 *		- returns position of vertex with position f_position along the edge,
	 *		  meassured from first vertex of first edge inserted here
	 */
	inline Vector3f v_Edge_Vertex(float f_position) const
	{
		return v_axis * f_position + v_begin;
	}

	/*
	 *	bool TEdgeLine::b_Colinear(const TEdge &r_t_edge, float f_epsilon_ex) const
	 *		- returns true in case r_t_edge is colinear with this edgeline
	 *		- f_epsilon_ex is maximal (positional, not angular) deviation of edge points from axis
	 */
	bool b_Colinear(const TEdge &r_t_edge, float f_epsilon_ex) const
	{
		float f_pos_0 = f_Vertex_Pos(r_t_edge.p_vertex[0]->v_pos);
		float f_pos_1 = f_Vertex_Pos(r_t_edge.p_vertex[1]->v_pos);
		// calculate positions along this edgeline

		/*
		if((f_distance_0 <= vertex_list.front().f_position &&
		   f_distance_1 <= vertex_list.front().f_position) ||
		   (f_distance_0 >= vertex_list.back().f_position &&
		   f_distance_1 >= vertex_list.back().f_position))
			return false;
		// if both vertices are on one side of edge axis, but not overlaping it,
		// then this edge comes elsewhere
		*/
		// we have to keep all colinear edges, otherwise we could end up with multiple edgelines
		// that should be connected, but because of order of adding edges, aren't

		float f_distance_0 = (r_t_edge.p_vertex[0]->v_pos - v_Edge_Vertex(f_pos_0)).f_Length();
		if(f_distance_0 > f_epsilon_ex)
			return false;
		float f_distance_1 = (r_t_edge.p_vertex[1]->v_pos - v_Edge_Vertex(f_pos_1)).f_Length();
		if(f_distance_1 > f_epsilon_ex)
			return false;
		// calculate distances to the edge

		return true;
	}

	/*
	 *	bool TEdgeLine::AddEdge(TEdge *p_edge, float f_epsilon_ex, float f_edge_epsilon_ex,
	 *		bool b_weld_vertices, bool b_stick_vertices_to_edge)
	 *		- adds edge p_edge into the list of edges, and it's vertices into the list of
	 *		  vertices, which is kept sorted by distance along the edge
	 *		- f_epsilon_ex is maximal (positional, not angular) deviation of edge points from axis
	 *		- f_edge_epsilon_ex is minimal distance along the edge between vertices to form edge
	 *		  (otherwise they are considered occupying the same position)
	 *		- if b_weld_vertices is true, then vertices with distance below f_edge_epsilon_ex
	 *		  are moved to the same position (other vertex coordinates are kept untouched)
	 *		- if b_stick_vertices_to_edge is true, vertices are moved towards their
	 *		  prependicular position on axis
	 *		- returns true, if there was not enough memory, returns false
	 *		- p_edge will be deleted by TEdgeLine destructor
	 */
	bool AddEdge(TEdge *p_edge, float f_epsilon_ex, float f_edge_epsilon_ex,
		bool b_weld_vertices, bool b_stick_vertices_to_edge)
	{
		_ASSERTE(b_Colinear(*p_edge, f_epsilon_ex)); // edge must be colinear

		for(int i = 0; i < 2; i ++) {
			float f_vertex_pos = f_Vertex_Pos(p_edge->p_vertex[i]->v_pos);
			// calculate position along this edgeline

			if(b_stick_vertices_to_edge)
				p_edge->p_vertex[i]->v_pos = v_Edge_Vertex(f_vertex_pos);
			// stick to axe if required

			if(f_vertex_pos < vertex_list.front().f_position - f_edge_epsilon_ex) {
				vertex_list.insert(vertex_list.begin(),
					TEdgeVertex(p_edge->p_vertex[i], f_vertex_pos));
				if(!vertex_list.size())
					return false;
				continue;
			} else if(f_vertex_pos > vertex_list.back().f_position + f_edge_epsilon_ex) {
				vertex_list.push_back(TEdgeVertex(p_edge->p_vertex[i], f_vertex_pos));
				if(!vertex_list.size())
					return false;
				continue;
			}
			// in case vertex lies somewehere outside edge

			float f_prev_pos = vertex_list.front().f_position, f_pos;
			if(f_vertex_pos > f_prev_pos - f_edge_epsilon_ex &&
			   f_vertex_pos < f_prev_pos + f_edge_epsilon_ex) {
				if(b_weld_vertices)
					p_edge->p_vertex[i]->v_pos = vertex_list.front().p_vertex->v_pos;
				vertex_list.front().next_vertex_list.push_back(
					TEdgeVertex(p_edge->p_vertex[i], f_vertex_pos));
				if(!vertex_list.front().next_vertex_list.size())
					return false;
				continue;
			}
			// in case positions are equal, doesn't try to sort them (if would cause
			// degenerate edges) adds them into list of the same vertices instead

			for(std::vector<TEdgeVertex>::iterator p_prev_vertex = vertex_list.begin(),
			   p_vertex = vertex_list.begin() + 1; p_vertex < vertex_list.end();
			   p_prev_vertex = p_vertex ++, f_prev_pos = f_pos) {
				f_pos = (*p_vertex).f_position;
				if(f_vertex_pos > f_pos - f_edge_epsilon_ex && f_vertex_pos < f_pos + f_edge_epsilon_ex) {
					if(b_weld_vertices)
						p_edge->p_vertex[i]->v_pos = (*p_vertex).p_vertex->v_pos;
					(*p_vertex).next_vertex_list.push_back(
						TEdgeVertex(p_edge->p_vertex[i], f_vertex_pos));
					if(!(*p_vertex).next_vertex_list.size())
						return false;
					break;
				}
				// in case positions are equal, doesn't try to sort them (if would cause
				// degenerate edges) adds them into list of the same vertices instead

				if(f_vertex_pos > f_prev_pos && f_vertex_pos < f_pos) {
					vertex_list.insert(p_vertex, TEdgeVertex(p_edge->p_vertex[i], f_vertex_pos)); // insert !after!
					if(!vertex_list.size())
						return false;
					break;
				}
				// in case it lies between two vertices, add them to it
			}
		}

		edge_list.push_back(p_edge);
		if(!edge_list.size())
			return false;
		// add edge

		return true;
	}

	/*
	 *	inline std::vector<TEdge*>::const_iterator TEdgeLine::p_EdgeList_Begin()
	 *		- returns begin iterator of edge list
	 */
	inline std::vector<TEdge*>::const_iterator p_EdgeList_Begin()
	{
		return edge_list.begin();
	}

	/*
	 *	inline std::vector<TEdge*>::const_iterator TEdgeLine::p_EdgeList_End()
	 *		- returns end iterator of edge list
	 */
	inline std::vector<TEdge*>::const_iterator p_EdgeList_End()
	{
		return edge_list.end();
	}

	/*
	 *	bool TEdgeLine::GetEdgeVertices(const TVertex *p_vertex_0, const TVertex *p_vertex_1,
	 *		std::vector<TEdgeVertex>::iterator &r_p_begin,
	 *		std::vector<TEdgeVertex>::iterator &r_p_end, bool &r_b_reverse)
	 *		- finds where p_vertex_0 and p_vertex_1 are in vertex list
	 *		- fills iterators r_p_begin (contains TEdgeVertex with p_vertex_0),
	 *		  r_p_end (contains TEdgeVertex with p_vertex_1) and boolean r_b_reverse
	 *		  (if r_b_reverse is false, r_p_end has bigger value than r_p_begin, otherwise
	 *		  we have to count backwards)
	 *		- returns false if it couldn't find given vertices, otherwise true
	 *		- intended use is to build list of edgelines of all polygons and then for each edge
	 *		  iterate r_p_begin to r_p_end, inserting vertices to the polygon (not r_p_begin or
	 *		  r_p_end themselves, they point to vertices of the edge itself)
	 */
	bool GetEdgeVertices(const TVertex *p_vertex_0, const TVertex *p_vertex_1,
		std::vector<TEdgeVertex>::iterator &r_p_begin,
		std::vector<TEdgeVertex>::iterator &r_p_end, bool &r_b_reverse)
	{
		bool b_found_0 = false;
		bool b_found_1 = false;
		for(std::vector<TEdgeVertex>::iterator p_line_vertex = vertex_list.begin();
		   p_line_vertex < vertex_list.end(); p_line_vertex ++) {
			for(int i = 0; i < 2; i ++) {
				bool b_found = false;
				if((*p_line_vertex).p_vertex == ((i)? p_vertex_1 : p_vertex_0))
					b_found = true;
				else {
					for(std::vector<TEdgeVertex>::iterator p_weld_vertex =
					   (*p_line_vertex).next_vertex_list.begin(); p_weld_vertex <
					   (*p_line_vertex).next_vertex_list.end(); p_weld_vertex ++) {
						if((*p_weld_vertex).p_vertex == ((i)? p_vertex_1 : p_vertex_0)) {
							b_found = true;
							break;
						}
					}
				}
				if(b_found) {
					if(i) {
						if(!b_found_0)
							r_b_reverse = true;
						else
							r_b_reverse = false;
						// if second is found sooner than first, have to iterate backwards

						r_p_end = p_line_vertex;

						b_found_1 = true;
						if(b_found_0)
							return true;
						// do we have both?
					} else {
						r_p_begin = p_line_vertex;
						
						b_found_0 = true;
						if(b_found_1)
							return true;
						// do we have both?
					}
				}
			}
		}
		// find first and last vertex of the edge

		return false;
	}
};

/*
 *	bool CGeometry::Remove_TJunctions(float f_epsilon_ex, float f_edge_epsilon_ex,
 *		bool b_weld_vertices, bool b_stick_vertices_to_edge)
 *		- remove T-junctions (T-junction is edge of one polygon, lying on the edge of other poly,
 *		  but their end vertices aren't on the same distance along the edge, yielding thin gap
 *		  artifact when rasterized)
 *		- f_epsilon_ex is max distance of vertex from edge to be considered on-edge, edges shorter
 *		  than f_edge_epsilon_ex are ignored
 *		- if b_weld_vertices is true, vertices within b_weld_vertices will be moved to single
 *		  position
 *		- if b_stick_vertices_to_edge is true, vertices will be moved towards edge axis,
 *		  it might eliminate some rasterizer artifacts later on
 */
bool CGeometry::Remove_TJunctions(float f_epsilon_ex, float f_edge_epsilon_ex,
	bool b_weld_vertices, bool b_stick_vertices_to_edge)
{
	std::vector<TEdgeLine*> edge_list;

	for(std::vector<TPolygon>::iterator p_polygon = m_poly_list.begin();
	   p_polygon < m_poly_list.end(); p_polygon ++) {
		CRefPolygon *p_poly = (*p_polygon).p_poly;
		for(int i = 0, n_prev_i = p_poly->n_Vertex_Num() - 1;
		   i < p_poly->n_Vertex_Num(); n_prev_i = i ++) {
			if((p_poly->t_Vertex(n_prev_i).v_pos -
			   p_poly->t_Vertex(i).v_pos).f_Length() < f_edge_epsilon_ex)
				continue;
			// ignore short edges

			TEdgeLine::TEdge *p_edge;
			if(!(p_edge = new TEdgeLine::TEdge)) {
				for(std::vector<TEdgeLine*>::iterator p_edgeline = edge_list.begin();
				   p_edgeline < edge_list.end();)
					delete *p_edgeline ++;
				return false;
			}
			p_edge->n_vertex_index = n_prev_i;
			p_edge->p_polygon = p_poly;
			p_edge->p_vertex[0] = p_poly->r_p_Vertex(n_prev_i);
			p_edge->p_vertex[1] = p_poly->r_p_Vertex(i);
			// fill edge struct

			bool b_found = false;
			for(std::vector<TEdgeLine*>::iterator p_edgeline = edge_list.begin();
			   p_edgeline < edge_list.end(); p_edgeline ++) {
				if((*p_edgeline)->b_Colinear(*p_edge, f_epsilon_ex)) {
					if(!(*p_edgeline)->AddEdge(p_edge, f_epsilon_ex, f_edge_epsilon_ex,
					   b_weld_vertices, b_stick_vertices_to_edge)) {
						delete p_edge;
						for(std::vector<TEdgeLine*>::iterator p_edgeline = edge_list.begin();
						   p_edgeline < edge_list.end();)
							delete *p_edgeline ++;
						return false;
					}

					b_found = true;
					break;
				}
			}
			// either find colinear edgeline and add edge to it

			if(!b_found) {
				TEdgeLine *p_new_edgeline;
				if(!(p_new_edgeline = new TEdgeLine(p_edge)) ||
				   !p_new_edgeline->n_Edge_Num() || p_new_edgeline->n_DifferentVertex_Num() < 2) {
					if(p_new_edgeline)
						delete p_new_edgeline; // p_edge is deleted by p_new_edgeline destructor
					else
						delete p_edge;
					for(std::vector<TEdgeLine*>::iterator p_edgeline = edge_list.begin();
					   p_edgeline < edge_list.end();)
						delete *p_edgeline ++;
					return false;
				}

				edge_list.push_back(p_new_edgeline);
				if(!edge_list.size()) {
					delete p_new_edgeline;
					for(std::vector<TEdgeLine*>::iterator p_edgeline = edge_list.begin();
					   p_edgeline < edge_list.end();)
						delete *p_edgeline ++;
					return false;
				}
			}
			// or create a new edgeline
		}
	}
	// create list of edgelines. now we have to insert t-junction vertices into the polys

	for(std::vector<TEdgeLine*>::const_iterator p_edgeline = edge_list.begin();
	   p_edgeline < edge_list.end(); p_edgeline ++) {
		for(std::vector<TEdgeLine::TEdge*>::const_iterator p_edge =
		   (*p_edgeline)->p_EdgeList_Begin(); p_edge <
		   (*p_edgeline)->p_EdgeList_End(); p_edge ++) {
			CRefPolygon *p_poly = (*p_edge)->p_polygon;
			std::vector<TEdgeLine::TEdgeVertex>::iterator p_begin;
			std::vector<TEdgeLine::TEdgeVertex>::iterator p_end;
			bool b_reverse;
			if(!(*p_edgeline)->GetEdgeVertices((*p_edge)->p_vertex[0], (*p_edge)->p_vertex[1],
			   p_begin, p_end, b_reverse)) {
				for(std::vector<TEdgeLine*>::iterator p_edgeline = edge_list.begin();
				   p_edgeline < edge_list.end();)
					delete *p_edgeline ++;
				return false;
			}
			// get list of vertices to be inserted

			int n_index_of_first_edge_vertex = -1;
			for(int i = (*p_edge)->n_vertex_index; i < p_poly->n_Vertex_Num(); i ++) {
				if(p_poly->r_p_Vertex(i) == (*p_edge)->p_vertex[0]) {
					n_index_of_first_edge_vertex = i;
					break;
				}
			}
			_ASSERTE(n_index_of_first_edge_vertex != -1);
			// find first edge vertex in polygon so we know where to insert new vertices (after it)

			float f_t1 = (*p_begin).f_position, f_t2 = (*p_end).f_position;
			for((b_reverse)? p_begin -- : p_begin ++; (b_reverse)? p_end > p_begin :
			   p_begin < p_end; (b_reverse)? p_begin -- : p_begin ++) {
				float f_ratio = ((*p_begin).f_position - f_t1) / (f_t2 - f_t1); // ratio for texcoord interpolation
				TVertex t_lerp_vertex = (*p_edge)->p_vertex[0]->Lerp(f_ratio,
					*(*p_edge)->p_vertex[1]);
				t_lerp_vertex.v_pos = (*p_begin).p_vertex->v_pos; // want precise position
				TRefVertex t_new_vertex = t_AddRefVertex(t_lerp_vertex);
				if(!t_new_vertex.p_vertex ||
				   !p_poly->Add_Vertex(++ n_index_of_first_edge_vertex, t_new_vertex)) {
					for(std::vector<TEdgeLine*>::iterator p_edgeline = edge_list.begin();
					   p_edgeline < edge_list.end();)
						delete *p_edgeline ++;
					return false;
				}
				// create a new vertex and add it to the poly
				// n_index_of_first_edge_vertex created index of first edge vertex, we need
				// to insert new vertices after it, but Add_Vertex adds vertices before index
				// also we need add next edge vertex after this one -> ++ n_index_of_first_edge_vertex
			}
		}
	}
	// insert t-junction vertices into the polys

	for(std::vector<TEdgeLine*>::iterator p_edgeline = edge_list.begin();
	   p_edgeline < edge_list.end();)
		delete *p_edgeline ++;
	// cleanup

	return true;
}

/*
 *	bool CGeometry::Generate_Texcoords(ECoordMethod n_method,
 *		const int *p_polygon_indices, int n_polygon_num,
 *		int n_tex_channel, int n_s_coord, int n_t_coord, int n_r_coord, int n_q_coord,
 *		const Matrix4f &m_world_transform, Vector3f v_scale)
 *		- generate texture coordinates using certain method (returns false for unknown method),
 *		  for defined group of polygons, into certain texture channel (currently 0 - 3),
 *		  you can choose where every texture coordinate goes (indices 0 - 3, -1 = disable),
 *		  then there is matrix, used to define position of gizmo and vector with mapping scale
 *		- return false on unknown method, on bad texture channel / coordinate index,
 *		  it doesn't check wheter polygon indices are correct
 *
 *	for n_method coord_XYZ_to_STR: (1 possible output for 1 vertex position)
 *		- s, t, r are x, y, z of vertex transformed by m_world_transform
 *		- w is always 1.0 in case you want it to be generated
 *
 *	for n_method coord_Planar: (1 possible output for 1 vertex position)
 *		- s, t are x, y of vertex transformed by m_world_transform and projected to x-y plane
 *		- r is always 0.0 (projection is on plane) and w is always 1.0
 *
 *	for n_method coord_Box:
 *		- s, t are x, y of vertex transformed by m_world_transform and projected to plane,
 *		  closest to polygon plane
 *		- r is always 0.0 (projection is on plane) and w is always 1.0
 *
 *	for n_method coord_Cylinder: (1 possible output for 1 vertex position)
 *		- s, t are x, y of vertex transformed by m_world_transform and projected to
 *		  coat of unit cylinder. it has no cap, though ...
 *		- r is always 0.0 (projection is on plane) and w is always 1.0
 *
 *	for n_method coord_Sphere: (1 possible output for 1 vertex position)
 *		- s, t, r are x, y, z of vertex transformed by m_world_transform and projected to
 *		  coat of unit sphere
 *		- w is always 1.0
 *
 *	for n_method coord_Tangent_of_x and coord_BiTangent_of_x:
 *		- n_s_coord and n_t_coord indicates source texture coordinates in texture channel
 *		  (chosen by method name)
 *		- n_r_coord is where to put desired vector (tangent or bitangent) and n_q_coord
 *		  is where to put texture polarity (either 1.0 or -1.0)
 *		- when n_r_coord == -1 or n_r_coord > 1, function returns false (n_q_coord can be -1)
 *		  the purpose for != -1 is you have to generate that coordinates in order to determine
 *		  texture polarity, the purpose for <= 1 is result is three-dimensional vector and it
 *		  must fit into four-dimensional vector
 *		- for example call with params coord_Tangent_of_1 and s = 0, t = 1, r = 0, q = 3
 *		  calculates tangent vector for first two texture coordinates from first texture
 *		  channel, resulting vector will be stored to coords 0, 1, 2 and polarity to 3
 *		- might be fuzzy for polygon vertices (here polygon means not triangle)
 *		- since coord_BiTangent_of_x is purely for the sake of completeness (since bitangent
 *		  can be calculated using cross product of tangent and normal (as we recommend to
 *		  ensure orthogonality of t / b / n vectors) all functionality, concerning
 *		  coord_BiTangent_of_x can be stripped away (64k app hint)
 *		- never use both coord_Tangent_of_x and coord_BiTangent_of_x, they may not be
 *		  prependicular, rather use cross product
 */
bool CGeometry::Generate_Texcoords(ECoordMethod n_method,
	const int *p_polygon_indices, int n_polygon_num,
	int n_tex_channel, int n_s_coord, int n_t_coord, int n_r_coord, int n_q_coord,
	const Matrix4f &m_world_transform, Vector3f v_scale)
{
	if(n_method < coord_XYZ_to_STR || (n_method > coord_Tangent_of_1 + max_Vertex_Texcoords &&
	   n_method < coord_BiTangent_of_1) || n_method > coord_BiTangent_of_1 + max_Vertex_Texcoords ||
	   n_tex_channel < 0 || n_tex_channel >= max_Vertex_Texcoords ||
	   n_s_coord < -1 || n_s_coord > 3 || n_t_coord < -1 || n_t_coord > 3 ||
	   n_r_coord < -1 || n_r_coord > 3 || n_q_coord < -1 || n_q_coord > 3)
		return false;
	// see if input data are all right

	if(n_method >= coord_Tangent_of_1 &&
	   n_method < coord_BiTangent_of_1 + max_Vertex_Texcoords) {
		if(n_r_coord == -1 || n_r_coord > 1)
			return false;
		// you HAVE to generate tangent/bitangent vector and it's index
		// can be either 0 or 1 only, because it has three coordinates

		if(!Break_SubMeshVertices(p_polygon_indices, n_polygon_num))
			return false;
	}
	// these functions generates texture coordinates iteratively
	// (it means it takes more than one step to calculate coordinates of single vertex)
	// so they need to have ensured they won't have to duplicate vertices during process
	// since it would be hard to determine which vertices has been already duplicated

	if(!UpdatePolygonHash() || !UpdateVertexHash())
		return false;

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = false; // set "unused"
	// clear vertex flags

	int n_mesh_vertex_num = m_vertex_list.size();

	{int n_index = 0;
	for(std::vector<TPolygon>::const_iterator p_polygon = (p_polygon_indices)?
	   m_poly_list.begin() + (n_index = (*p_polygon_indices)) : m_poly_list.begin();
	   n_polygon_num > 0; n_polygon_num --, (p_polygon_indices)? p_polygon = m_poly_list.begin() +
	   (n_index = (*(++ p_polygon_indices))) : (n_index ++, p_polygon ++)) { // todo - see what g++ thinks of such a construct
		CRefPolygon *p_poly = (*p_polygon).p_poly;
		for(int i = 0; i < p_poly->n_Vertex_Num(); i ++) {
			if(p_poly->t_Vertex(i).n_flag)
				continue;
			// this vertex was marked as used already

			std::vector<int> *p_hash_table = m_p_poly_hash_table[n_HashFunction(
				p_poly->r_p_Vertex(i), m_n_poly_hash_table_size)];
			// get hash table

			bool b_found_isolated_poly = false;
			for(std::vector<int>::const_iterator p_index = p_hash_table->begin();
			   p_index < p_hash_table->end(); p_index ++) {
				if(*p_index == n_index)
					continue;
				// this polygon definitely will reference it's vertex

				CRefPolygon *p_potential_poly = m_poly_list[*p_index].p_poly;
				bool b_has_vertex = false;
				for(int j = 0; j < p_potential_poly->n_Vertex_Num(); j ++) {
					if(p_potential_poly->r_p_Vertex(j) == p_poly->r_p_Vertex(i)) {
						b_has_vertex = true;
						break;
					}
				}
				if(b_has_vertex) {
					b_found_isolated_poly = true;
					// we might have found it

					for(int j = 0; j < n_polygon_num; j ++) {
						if(((p_polygon_indices)? p_polygon_indices[j] : j) == *p_index) {
							b_found_isolated_poly = false;
							break;
						}
					}
					// nope, it's in the list
				}
			}
			// try to find polygon, referencing current vertex, that is not in list

			if(b_found_isolated_poly) {
				p_poly->r_p_Vertex(i)->n_flag = true;
				n_mesh_vertex_num --;
			}
			// this vertex is used so when we want to change it's
			// texture coords, he have to duplicate it
		}

	}}
	// mark vertices whose texture coordinate mustn't be changed
	// (i.e. vertices, used by polygons not in list; might be slow)

	Vector3f *p_helper_data = 0;

	if(n_method >= coord_Tangent_of_1 &&
	   n_method < coord_BiTangent_of_1 + max_Vertex_Texcoords && n_q_coord != -1) {
		if(!(p_helper_data = new Vector3f[n_mesh_vertex_num]))
			return false;
		for(int i = 0, j = 0; i < m_vertex_list.size(); i ++) {
			if(m_vertex_list[i]->n_flag == false) {
				p_helper_data[j] = Vector3f(0, 0, 0);
				m_vertex_list[i]->n_flag = j ++;
				_ASSERTE(j <= n_mesh_vertex_num);
			} else
				m_vertex_list[i]->n_flag = -1;
		}
		// we need second temporal array for bi-tangents to determine texture polarity
	}

	for(std::vector<TPolygon>::const_iterator p_polygon = (p_polygon_indices)?
	   m_poly_list.begin() + (*p_polygon_indices) : m_poly_list.begin(); n_polygon_num > 0;
	   n_polygon_num --, (p_polygon_indices)? p_polygon = m_poly_list.begin() +
	   (*(++ p_polygon_indices)) : p_polygon ++) {
		CRefPolygon *p_poly = (*p_polygon).p_poly;

		Vector3f v_s_dir;
		Vector3f v_t_dir;
		if(n_method >= coord_Tangent_of_1 && n_method < coord_BiTangent_of_1 + max_Vertex_Texcoords) {
			int n_tex_channel = (int)n_method - (int)((n_method < coord_Tangent_of_1 +
				max_Vertex_Texcoords)? coord_Tangent_of_1 : coord_BiTangent_of_1);
			// source texture channel

			Vector3f v_delta_pos1 = p_poly->r_p_Vertex(0)->v_Pos() - p_poly->r_p_Vertex(1)->v_Pos();
			Vector3f v_delta_pos2;
			int n = p_poly->n_Vertex_Num() - 1;
			for(int i = 2; i < p_poly->n_Vertex_Num(); i ++) {
				v_delta_pos2 = p_poly->r_p_Vertex(0)->v_Pos() - p_poly->r_p_Vertex(i)->v_Pos();
				if(v_delta_pos1.f_Dot(v_delta_pos2) /
				   (v_delta_pos1.f_Length() * v_delta_pos2.f_Length()) < 1.0f - f_epsilon) {
					n = i;
					break;
				}
			}
			// try to find two not-colinear vectors

			Vector2f v_st[3];
			for(int i = 0; i < 3; i ++) {
				int n_index = (i < 2)? i : n;
				v_st[i].x = p_poly->r_p_Vertex(n_index)->v_texcoord[n_tex_channel][n_s_coord];
				v_st[i].y = p_poly->r_p_Vertex(n_index)->v_texcoord[n_tex_channel][n_t_coord];
				// get s, t
			}
			Vector2f v_delta_st1 = v_st[0] - v_st[1];
			Vector2f v_delta_st2 = v_st[0] - v_st[2];
			// the same vectors, but in texture space

			float f_r = 1 / (v_delta_st1.x * v_delta_st2.y - v_delta_st2.x * v_delta_st1.y);
			v_s_dir = Vector3f(
				(v_delta_st2.y * v_delta_pos1.x - v_delta_st1.y * v_delta_pos2.x) * f_r,
				(v_delta_st2.y * v_delta_pos1.y - v_delta_st1.y * v_delta_pos2.y) * f_r,
				(v_delta_st2.y * v_delta_pos1.z - v_delta_st1.y * v_delta_pos2.z) * f_r);
			v_t_dir = Vector3f(
				(v_delta_st1.x * v_delta_pos2.x - v_delta_st2.x * v_delta_pos1.x) * f_r,
				(v_delta_st1.x * v_delta_pos2.y - v_delta_st2.x * v_delta_pos1.y) * f_r,
				(v_delta_st1.x * v_delta_pos2.z - v_delta_st2.x * v_delta_pos1.z) * f_r);
			// calculate s & t directions over this polygon (actualy it works
			// for triangles and uniformly textured polygons only)
		}

		for(int j = 0; j < p_poly->n_Vertex_Num(); j ++) {
			TVertex *p_vertex = p_poly->r_p_Vertex(j);
			float f_texcoord[4] = {0, 0, 0, 1};
			Vector3f v_transformed_pos = m_world_transform.v_Transform_Pos(p_vertex->v_Pos());

			switch(n_method) {
			case coord_XYZ_to_STR:
				for(int n = 0; n < 3; n ++)
					f_texcoord[n] = v_transformed_pos[n];
				break;
			case coord_Planar:
				for(int n = 0; n < 2; n ++)
					f_texcoord[n] = v_transformed_pos[n];
				f_texcoord[2] = 0;
				break;
			case coord_Box:
				{
					Vector3f v_normal =
						m_world_transform.v_Transform_Dir(p_poly->t_Normal().v_normal);
					for(int n = 0; n < 3; n ++) {
						if(((float)fabs(v_normal[n]) > (float)fabs(v_normal[(n + 1) % 3]) &&
						   (float)fabs(v_normal[n]) > (float)fabs(v_normal[(n + 2) % 3])) || n == 2) {
							// find maximal coordinate

							f_texcoord[0] = v_transformed_pos[(n + 1) % 3];
							f_texcoord[1] = v_transformed_pos[(n + 2) % 3];
							// mapping plane lies on the other two
							break;
						}
					}
				}
				break;
			case coord_Cylinder:
			case coord_Sphere:
				{
					bool b_changing_period = false, b_have_first_quadrant = false;
					for(int k = 0; k < p_poly->n_Vertex_Num(); k ++) {
						Vector3f v_pos = 
							m_world_transform.v_Transform_Pos(p_poly->r_p_Vertex(k)->v_Pos());
						if(v_pos.z >= 0 && v_pos.x >= 0)
							b_have_first_quadrant = true;
						else if(v_pos.x >= 0)
							b_changing_period = true;
					}
					b_changing_period &= b_have_first_quadrant;

					float f_radius = v_transformed_pos.f_Length();
					f_texcoord[0] = (atanf(v_transformed_pos[2] /
						v_transformed_pos[0]) +
						(v_transformed_pos[0] < 0) * f_pi) / (2 * f_pi); // s ~ theta
					if(f_texcoord[0] < 0)
						f_texcoord[0] += 1;
					if(b_changing_period && v_transformed_pos.x >= 0 && v_transformed_pos.z >= 0)
						f_texcoord[0] += 1;
					f_texcoord[1] = (n_method == coord_Cylinder)? v_transformed_pos[1] : // t ~ y
						acosf(v_transformed_pos[1] / f_radius) / f_pi; // r ~ azimuth
					f_texcoord[2] = 0;
				}
				// convert to spherical coordinates, trying to always generate positive coords
				break;
			default:
				if(n_method >= coord_Tangent_of_1 && n_method <
				   coord_Tangent_of_1 + max_Vertex_Texcoords) {
					for(int n = 0; n < 3; n ++)
						p_vertex->v_texcoord[n_tex_channel][n_r_coord + n] += v_s_dir[n];
					if(n_q_coord != -1)
						p_helper_data[p_vertex->n_flag] += v_t_dir;
				} else if(n_method >= coord_BiTangent_of_1 && n_method <
				   coord_BiTangent_of_1 + max_Vertex_Texcoords) {
					for(int n = 0; n < 3; n ++)
						p_vertex->v_texcoord[n_tex_channel][n_r_coord + n] += v_t_dir[n];
					if(n_q_coord != -1)
						p_helper_data[p_vertex->n_flag] += v_s_dir;
				}
				break;
			}

			f_texcoord[0] *= v_scale.x;
			f_texcoord[1] *= v_scale.y;
			f_texcoord[2] *= v_scale.z;
			// w should stay either 0 or +-1

			if(n_method < coord_Tangent_of_1 ||
			   n_method > coord_BiTangent_of_1 + max_Vertex_Texcoords) {
				// different way of generating texcoords for those methods
				if((n_s_coord != -1 && p_vertex->v_texcoord[n_tex_channel][n_s_coord] != f_texcoord[0]) ||
				   (n_t_coord != -1 && p_vertex->v_texcoord[n_tex_channel][n_t_coord] != f_texcoord[1]) ||
				   (n_r_coord != -1 && p_vertex->v_texcoord[n_tex_channel][n_r_coord] != f_texcoord[2]) ||
				   (n_q_coord != -1 && p_vertex->v_texcoord[n_tex_channel][n_q_coord] != f_texcoord[3])) {
					// if newly generated texture coordinates are different

					if(p_vertex->n_flag == false)
						p_vertex->n_flag = true; // it was unused
					else {
						m_b_poly_hash_dirty_flag = true;
						if(!(p_poly->r_p_Vertex(j) = p_vertex = p_AddVertex(*p_vertex)))
							return false;
					}
					// add duplicate when necessary
				} else
					p_vertex->n_flag = true;

				if(n_s_coord != -1)
					p_vertex->v_texcoord[n_tex_channel][n_s_coord] = f_texcoord[0];
				if(n_t_coord != -1)
					p_vertex->v_texcoord[n_tex_channel][n_t_coord] = f_texcoord[1];
				if(n_r_coord != -1)
					p_vertex->v_texcoord[n_tex_channel][n_r_coord] = f_texcoord[2];
				if(n_q_coord != -1)
					p_vertex->v_texcoord[n_tex_channel][n_q_coord] = f_texcoord[3];
				// assign new texture coordinates
			}
		}
	}

	if(n_method >= coord_Tangent_of_1 &&
	   n_method < coord_BiTangent_of_1 + max_Vertex_Texcoords) {
		for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin();
		   p_vertex < m_vertex_list.end(); p_vertex ++) {
			if((*p_vertex)->n_flag == -1)
				continue;

			Vector3f v_t_dir, v_s_dir;

			//if(n_method > coord_Tangent_of_1 + max_Vertex_Texcoords) {
				v_s_dir = Vector3f(
					(*p_vertex)->v_texcoord[n_tex_channel][n_r_coord],
					(*p_vertex)->v_texcoord[n_tex_channel][n_r_coord + 1],
					(*p_vertex)->v_texcoord[n_tex_channel][n_r_coord + 2]);
				if(n_q_coord != -1)
					v_t_dir = p_helper_data[(*p_vertex)->n_flag];
			/*} else {
				v_t_dir = Vector3f(
					(*p_vertex)->v_texcoord[n_tex_channel][n_r_coord],
					(*p_vertex)->v_texcoord[n_tex_channel][n_r_coord + 1],
					(*p_vertex)->v_texcoord[n_tex_channel][n_r_coord + 2]);
				if(n_q_coord != -1)
					v_s_dir = p_helper_data[(*p_vertex)->n_flag];
			}*/

			Vector3f v_tangent = v_t_dir - (*p_vertex)->v_normal *
				(*p_vertex)->v_normal.f_Dot(v_t_dir);
			v_tangent.Normalize();

			if(n_r_coord != -1) {
				//if(n_method > coord_Tangent_of_1 + max_Vertex_Texcoords)
				//	v_tangent = (*p_vertex)->v_normal.v_Cross(v_tangent);
				// we wanted bi-tangent which is prependicular to both normal and tangent

				for(int j = 0; j < 3; j ++)
					(*p_vertex)->v_texcoord[n_tex_channel][n_r_coord + j] = v_tangent[j];
			}
			// store whatever you wanted to calculate
			
			if(n_q_coord != -1) {
				(*p_vertex)->v_texcoord[n_tex_channel][n_q_coord] = 1 - 2 *
					((*p_vertex)->v_normal.v_Cross(v_t_dir).f_Dot(v_s_dir) < 0);
			}
			// determine texture polarity
		}

		if(n_q_coord != -1) {
			_ASSERTE(p_helper_data);
			delete[] p_helper_data;
		}
	}

	return true;
}

bool CGeometry::Find_MinPolygonAlignedBoundingBoxBasis(const CRefPolygon *p_poly,
	Vector3f &r_v_x, Vector3f &r_v_y, Vector3f &r_v_z)
{
	float f_max_area = -1.0f;

	Vector3f v_prev_pos, v_pos = p_poly->t_Vertex(p_poly->n_Vertex_Num() - 1).v_Pos();
	for(int i = 0; i < p_poly->n_Vertex_Num(); i ++, v_prev_pos = v_pos) {
		v_pos = p_poly->t_Vertex(i).v_Pos();

		Vector3f v_x = v_pos - v_prev_pos;
		float f_edge_len;
		if((f_edge_len = (v_pos - v_prev_pos).f_Length()) < f_epsilon)
			continue;
		// can't calc for very short edges

		v_x *= 1 / f_edge_len;
		Vector3f v_y = p_poly->t_Normal().v_normal;
		Vector3f v_z = r_v_x.v_Cross(r_v_y);
		// calc vector basis

		Vector3f v_min, v_max;
		for(int j = 0; j < p_poly->n_Vertex_Num(); j ++) {
			Vector3f v_pos = p_poly->t_Vertex(j).v_Pos();
			v_pos = Vector3f(v_pos.f_Dot(v_x), v_pos.f_Dot(v_y), v_pos.f_Dot(v_z));
			// calc a new position in our edge-space

			if(!j) {
				v_min = v_pos;
				v_max = v_pos;
			} else {
				for(int n = 0; n < 3; n ++) {
					if(v_min[n] > v_pos[n] || !j)
						v_min[n] = v_pos[n];
					if(v_max[n] > v_pos[n] || !j)
						v_max[n] = v_pos[n];
				}
			}
		}
		// find min/max in our bounding-box space (.z should be 0
		// if polygon is planar - as it should be)

		if(f_max_area > (v_min - v_max).f_Length2() || f_max_area < 0) {
			f_max_area = (v_min - v_max).f_Length2();
			// bounding box projected to x-y area

			r_v_x = v_x;
			r_v_y = v_y;
			r_v_z = v_z;
			// copy vectors
		}
	}

	return f_max_area > -.5f;
}

/*
 *	bool CGeometry::Alloc_Surfaces(ESurfMethod n_method, const int *p_polygon_indices,
 *		int n_polygon_num, float f_max_deviation, Vector3f v_scale, int n_max_texture_size,
 *		int n_border_width, int n_surface_channel, int n_tex_channel, int n_s_coord,
 *		int n_t_coord)
 *		- calc per-surface texture coordinates using specified method (return false for unknown)
 *		  for polygons specified by p_polygon_indices and n_polygon_num (if p_polygon_indices is
 *		  0, n_polygon_num first polygons will be processed)
 *		- v_scale is no. of texels / worldspace unit ratio, n_max_texture_size is maximal texture
 *		  size (polygons too large to fit will be subdivided), n_border_width is space in texels
 *		  which should be left unused between separate surfaces (surfaces using same texture,
 *		  not smoothing surfaces)
 *		- n_tex_channel, n_s_coord and n_t_coord specify where to put a new texture coordinates
 *		- n_surface_channel specify which per-surface texture id to set
 *		- return false if there's not enough memory or on unknown method
 */
bool CGeometry::Alloc_Surfaces(ESurfMethod n_method, const int *p_polygon_indices,
	int n_polygon_num, float f_max_deviation, Vector3f v_scale, int n_max_texture_size,
	int n_border_width, int n_surface_channel, int n_tex_channel, int n_s_coord, int n_t_coord)
{
	if((n_method != surf_Planar && n_method != surf_Unwrap) ||
	   n_surface_channel < 0 || n_surface_channel >= max_Surface_Textures ||
	   n_tex_channel < 0 || n_tex_channel >= max_Vertex_Texcoords ||
	   n_s_coord < 0 || n_s_coord > 3 || n_t_coord < 0 || n_t_coord > 3)
		return false;
	// see if the params are right

	{int n_index = 0;
	for(std::vector<TPolygon>::iterator p_polygon = (p_polygon_indices)?
	   m_poly_list.begin() + (n_index = (*p_polygon_indices)) : m_poly_list.begin();
	   n_polygon_num > 0; n_polygon_num --, (p_polygon_indices)? p_polygon = m_poly_list.begin() +
	   (n_index = (*(++ p_polygon_indices))) : (n_index ++, p_polygon ++)) { // todo - see what g++ thinks of such a construct
		(*p_polygon).n_surface_map[n_surface_channel] = -1;
		//if(!
	}}
	// mark all polys as un-allocated

	return true;
}

/*
 *	bool CGeometry::MakeTris()
 *		- break polygons into tris
 *		- return false if there's a problem with low memory
 */
bool CGeometry::MakeTris()
{
	m_b_poly_neighbours_dirty_flag = true;
	m_b_poly_hash_dirty_flag = true;

	for(std::vector<TPolygon>::iterator p_poly = m_poly_list.begin(),
	   p_end = m_poly_list.end(); p_poly < p_end; p_poly ++) {
		CRefPolygon *p_polygon = (*p_poly).p_poly;
		if(p_polygon->n_Vertex_Num() <= 3) {
			int n_good_vertex = 0;
			for(int j = 0; j < p_polygon->n_Vertex_Num(); j ++) {
				for(int s = -1, n = 0; s <= 1; s += 2) {
					int n_i0 = (j - 2 * s + p_polygon->n_Vertex_Num()) %
						p_polygon->n_Vertex_Num();
					int n_i1 = (j - s + p_polygon->n_Vertex_Num()) %
						p_polygon->n_Vertex_Num();

					Vector3f v1 = p_polygon->t_Vertex(j).v_Pos() -
						p_polygon->t_Vertex(n_i0).v_Pos();
					v1.Normalize();
					Vector3f v2 = p_polygon->t_Vertex(j).v_Pos() -
						p_polygon->t_Vertex(n_i1).v_Pos();
					v2.Normalize();

					if(v1.f_Dot(v2) < 1.0f - f_epsilon) {
						if(n ++) {
							n_good_vertex = j;
							j = p_polygon->n_Vertex_Num();
							break;
						}
					}
				}
			}
			// select good vertex to begin at (vertex that doesn't
			// lie before or after two colinear edges)

			for(int j = 3; j < p_polygon->n_Vertex_Num(); j ++) {
				TPolygon t_poly;
				t_poly = (*p_poly);
				t_poly.n_neighbour_edge_num = 0;
				t_poly.p_neighbour_index_list = 0;

				if(!(t_poly.p_poly = new CRefPolygon))
					return false;

				TRefVertex t_vertex[3];
				t_vertex[0].p_geom = this;
				t_vertex[0].p_vertex = p_polygon->r_p_Vertex(n_good_vertex);
				t_vertex[1].p_geom = this;
				t_vertex[1].p_vertex = p_polygon->r_p_Vertex((n_good_vertex +
					j - 1) % p_polygon->n_Vertex_Num());
				t_vertex[2].p_geom = this;
				t_vertex[2].p_vertex = p_polygon->r_p_Vertex((n_good_vertex + j) %
					p_polygon->n_Vertex_Num());

				if(!t_poly.p_poly->Add_Vertex(t_vertex, 3)) {
					delete t_poly.p_poly;
					return false;
				}

				t_poly.p_poly->r_t_Normal() = p_polygon->t_Normal();

				m_b_poly_hash_dirty_flag = true;
				m_b_vertex_normal_smooth_group_dirty_flag = true;
				m_b_vertex_normal_smooth_all_dirty_flag = true;
				m_b_poly_neighbours_dirty_flag = true;

				m_poly_list.push_back(t_poly);
				if(!m_poly_list.size()) {
					delete t_poly.p_poly;
					return false;
				}
			}
			// create tris ...

			// first tri should be n_good_vertex, n_good_vertex + 1, n_good_vertex + 2

			if(n_good_vertex)
				p_polygon->Delete_Vertices(0, n_good_vertex);
			// delete vertices before n_good_vertex

			if(n_good_vertex + 3 < p_polygon->n_Vertex_Num()) {
				p_polygon->Delete_Vertices(n_good_vertex + 3,
					p_polygon->n_Vertex_Num() - n_good_vertex - 3);
			}
			// delete vertices after n_good_vertex + 3
		}
	}

	return true;
}

/*
 *	bool CGeometry::Tesselate(float f_normal_tension)
 *		- break polygons into tris different way and offset
 *		  vertices along their normals times f_normal_tension
 *		- return false if there's a problem with low memory
 */
bool CGeometry::Tesselate(float f_normal_tension)
{
	if(!UpdateVertexNormals_SmoothAll())
		return false;
	// need smooth normals ...

	m_b_poly_neighbours_dirty_flag = true;
	m_b_bounding_box_dirty_flag = true;
	m_b_poly_hash_dirty_flag = true;
	m_b_vertex_hash_dirty_flag = true;
	m_b_vertex_normal_smooth_all_dirty_flag = true;
	m_b_vertex_normal_smooth_group_dirty_flag = true;

	for(std::vector<TPolygon>::iterator p_poly = m_poly_list.begin(),
	   p_end = m_poly_list.end(); p_poly < p_end; p_poly ++) {
		CRefPolygon *p_polygon = (*p_poly).p_poly;
		TVertex t_center;
		memset(&t_center, 0, sizeof(TVertex)); // hack! .. ale (float)0 je 0x00000000
		for(int j = 0; j < p_polygon->n_Vertex_Num(); j ++)
			t_center = t_center.Lerp(1.0f / (j + 1), p_polygon->t_Vertex(j));
		t_center.v_normal.Normalize();
		if((float)fabs(t_center.v_normal.f_Length() - 1) > f_epsilon)
			t_center.v_normal = p_polygon->t_Normal().v_normal;
		// calc center (including texcoords & normal)

		t_center.v_Pos() += t_center.v_normal * f_normal_tension;
		// displace vertex (todo - maybe displace all vertices - see what is the result first)

		TRefVertex t_ref_vertex = t_AddRefVertex(t_center);
		if(!t_ref_vertex.p_vertex)
			return false;

		for(int j = 1; j < p_polygon->n_Vertex_Num(); j ++) {
			TPolygon t_poly;
			t_poly = (*p_poly);
			t_poly.n_neighbour_edge_num = 0;
			t_poly.p_neighbour_index_list = 0;

			if(!(t_poly.p_poly = new CRefPolygon))
				return false;

			TRefVertex t_vertex[3];
			t_vertex[0].p_geom = this;
			t_vertex[0].p_vertex = p_polygon->r_p_Vertex(j);
			t_vertex[1].p_geom = this;
			t_vertex[1].p_vertex = p_polygon->r_p_Vertex((j + 1) %
			p_polygon->n_Vertex_Num());
			t_vertex[2] = t_ref_vertex;

			if(!t_poly.p_poly->Add_Vertex(t_vertex, 3)) {
				delete t_poly.p_poly;
				return false;
			}

			t_poly.p_poly->r_t_Normal() = p_polygon->t_Normal();

			m_b_poly_hash_dirty_flag = true;
			m_b_vertex_normal_smooth_group_dirty_flag = true;
			m_b_vertex_normal_smooth_all_dirty_flag = true;
			m_b_poly_neighbours_dirty_flag = true;

			m_poly_list.push_back(t_poly);
			if(!m_poly_list.size()) {
				delete t_poly.p_poly;
				return false;
			}
		}
		// create tris ...

		p_polygon->Delete_Vertices(3, p_polygon->n_Vertex_Num() - 3);
		// delete vertices of first poly to make it triangle

		p_polygon->r_p_Vertex(2) = t_ref_vertex.p_vertex;
		// third vertex is going to be the centre one
	}
	// subdivide polygons

	return true;
}

/*
 *	bool CGeometry::CalcVertexNormals(bool b_ignore_groups)
 *		- calculate vertex normals
 *		- if b_ignore_groups is false (default), vertex normals
 *		  are computed so only individual surfaces are smooth
 *		  otherwise the whole mesh is smooth
 *		- can return false if there's not enough memory for look-up's
 */
bool CGeometry::CalcVertexNormals(bool b_ignore_groups)
{
	if(!UpdatePolygonHash())
		return false;
	// update polygon hash table if necessary

	m_b_vertex_normal_smooth_all_dirty_flag = true;
	m_b_vertex_normal_smooth_group_dirty_flag = true;
	if(!b_ignore_groups) {
		for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
		   p_vertex < m_vertex_list.end(); p_vertex ++) {
			int n_surface = -1;
			std::vector<int> *p_hash = m_p_poly_hash_table[n_HashFunction(*p_vertex,
				m_n_poly_hash_table_size)];
			for(std::vector<int>::iterator p_poly_index = p_hash->begin();
			   p_poly_index < p_hash->end(); p_poly_index ++) {
				CRefPolygon *p_poly = m_poly_list[*p_poly_index].p_poly;
				TVertex *p_new_vertex = 0;
				int n_new_hash;
				for(int k = 0; k < p_poly->n_Vertex_Num(); k ++) {
					if(p_poly->r_p_Vertex(k) == *p_vertex) {
						if(n_surface == -1)
							n_surface = m_poly_list[*p_poly_index].n_surface;
						else if(n_surface != m_poly_list[*p_poly_index].n_surface) {
							// we've got a single vertex, used by multiple surfaces,
							// we have to generate a new vertex and replace it
							// in the other polys

							if(!p_new_vertex) {
								if(!(p_new_vertex = p_AddVertex(**p_vertex)))
									return false;
								p_new_vertex->n_flag = (*p_vertex)->n_flag;
								// create new vertex

								n_new_hash = n_HashFunction(p_new_vertex,
									m_n_poly_hash_table_size);
								// calc polygon hash table index for the new vertex
							}
							// create a new vertex (we're going to get to it
							// to separate more possible aliases later on)

							int n_poly_index = *p_poly_index;
							p_hash->erase(p_poly_index --);
							// this poly will reference that vertex no longer // todo - check if this is ok

							m_p_poly_hash_table[n_new_hash]->push_back(n_poly_index);
							if(!m_p_poly_hash_table[n_new_hash]->size()) {
								m_b_poly_hash_dirty_flag = true;
								// poly hash is damaged
								return false;
							}
							// update polygon hash

							p_poly->r_p_Vertex(k) = p_new_vertex;
							// replace vertex
						}
						break;
					}
				}
			}
		}
	}
	// separate vertex surfaces, keep poly hash up-to-date

	std::vector<std::vector<TVertex*>*> *p_merge_table;
	if(!(p_merge_table = p_VertexMergeTable(true, f_epsilon)))
		return false;
	// create merge table ...

	for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = -1;
	// clear vertex flags

	{int n_index = 0;
	for(std::vector<std::vector<TVertex*>*>::iterator p_list = p_merge_table->begin();
	   p_list < p_merge_table->end(); p_list ++, n_index ++) {
		for(std::vector<TVertex*>::iterator p_vertex = (*p_list)->begin(); p_vertex < (*p_list)->end();)
			(*(p_vertex ++))->n_flag = n_index;
		// set vertex flags ...
	}}

	for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end(); p_vertex ++) {
		int n_surface = -1;
		if(!b_ignore_groups) {
			std::vector<int> *p_hash = m_p_poly_hash_table[n_HashFunction(*p_vertex,
				m_n_poly_hash_table_size)];
			for(std::vector<int>::const_iterator p_poly_index = p_hash->begin();
			   p_poly_index < p_hash->end(); p_poly_index ++) {
				CRefPolygon *p_poly = m_poly_list[*p_poly_index].p_poly;
				for(int k = 0; k < p_poly->n_Vertex_Num(); k ++) {
					if(p_poly->r_p_Vertex(k) == *p_vertex) {
						n_surface = m_poly_list[*p_poly_index].n_surface;
						break;
					}
				}
				if(n_surface != -1)
					break;
			}
			if(n_surface == -1)
				continue;
		}
		// determine vertex surface if necessary
		// (single vertex shoudln't definitely be used by more different surfaces)

		(*p_vertex)->v_normal = Vector3f(0, 0, 0);
		// don't affect normals of unused vertices

		for(std::vector<TVertex*>::iterator p_same_vertex =
		   (*p_merge_table)[(*p_vertex)->n_flag]->begin();
		   p_same_vertex < (*p_merge_table)[(*p_vertex)->n_flag]->end(); p_same_vertex ++) {
			std::vector<int> *p_hash = m_p_poly_hash_table[n_HashFunction(*p_same_vertex,
				m_n_poly_hash_table_size)];
			for(std::vector<int>::const_iterator p_poly_index = p_hash->begin();
			   p_poly_index < p_hash->end(); p_poly_index ++) {
				if(!b_ignore_groups &&
				   m_poly_list[*p_poly_index].n_surface != n_surface)
					continue;

				CRefPolygon *p_poly = m_poly_list[*p_poly_index].p_poly;
				for(int l = 0; l < p_poly->n_Vertex_Num(); l ++) {
					if(p_poly->r_p_Vertex(l) == *p_same_vertex) {
						Vector3f v_edge_a = p_poly->r_p_Vertex((l + 1) % p_poly->n_Vertex_Num())->v_Pos() -
							p_poly->r_p_Vertex(l)->v_Pos();
						Vector3f v_edge_b = p_poly->r_p_Vertex((l - 1 + p_poly->n_Vertex_Num()) %
							p_poly->n_Vertex_Num())->v_Pos() - p_poly->r_p_Vertex(l)->v_Pos();
						// two edges, sharing vertex

						float f_angle = acosf(v_edge_a.f_Dot(v_edge_b) /
							(float)sqrt(v_edge_a.f_Length2() * v_edge_b.f_Length2()));
						// angle between them (radians)

						(*p_vertex)->v_normal += p_poly->t_Normal().v_normal * f_angle;
						// weighted normal average (removes some glitches for triangle-based geometry)
						break;
					}
				}
			}
		}
		(*p_vertex)->v_normal.Normalize();
		if((float)fabs((*p_vertex)->v_normal.f_Length() - 1) > f_epsilon)
			(*p_vertex)->v_normal = Vector3f(0, 0, 0);
		// calc vertex normal
	}
	// calc vertex normals averaging referencing polygon normals

	for(std::vector<std::vector<TVertex*>*>::iterator p_list = p_merge_table->begin();
	   p_list < p_merge_table->end();)
		delete *(p_list ++);
	delete p_merge_table;
	// free merge table

	if(b_ignore_groups)
		m_b_vertex_normal_smooth_all_dirty_flag = false;
	else
		m_b_vertex_normal_smooth_group_dirty_flag = false;

	return true;
}

/*
 *	bool CGeometry::b_IsEnclosed_Euler(bool &r_b_enclosed)
 *		- return true (by reference) when mesh is enclosed
 *		- this does use Euler rule so it work for convex objects only
 */
bool CGeometry::b_IsEnclosed_Euler(bool &r_b_enclosed)
{
	if(!UpdatePolygonHash() || !UpdatePolygonNeighbours())
		return false;
	// update polygon hash table & neighbour lists if necessary

	std::vector<std::vector<TVertex*>*> *p_merge_table;
	if(!(p_merge_table = p_VertexMergeTable(true, f_epsilon)))
		return false;
	// create merge table ...

	for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = -1;
	// clear vertex flags

	for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end(); p_vertex ++) {
		std::vector<int> *p_hash = m_p_poly_hash_table[n_HashFunction(*p_vertex,
			m_n_poly_hash_table_size)];
		for(std::vector<int>::const_iterator p_poly_index = p_hash->begin();
		   p_poly_index < p_hash->end(); p_poly_index ++) {
			CRefPolygon *p_poly = m_poly_list[*p_poly_index].p_poly;
			for(int k = 0; k < p_poly->n_Vertex_Num(); k ++) {
				if(p_poly->r_p_Vertex(k) == *p_vertex) {
					(*p_vertex)->n_flag = m_poly_list[*p_poly_index].n_surface;
					break;
				}
			}
			if((*p_vertex)->n_flag != -1)
				break;
		}
		if((*p_vertex)->n_flag == -1)
			continue;
		// determine vertex surface
	}
	// mark used vertices ...

	int n_vertex_num = 0;
	int n_edge_num = 0;
	int n_poly_num = m_poly_list.size();

	for(std::vector<std::vector<TVertex*>*>::const_iterator p_list = p_merge_table->begin();
	   p_list < p_merge_table->end(); p_list ++) {
		for(std::vector<TVertex*>::const_iterator p_vertex = (*p_list)->begin();
		   p_vertex < (*p_list)->end();) {
			if((*(p_vertex ++))->n_flag != -1) {
				n_vertex_num ++;
				break;
			}
		}
	}
	// sum up used vertices

	{int n_index = 0;
	for(std::vector<TPolygon>::const_iterator p_poly_wrapper = m_poly_list.begin();
	   p_poly_wrapper < m_poly_list.end(); p_poly_wrapper ++, n_index ++) {
		for(int j = 0; j < (*p_poly_wrapper).n_neighbour_edge_num; j ++) {
			bool b_marked_already = false;
			for(std::vector<int>::const_iterator p_poly_index =
			   (*p_poly_wrapper).p_neighbour_index_list[j]->begin(); p_poly_index <
			   (*p_poly_wrapper).p_neighbour_index_list[j]->end();) {
				if(*(p_poly_index ++) < n_index) {
					b_marked_already = true;
					break;
				}
			}
			if(b_marked_already)
				continue;
			// see if this edge was marked already

			if(((*p_poly_wrapper).p_poly->t_Vertex(j).v_pos -
			   (*p_poly_wrapper).p_poly->t_Vertex((j + 1) %
			   (*p_poly_wrapper).p_poly->n_Vertex_Num()).v_pos).f_Length() >= f_epsilon)
				n_edge_num ++;
			// don't count degenerate edges as their vertices are merged into single position
		}
	}}
	// sum up edges

	for(int i = 0; i < p_merge_table->size(); i ++)
		delete (*p_merge_table)[i];
	delete p_merge_table;
	// free merge table

	r_b_enclosed = n_vertex_num + n_poly_num == n_edge_num + 2;
	// Euler rule

	return true;
}

/*
 *	bool CGeometry::b_IsEnclosed_Adjacency(bool &r_b_enclosed)
 *		- return true (by reference) when mesh is enclosed
 *		- this does use edge adjacency so it should work for
 *		  any object with simple edges (1 edge is shared by 1 or 2 polys)
 *		- it considers mesh enclosed even if it contain multiple
 *		  enclosed objects
 */
bool CGeometry::b_IsEnclosed_Adjacency(bool &r_b_enclosed)
{
	if(!UpdatePolygonNeighbours())
		return false;
	// update polygon neighbour lists if necessary

	for(std::vector<TPolygon>::const_iterator p_poly_wrapper = m_poly_list.begin();
	   p_poly_wrapper < m_poly_list.end(); p_poly_wrapper ++) {
		for(int j = 0; j < (*p_poly_wrapper).n_neighbour_edge_num; j ++) {
			if((*p_poly_wrapper).p_neighbour_index_list[j]->size() == 0) {
				r_b_enclosed = false;
				return true;
			}
		}
		// see if there is edge with no neighbours -> it's
		// boundary edge, i.e. object is not enclosed
	}

	r_b_enclosed = true;

	return true;
}

/*
 *	bool Cut(const Plane3f &r_t_plane, EPlanePos n_desired_half, float f_epsilon_ex, CGeometry *p_other_half = 0)
 *		- cut mesh in half, defined by t_plane
 *		- <this> mesh is going to contain all geometry, lying in position n_desired_half
 *		- there is possibility to store all other (other = being deleted from <this>) geometry
 *		  in p_other_half (it's original contents aren't erased, we just add new geometry)
 *		- it's recommended to call OptimizeVertices() afterwards, since polygon splitting
 *		  may generate vertices that won't be used in <this> mesh. (p_other_half is going
 *		  to get used vertices only)
 *		- can be called with p_other_half set to <this>, but it generates double vertices
 *		  on edges cut by plane (can be eliminated by optimizing as well)
 */
bool CGeometry::Cut(const Plane3f &r_t_plane, EPlanePos n_desired_half,
	float f_epsilon_ex, CGeometry *p_other_half)
{
	if(p_other_half && p_other_half != this) {
		UpdateBoundingBox();
		// update bounding box so we know dimensions of our geometry

		p_other_half->UpdateBoundingBox();
		for(int i = 0; i < 3; i ++) {
			if(p_other_half->m_v_bounding_box[0][i] > m_v_bounding_box[0][i])
				p_other_half->m_v_bounding_box[0][i] = m_v_bounding_box[0][i];
			if(p_other_half->m_v_bounding_box[1][i] < m_v_bounding_box[1][i])
				p_other_half->m_v_bounding_box[1][i] = m_v_bounding_box[1][i];
		}
		// extend bounding box so all possible inserted vertices fit ...

		p_other_half->m_b_bounding_box_dirty_flag = false;
		// set bounding box flag to false, because we don't want it to be recomputed until we're done

		for(std::vector<TVertex*>::iterator p_vertex = p_other_half->m_vertex_list.begin();
		   p_vertex < p_other_half->m_vertex_list.end();)
			(*(p_vertex ++))->n_flag = -1;
		// set flags to vertices that already were part of geometry
		// so we don't weld new vertices to them

		{int n_index = 0;
		for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin();
		   p_vertex < m_vertex_list.end();)
			(*(p_vertex ++))->n_flag = n_index ++;
		}
		// set flags to our vertices so we don't weld vertices we're adding

		if(!p_other_half->UpdateVertexHash()) {
			p_other_half->m_b_bounding_box_dirty_flag = true;
			return false;
		}
		// update vertex hash
	}
	// in case we're going to add vertices to the other geometry, we'll need vertex hash table
	// to speed-up inserting vertices, used by new geometry. to avoid re-computing vertex hash
	// table after each inserted vertex, we maximize bounding box so we can easily update vertex
	// hash after adding each vertex without need to rebuild it

	for(std::vector<TPolygon>::iterator p_poly = m_poly_list.begin(),
	   p_end = m_poly_list.end(); p_poly < p_end; p_poly ++) {
		if((*p_poly).p_poly->n_Plane_Pos(r_t_plane, f_epsilon_ex) == n_desired_half)
			continue; // in case nothing is about to happen ...
		else {
			if(p_other_half && p_other_half != this) {
				p_other_half->m_b_poly_hash_dirty_flag = true;
				p_other_half->m_b_poly_neighbours_dirty_flag = true;
				p_other_half->m_b_vertex_normal_smooth_all_dirty_flag = true;
				p_other_half->m_b_vertex_normal_smooth_group_dirty_flag = true;
			}
			m_b_poly_hash_dirty_flag = true;
			m_b_poly_neighbours_dirty_flag = true;
			m_b_vertex_normal_smooth_all_dirty_flag = true;
			m_b_vertex_normal_smooth_group_dirty_flag = true;
			// set dirty flags ...

			TPolygon t_poly = (*p_poly);
			CRefPolygon *p_polygon = t_poly.p_poly;

			if(p_other_half) {
				t_poly.n_neighbour_edge_num = 0;
				t_poly.p_neighbour_index_list = 0;
				if(!(t_poly.p_poly = new CRefPolygon) ||
				   !p_polygon->Split(r_t_plane,
				   *t_poly.p_poly, n_desired_half, f_epsilon_ex)) {
					if(t_poly.p_poly)
						delete t_poly.p_poly;
					return false;
				}
			} else if(!p_polygon->Cut(r_t_plane, n_desired_half, f_epsilon_ex))
				return false;
			// create new polygon

			if(!p_polygon->n_Vertex_Num()) {
				delete p_polygon;
				if((*p_poly).p_neighbour_index_list) {
					for(std::vector<int> **p_list = (*p_poly).p_neighbour_index_list,
					   **p_end = (*p_poly).p_neighbour_index_list + (*p_poly).n_neighbour_edge_num;
					   p_list < p_end;)
						delete *p_list ++;
					delete[] (*p_poly).p_neighbour_index_list;
				}
				m_poly_list.erase(p_poly --);
				p_end --;
			}
			// in case source polygon was erased, free it's wrapper ...

			if(p_other_half && p_other_half != this) {
				_ASSERTE(!p_other_half->m_b_vertex_hash_dirty_flag);
				_ASSERTE(p_other_half->m_p_vertex_hash_table);
				// should be up-to-date and available

				for(int j = 0; j < t_poly.p_poly->n_Vertex_Num(); j ++) {
					std::vector<TVertex*> *p_hash_table =
						p_other_half->m_p_vertex_hash_table[p_other_half->n_HashFunction(
						t_poly.p_poly->t_Vertex(j).v_pos, p_other_half->m_n_vertex_hash_table_size,
						p_other_half->m_f_vertex_hash_table_cell_size)];
					TVertex *p_found = 0;
					for(std::vector<TVertex*>::const_iterator p_vertex = p_hash_table->begin();
					   p_vertex = p_hash_table->end(); p_vertex ++) {
						if((*p_vertex)->b_Compare(t_poly.p_poly->t_Vertex(j), f_epsilon_ex) &&
						   (*p_vertex)->n_flag == t_poly.p_poly->t_Vertex(j).n_flag) {
							p_found = (*p_vertex);
							break;
						}
					}
					if(!p_found &&
					   !(p_found = p_other_half->p_AddVertex(t_poly.p_poly->t_Vertex(j)))) {
						delete t_poly.p_poly;
						return false;
					}
					// either find vertex or add new

					t_poly.p_poly->r_p_Vertex(j) = p_found;
					t_poly.p_poly->SetVertexGeometryPointer(j, p_other_half);
					// replace vertex and set new geometry pointer
				}
				// in case we're going to transfer vertices, transfer vertices

				_ASSERTE(!p_other_half->m_b_bounding_box_dirty_flag);
				// vertices should fit in allright
			}
			if(p_other_half) {
				p_other_half->m_poly_list.push_back(t_poly);
				if(!p_other_half->m_poly_list.size()) {
					/*p_other_half->m_b_poly_hash_dirty_flag = true;
					p_other_half->m_b_poly_neighbours_dirty_flag = true;
					p_other_half->m_b_vertex_normal_smooth_all_dirty_flag = true;
					p_other_half->m_b_vertex_normal_smooth_group_dirty_flag = true;*/ // set before
					p_other_half->m_b_vertex_hash_dirty_flag = true;
					p_other_half->m_b_bounding_box_dirty_flag = true;
					return false;
				}
			}
		}
	}

	if(p_other_half && p_other_half != this) {
		p_other_half->m_b_vertex_hash_dirty_flag = true;
		p_other_half->m_b_bounding_box_dirty_flag = true;
		// we should let p_other_half re-compute vertex hash and bounding box
	}

	return true;
}

/*
 *	bool CGeometry::Slice(const Plane3f &r_t_plane,
 *		const int *p_poly_list, int n_poly_num, float f_epsilon_ex)
 *		- function for slicing some polygons
 *		- cares to create no t-junctions
 */
bool CGeometry::Slice(const Plane3f &r_t_plane,
	const int *p_polygon_indices, int n_polygon_num, float f_epsilon_ex)
{
	if(!UpdatePolygonNeighbours())
		return false;
	// need neighbour info to be able to care for t-junctions

	{int n_index = 0;
	for(std::vector<TPolygon>::const_iterator p_polygon = (p_polygon_indices)?
	   m_poly_list.begin() + (n_index = (*p_polygon_indices)) : m_poly_list.begin();
	   n_polygon_num > 0; n_polygon_num --, (p_polygon_indices)? p_polygon = m_poly_list.begin() +
	   (n_index = (*(++ p_polygon_indices))) : (n_index ++, p_polygon ++)) { // todo - see what g++ thinks of such a construct
		CRefPolygon t_tmp_poly[2];

		if(!(t_tmp_poly[0] = *(*p_polygon).p_poly) ||
		   !t_tmp_poly[0].Split(r_t_plane, t_tmp_poly[1], plane_Front, f_epsilon_ex))
			return false;
		// try to split the polygon

		if(t_tmp_poly[0].b_IsTiny() || t_tmp_poly[1].b_IsTiny())
			continue;
		// not interested in polygons that aren't split

		{int n_edge = 0;
		for(std::vector<int> **p_list = (*p_polygon).p_neighbour_index_list,
		   **p_end = (*p_polygon).p_neighbour_index_list + (*p_polygon).n_neighbour_edge_num;
		   p_list < p_end; p_list ++, n_edge ++) {
			// t_odo - make clear which vertices are associated with which neighbour index
			// mark all ussages of polygon neighbour lists with some keyword
			// marked "// using polyneighbours with vertex indices"
			// edge i is vertex i and the next one

			TVertex *p_edge_vertex[2] = {(*p_polygon).p_poly->r_p_Vertex(n_edge),
				(*p_polygon).p_poly->r_p_Vertex((n_edge + 1) % (*p_polygon).p_poly->n_Vertex_Num())};
			// get edge vertices

			int n_pos[2] = {r_t_plane.n_Vector_Pos(p_edge_vertex[0]->v_pos, f_epsilon_ex),
				r_t_plane.n_Vector_Pos(p_edge_vertex[1]->v_pos, f_epsilon_ex)};
			// see positions

			if(n_pos[0] == n_pos[1] || n_pos[0] == plane_Onplane || n_pos[1] == plane_Onplane)
				continue;
			// if both vertices lie on the same side of the plane, edge won't be cut

			for(std::vector<int>::const_iterator p_index = (*p_list)->begin();
			   p_index < (*p_list)->end(); p_index ++) {
				bool b_found = false;
				if(p_polygon_indices)
					b_found = *p_index >= 0 && *p_index < n_polygon_num;
				else {
					for(const int *p_index2 = p_polygon_indices, *p_end = p_polygon_indices +
					   n_polygon_num; p_index2 < p_end; p_index2 ++) {
						if(*p_index2 ++ == *p_index) {
							b_found = true;
							break;
						}
					}
				}
				if(b_found)
					continue;
				// this polygon will be split anyway so no t-junction originates

				CRefPolygon *p_polygon2 = m_poly_list[*p_index].p_poly;
				for(int i = 0, n_prev_i = p_polygon2->n_Vertex_Num() - 1;
				   i < m_poly_list[*p_index].n_neighbour_edge_num; i ++) {
					bool b_found = false;
					for(std::vector<int>::const_iterator p_index2 =
					   m_poly_list[*p_index].p_neighbour_index_list[n_prev_i]->begin(); p_index2 <
					   m_poly_list[*p_index].p_neighbour_index_list[n_prev_i]->end(); p_index2 ++) {
						if(*p_index2 == *p_index) {
							b_found = true;
							break;
						}
					}
					if(b_found) {
						// this is shared edge, it has to be split

						TVertex *p_edge_vertex[2] = {p_polygon2->r_p_Vertex(n_edge),
							p_polygon2->r_p_Vertex((n_edge + 1) % p_polygon2->n_Vertex_Num())};
						// get edge vertices

						float f_t;
						try {
							f_t = r_t_plane.f_Intersect_Ray_t(p_edge_vertex[0]->v_pos,
								p_edge_vertex[1]->v_pos - p_edge_vertex[0]->v_pos);
						} catch(Plane3f::CRayIsParallel_Exception) {
							_ASSERTE(0); // ray shouldn't be parallel with plane
							f_t = .5f;
						}
						// calculate intersection of edge with plane

						_ASSERTE(f_t > f_epsilon && f_t < 1.0f - f_epsilon);
						// intersection should be inside edge, not on vertices or even outside

						//m_b_vertex_hash_dirty_flag = true; // raised by t_AddRefVertex()
						m_b_poly_hash_dirty_flag = true; // polygon-vertex mapping changed
						m_b_poly_neighbours_dirty_flag = true; // new edges, new neighbours
						m_b_vertex_normal_smooth_all_dirty_flag = true;
						m_b_vertex_normal_smooth_group_dirty_flag = true;
						// raise dirty flags

						TRefVertex t_new_vertex;
						t_new_vertex = t_AddRefVertex(p_edge_vertex[0]->Lerp(f_t,
						   *p_edge_vertex[1]));
						if(!t_new_vertex.p_vertex || !p_polygon2->Add_Vertex(i, t_new_vertex)/*p_polygon2->Add_Vertex(t_new_vertex)*/)
							return false;
						// create a new vertex

						t_new_vertex.p_vertex->v_normal.Normalize();
						if((float)fabs(t_new_vertex.p_vertex->v_normal.f_Length() - 1) > f_epsilon)
							t_new_vertex.p_vertex->v_normal = p_polygon2->t_Normal().v_normal;
						// re-normalize normal

						/*for(int j = p_polygon2->n_Vertex_Num() - 1; j > i; j --) {
							TVertex *p_tmp_vertex = p_polygon2->r_p_Vertex(j);
							p_polygon2->r_p_Vertex(j) = p_polygon2->r_p_Vertex(j - 1);
							p_polygon2->r_p_Vertex(j - 1) = p_tmp_vertex;
						}*/
						// vertex was inserted to back of polygon, we need to shift it
						// between edge vertices

						break;
					}
				}
				// take edge of the other polygon and split it
			}
		}}
	}}
	// cut edges of neighbour polygons so no T-junctions originate

	{int n_index = 0;
	for(std::vector<TPolygon>::const_iterator p_polygon = (p_polygon_indices)?
	   m_poly_list.begin() + (n_index = (*p_polygon_indices)) : m_poly_list.begin();
	   n_polygon_num > 0; n_polygon_num --, (p_polygon_indices)? p_polygon = m_poly_list.begin() +
	   (n_index = (*(++ p_polygon_indices))) : (n_index ++, p_polygon ++)) { // todo - see what g++ thinks of such a construct
		CRefPolygon t_tmp_poly[2];

		if(!(t_tmp_poly[0] = *(*p_polygon).p_poly) ||
		   !t_tmp_poly[0].Split(r_t_plane, t_tmp_poly[1], plane_Front, f_epsilon_ex))
			return false;
		// try to split the polygon

		if(t_tmp_poly[0].b_IsTiny() || t_tmp_poly[1].b_IsTiny())
			continue;
		// not interested in polygons that aren't split

		TPolygon t_poly;
		t_poly = (*p_polygon);
		t_poly.n_neighbour_edge_num = 0;
		t_poly.p_neighbour_index_list = 0;
		if(!(t_poly.p_poly = new CRefPolygon))
			return false;
		// new polygon wrapper

		if(!(*t_poly.p_poly = t_tmp_poly[1]) || !(*(*p_polygon).p_poly = t_tmp_poly[0])) {
			delete t_poly.p_poly;
			return false;
		}
		// copy polygons

		m_b_poly_hash_dirty_flag = true;
		m_b_vertex_normal_smooth_group_dirty_flag = true;
		m_b_vertex_normal_smooth_all_dirty_flag = true;
		m_b_poly_neighbours_dirty_flag = true;
		// raise dirty flags

		m_poly_list.push_back(t_poly);
		if(!m_poly_list.size()) {
			delete t_poly.p_poly;
			return false;
		}
	}}
	// now split polygons themselves (have to do it now, because we
	// needed to keep indices of polygons and neighbour lists intact)

	// t_odo - create new CRefPolygon, copy t_tmp_poly[1] inside it and push_back into m_poly_list,
	// copy t_tmp_poly[0] into (*p_polygon)->p_poly, raise dirty flags

	return true;
}

/*
 *	bool CGeometry::Extrude(const int *p_polygon_indices, int n_polygon_num,
 *		float f_amount, bool b_smooth, bool b_smooth_sides)
 *		- extrude selected polygons by their averaged normal times f_amount
 *		  (normals are averaged per group of neighbour polygons)
 *		  and create new polygons (quads) at their edges to connect them with
 *		  the rest of the mesh
 *		- for f_amount under f_epsilon, no side polygons are created
 *		- if b_smooth is true, new surface will be smoothed with the rest of object
 *		  (b_smooth_sides is ignored in such a case)
 *		- if b_smooth_sides is true, new polygons at edges of extruded ones will
 *		  be smooth. otherwise sides will be flat
 */
bool CGeometry::Extrude(const int *p_polygon_indices, int n_polygon_num,
	float f_amount, bool b_smooth, bool b_smooth_sides)
{
	if(!UpdatePolygonNeighbours() || !UpdatePolygonHash())
		return false;

	m_b_poly_neighbours_dirty_flag = true;
	m_b_vertex_normal_smooth_all_dirty_flag = true;
	m_b_vertex_normal_smooth_group_dirty_flag = true;
	m_b_bounding_box_dirty_flag = true;
	m_b_vertex_hash_dirty_flag = true;
	// set flags

	std::vector<int> poly_group; // current poly group, being extruded

	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = -1;
	// clear vertex flags ...

	for(int i = (p_polygon_indices)? *p_polygon_indices : 0; n_polygon_num > 0;
	   n_polygon_num --, (p_polygon_indices)? i = *(++ p_polygon_indices) : i ++) { // todo - see what g++ thinks of such a construct
		poly_group.push_back(i);
		bool b_added;
		do {
			b_added = false;
			int n_poly_num = n_polygon_num - 1;
			const int *p_poly_indices = (p_polygon_indices)? p_polygon_indices + 1 : 0;
			for(int n = (p_poly_indices)? *p_poly_indices : i + 1; n_poly_num > 0 && !b_added;
			   n_poly_num --, (p_poly_indices)? n = *(++ p_poly_indices) : n ++) {
				std::vector<TPolygon>::const_iterator p_wrapper = m_poly_list.begin() + n;
				for(int j = 0; j < poly_group.size() && !b_added; j ++) {
					for(std::vector<int> **p_list = (*p_wrapper).p_neighbour_index_list,
					   **p_end = (*p_wrapper).p_neighbour_index_list +
					   (*p_wrapper).n_neighbour_edge_num; p_list < p_end; p_list ++) {
						for(std::vector<int>::const_iterator p_poly_index = (*p_list)->begin();
						   p_poly_index < (*p_list)->end(); p_poly_index ++) {
							if(*p_poly_index == poly_group[j]) {
								poly_group.push_back(*p_poly_index);
								if(!poly_group.size())
									return false;
								n_polygon_num --;
								if(p_polygon_indices)
									i = *(++ p_polygon_indices);
								else
									i ++;
								b_added = true;
								break;
							}
						}
					}
				}
			}
		} while(b_added);
		// create polygon group, removing polygon indices from to-do list

		Vector3f v_extrude(0, 0, 0);
		if(f_amount >= f_epsilon) {
			for(std::vector<int>::const_iterator p_wrapper_index = poly_group.begin();
			   p_wrapper_index < poly_group.end();)
				v_extrude += m_poly_list[*(p_wrapper_index ++)].p_poly->t_Normal().v_normal;
			v_extrude.Normalize(f_amount);
			_ASSERTE((float)fabs(v_extrude.f_Length() - f_amount) < f_epsilon);
		}
		// calc extrude vector

		int n_cap_surface = (b_smooth)? -1 : m_n_surface_space ++;
		// get new surface id for cap when necessary

		for(std::vector<int>::const_iterator p_wrapper_index = poly_group.begin();
		   p_wrapper_index < poly_group.end(); p_wrapper_index ++) {
			TPolygon *p_wrapper = &m_poly_list[*p_wrapper_index];
			if(!b_smooth)
				p_wrapper->n_surface = n_cap_surface;
			CRefPolygon extruded_vertex_holder;
			std::vector<bool> active_vertex_list; // contain true if vertex is on edge that needs side polygon
			for(int j = 0; j < p_wrapper->p_poly->n_Vertex_Num(); j ++) {
				TVertex *p_vertex = p_wrapper->p_poly->r_p_Vertex(j);
				if(p_vertex->n_flag == -1) {
					bool b_make_copy = false;

					std::vector<int> *p_hash = m_p_poly_hash_table[n_HashFunction(p_vertex,
						m_n_poly_hash_table_size)];
					for(std::vector<int>::const_iterator p_poly_index = p_hash->begin();
					   p_poly_index < p_hash->end(); p_poly_index ++) {
						bool b_referencing = false;
						bool b_surrounded_by_edges_with_neighbours = true;
						TPolygon *p_ref_wrapper = &m_poly_list[*p_poly_index];
						for(int l = 0, n_prev_l = p_ref_wrapper->p_poly->n_Vertex_Num() - 1;
						   l < p_ref_wrapper->p_poly->n_Vertex_Num(); n_prev_l = l ++) {
							if(p_ref_wrapper->p_poly->r_p_Vertex(l) == p_vertex) {
								b_referencing = true;
								if(!p_ref_wrapper->p_neighbour_index_list[l]->size() ||
									!p_ref_wrapper->p_neighbour_index_list[n_prev_l]->size())
									b_surrounded_by_edges_with_neighbours = false;
								else {
									bool b_found = false;
									for(int n = 0, n_index = n_prev_l; n < 2 && !b_found; n ++, n_index = l) {
										for(std::vector<int>::const_iterator p_poly_index =
										   p_ref_wrapper->p_neighbour_index_list[n_index]->begin();
										   p_poly_index < p_ref_wrapper->p_neighbour_index_list[
										   n_index]->end(); p_poly_index ++) {
											for(std::vector<int>::const_iterator p_index =
											   poly_group.begin(); p_index < poly_group.end();) {
												if(*(p_index ++) == *p_poly_index) {
													b_found = true;
													break;
												}
											}
										}
									}
									if(!b_found)
										b_surrounded_by_edges_with_neighbours = false;
								}
								break;
							}
						}
						if(!b_referencing)
							continue;
						// does the polygon really reference this vertex ?

						bool b_found = false;
						if(b_surrounded_by_edges_with_neighbours) {
							for(std::vector<int>::const_iterator p_index =
							   poly_group.begin(); p_index < poly_group.end();) {
								if(*(p_index ++) == *p_poly_index) {
									b_found = true;
									break;
								}
							}
						}
						// does the polygon belong to our group?

						if(!b_found || !b_surrounded_by_edges_with_neighbours) {
							b_make_copy = true;
							break;
						}
					}
					// see if vertex is referenced by other polygons
					// or it's surrounded by edges with no neighbours
					// (in that case we have to make it's copy for side polys)

					if(b_make_copy) {
						p_vertex->n_flag = m_vertex_list.size(); // set flag to index of new extruded vertex
						if(!(p_vertex = p_AddVertex(*p_vertex)))
							return false;
						p_vertex->v_pos += v_extrude;
						// copy vertex
					} else {
						p_vertex->n_flag = -2;
						p_vertex->v_pos += v_extrude;
					}
				} else {
					if(p_vertex->n_flag != -2)
						p_vertex = m_vertex_list[p_vertex->n_flag];
				}
				// find appropriate vertex or it's extruded version

				TRefVertex t_ref_vertex;
				t_ref_vertex.p_geom = this;
				t_ref_vertex.p_vertex = p_vertex;

				if(!extruded_vertex_holder.Add_Vertex(t_ref_vertex))
					return false;
			}
			// create polygon which holds offset versions of all vertices of curent poly

			for(int j = 0, n_prev_j = extruded_vertex_holder.n_Vertex_Num() - 1;
			   j < extruded_vertex_holder.n_Vertex_Num(); n_prev_j = j ++) {
				bool b_neighbours_belong_to_extrude_group = false;
				for(std::vector<int>::const_iterator p_poly_index =
				   p_wrapper->p_neighbour_index_list[n_prev_j]->begin(); p_poly_index <
				   p_wrapper->p_neighbour_index_list[n_prev_j]->end(); p_poly_index ++) { // using polyneighbours with vertex indices
					for(std::vector<int>::const_iterator p_index =
					   poly_group.begin(); p_index < poly_group.end();) {
						if(*(p_index ++) == *p_poly_index) {
							b_neighbours_belong_to_extrude_group = true;
							break;
						}
					}
				}
				if(b_neighbours_belong_to_extrude_group)
					continue;
				// side polys on active edges only ... // f_ixme - should it search rather in p_wrapper->p_neighbour_index_list[j]? no, edge are vertices i and (i+1)%vertex_num

				TRefVertex t_quad[4];
				t_quad[0].p_geom = this;
				t_quad[1].p_geom = this;
				t_quad[2].p_geom = this;
				t_quad[3].p_geom = this;
				//
				if(b_smooth) {
					t_quad[0].p_vertex = extruded_vertex_holder.r_p_Vertex(j);
					t_quad[1].p_vertex = extruded_vertex_holder.r_p_Vertex(n_prev_j);
					t_quad[2].p_vertex = p_wrapper->p_poly->r_p_Vertex(n_prev_j);
					t_quad[3].p_vertex = p_wrapper->p_poly->r_p_Vertex(j); // using polyneighbours with vertex indices
				} else {
					if(!(t_quad[0].p_vertex = p_AddVertex(extruded_vertex_holder.t_Vertex(j))) ||
					   !(t_quad[1].p_vertex = p_AddVertex(extruded_vertex_holder.t_Vertex(n_prev_j))) ||
					   !(t_quad[2].p_vertex = p_AddVertex(p_wrapper->p_poly->t_Vertex(n_prev_j))) ||
					   !(t_quad[3].p_vertex = p_AddVertex(p_wrapper->p_poly->t_Vertex(j)))) // using polyneighbours with vertex indices
						return false;
					// we have to copy vertices for non-smooth side polys (we could save
					// vertex pairs for not-smooth extrusion, with smooth side polys, but
					// this code is mess already)
				}
				// prepare quad vertices ...

				TPolygon t_poly;
				t_poly = *p_wrapper;
				t_poly.n_neighbour_edge_num = 0;
				t_poly.p_neighbour_index_list = 0;
				if(b_smooth)
					t_poly.n_surface = p_wrapper->n_surface;
				else {
					if(b_smooth_sides)
						t_poly.n_surface = m_n_surface_space;
					else
						t_poly.n_surface = m_n_surface_space ++;
				}
				if(!(t_poly.p_poly = new CRefPolygon))
					return false;
				if(!t_poly.p_poly->Add_Vertex(t_quad, 4)) {
					delete t_poly.p_poly;
					return false;
				}

				t_poly.p_poly->Calc_Normal();

				m_b_poly_hash_dirty_flag = true;
				m_b_vertex_normal_smooth_group_dirty_flag = true;
				m_b_vertex_normal_smooth_all_dirty_flag = true;
				m_b_poly_neighbours_dirty_flag = true;

				m_poly_list.push_back(t_poly);
				if(!m_poly_list.size()) {
					delete t_poly.p_poly;
					return false;
				}

				p_wrapper = &m_poly_list[poly_group[i]]; // p_wrapper changed address, we have to re-locate him ...
			}
			// create polygon sides ...

			_ASSERTE(extruded_vertex_holder.n_Vertex_Num() == p_wrapper->p_poly->n_Vertex_Num());

			for(int j = 0; j < extruded_vertex_holder.n_Vertex_Num(); j ++)
				p_wrapper->p_poly->r_p_Vertex(j) = extruded_vertex_holder.r_p_Vertex(j);
			// copy new vertices ...

			extruded_vertex_holder.Delete();
		}

		if(!b_smooth && b_smooth_sides)
			m_n_surface_space ++;
		// increase surface space for smooth sides arround whole group

		poly_group.clear();
	}

	return true;
}

/*
 *	bool CGeometry::_Sort_PolyList(const Plane3f &r_plane, std::vector<int> &r_onplane_list,
 *		std::vector<int> &r_list, int &r_n_lo, int &r_n_hi, int &r_n_duplicated,
 *		std::vector<Vector2i> &r_collider_list, bool b_fill_remove_onplane, bool b_fill_collider)
 *		- function for sorting index list of polygons by plane r_plane
 *		- r_onplane_list may contain list of all polygons collinear to r_plane
 *		  for collider search purposes
 *		- r_list contain list of polygon indices to be sorted (sorting means everything in lower
 *		  half of index list will lie back to r_plane, the other half will lie in front of it.
 *		  split and onplane polygons are duplicated so they lie in both halves)
 *		- r_n_lo and r_n_hi contain lower and upper bounds of r_list
 *		- the middle of the list is r_n_lo == r_n_hi (after this function call)
 *		- collider list can be filled by index pairs of colliding polygons, first in pair
 *		  is from r_list, second is from r_onplane_list
 *		- when b_fill_remove_onplane is true, all onplane polygons are removed from the list
 *		  and added to r_onplane_list
 *		- when b_fill_collider is true, for every polygon, being split by r_plane it run search
 *		  trough r_onplane_list and record all collisions (doesn't perform any splits)
 */
bool CGeometry::_Sort_PolyList(const Plane3f &r_plane, std::vector<int> &r_onplane_list,
	std::vector<int> &r_list, int &r_n_lo, int &r_n_hi, int &r_n_duplicated,
	std::vector<Vector2i> &r_collider_list, bool b_fill_remove_onplane, bool b_fill_collider)
{
	r_n_duplicated = 0;

	while(r_n_lo < r_n_hi) {
		do {
			int n_pos = m_poly_list[r_list[r_n_lo]].p_poly->n_Plane_Pos(r_plane);

			if(n_pos == plane_Front)
				break;

			if(n_pos == plane_Split && b_fill_collider) {
				CRefPolygon *p_tested_poly = m_poly_list[r_list[r_n_lo]].p_poly;
				for(int i = 0; i < r_onplane_list.size(); i ++) {
					if(p_tested_poly->b_Collide(*m_poly_list[r_onplane_list[i]].p_poly)) {
						r_collider_list.push_back(Vector2i(r_list[r_n_lo], r_onplane_list[i]));
						if(!r_collider_list.size())
							return false;
					}
				}
			}
			// in case of split position and request for collider list, we look
			// for intersections between tested poly and all polygons in r_onplane_list

			if(n_pos == plane_Split || (n_pos == plane_Onplane && !b_fill_remove_onplane)) {
				r_n_duplicated ++;

				r_list.insert(&r_list[r_n_hi + 1], r_list[r_n_lo]);
				if(!r_list.size())
					return false;
				// we found poly that should be on both planes of the final list
				// we should leave it here and copy it past current r_n_hi pointer as well
			} else if(n_pos == plane_Onplane && b_fill_remove_onplane) {
				r_onplane_list.push_back(r_list[r_n_lo]);
				if(!r_onplane_list.size())
					return false;

				r_list.erase(&r_list[r_n_lo --]);

				r_n_duplicated --;
				r_n_hi --;
				// when we find onplane polygon, we add it to the list and erase from
				// the index list (we don't want it to be reused as pivot index isn't)
			}
		} while(++ r_n_lo < r_n_hi);

		CRefPolygon *p_tested_poly = m_poly_list[r_list[r_n_lo]].p_poly;
		for(int i = 0; i < r_onplane_list.size(); i ++) {
			if(p_tested_poly->b_Collide(*m_poly_list[r_onplane_list[i]].p_poly)) {
				r_collider_list.push_back(Vector2i(r_list[r_n_lo], r_onplane_list[i]));
				if(!r_collider_list.size())
					return false;
			}
		}

		if(r_n_lo < r_n_hi) {
			do {
				int n_pos = m_poly_list[r_list[r_n_hi]].p_poly->n_Plane_Pos(r_plane);

				if(n_pos == plane_Back)
					break;

				if(n_pos == plane_Split && b_fill_collider) {
					CRefPolygon *p_tested_poly = m_poly_list[r_list[r_n_hi]].p_poly;
					for(int i = 0; i < r_onplane_list.size(); i ++) {
						if(p_tested_poly->b_Collide(*m_poly_list[r_onplane_list[i]].p_poly)) {
							r_collider_list.push_back(Vector2i(r_list[r_n_hi], r_onplane_list[i]));
							if(!r_collider_list.size())
								return false;
						}
					}
				}
				// in case of split position and request for collider list, we look
				// for intersections between tested poly and all polygons in r_onplane_list

				if(n_pos == plane_Split || (n_pos == plane_Onplane && !b_fill_remove_onplane)) {
					r_n_duplicated ++;

					int n_temp = r_list[r_n_hi ++]; // for some reason it inserts bad array element when used directly
					r_list.insert(&r_list[r_n_lo ++], n_temp);
					if(!r_list.size())
						return false;
					// we found poly that should be on both planes of the final list
					// we should leave it here and copy it preceding current r_n_lo pointer as well
				} else if(n_pos == plane_Onplane && b_fill_remove_onplane) {
					r_onplane_list.push_back(r_list[r_n_hi]);
					if(!r_onplane_list.size())
						return false;

					r_list.erase(&r_list[r_n_hi]);

					r_n_duplicated --;
					// when we find onplane polygon, we add it to the list and erase from
					// the index list (we don't want it to be reused as pivot index isn't)
				}
			} while(r_n_lo < -- r_n_hi);
		}

		p_tested_poly = m_poly_list[r_list[r_n_hi]].p_poly;
		for(int i = 0; i < r_onplane_list.size(); i ++) {
			if(p_tested_poly->b_Collide(*m_poly_list[r_onplane_list[i]].p_poly)) {
				r_collider_list.push_back(Vector2i(r_list[r_n_hi], r_onplane_list[i]));
				if(!r_collider_list.size())
					return false;
			}
		}

        if(r_n_lo < r_n_hi) {
            int n_temp = r_list[r_n_lo];
            r_list[r_n_lo] = r_list[r_n_hi];
            r_list[r_n_hi] = n_temp;
        }
	}

	return true;
}

/*
 *	bool CGeometry::FindColliders(std::vector<int> &r_poly_indices_a,
 *		std::vector<int> &r_poly_indices_b, std::vector<Vector2i> &r_collider_list)
 *		- find collider pairs, firs of colliders is from list r_poly_indices_a, the second
 *		  one from r_poly_indices_b
 *		- r_poly_indices_a and r_poly_indices_b are modified during process, which means
 *		  some indices are copied multiple times into array. you can sort them and remove
 *		  repeating indices if you need lists after call of this function (or you can make
 *		  copies of them before calling this function)
 *		- return true if everything was ok, false in case there as not enough memory
 *		- collider list *might* contain duplicate records, you should sort it and skip
 *		  repeating collisions // fixme - does it contain those duplicate records?
 */
bool CGeometry::FindColliders(std::vector<int> &r_poly_indices_a,
	std::vector<int> &r_poly_indices_b, std::vector<Vector2i> &r_collider_list)
{
	std::stack<Vector4i> stack;
	stack.push(Vector4i(0, r_poly_indices_a.size() - 1, 0, r_poly_indices_b.size() - 1));
	if(!stack.size())
		return false;

	Vector4i v_add(0, 0, 0, 0);
	// stack contains sub-array bounds, sorted ascendingly so when we add items into this array,
	// we need to shift all the bounds in the stack. it's better to offset them 

	while(stack.size()) {
		Vector4i v_array_bounds = stack.top() + v_add;

		int lo_a = v_array_bounds.x;
		int hi_a = v_array_bounds.y;
		int lo_b = v_array_bounds.z;
		int hi_b = v_array_bounds.w;

		if(lo_a >= hi_a || lo_b >= hi_b) {
			stack.pop();
			continue;
		} else if(lo_b == hi_b - 1 || lo_a == hi_a - 1) {
			for(int i = lo_b; i <= hi_b; i ++) {
				for(int j = lo_a; j <= hi_a; j ++) {
					if(m_poly_list[r_poly_indices_a[j]].p_poly->b_Collide(
					   *m_poly_list[r_poly_indices_b[i]].p_poly)) {
						r_collider_list.push_back(Vector2i(r_poly_indices_a[j],
							r_poly_indices_b[i]));
						if(!r_collider_list.size())
							return false;
					}
				}
			}
			// simple o(n2) test all polygons (there shouldn't be >4 of them by now ...)

            stack.pop();
			continue;
		}

		int n_pivot = r_poly_indices_b[(lo_b + hi_b) / 2];
        r_poly_indices_b[(lo_b + hi_b) / 2] = r_poly_indices_b[hi_b];
        r_poly_indices_b[hi_b] = n_pivot;
		// pick a pivot, copy it to the end of the array

		std::vector<int> split_poly_list;

		split_poly_list.push_back(n_pivot);
		if(!split_poly_list.size())
			return false;
		// we manage list of all polys, onplane with pivot (we don't want them
		// in final array since they would cause re-splitting polygons, lying
		// on both sides of the index list and thus creating new polygons, missing
		// in one of the lists)

		int n_duplicated_polys;

		if(!_Sort_PolyList(m_poly_list[n_pivot].p_poly->t_Normal(), split_poly_list,
		   r_poly_indices_b, lo_b, hi_b, n_duplicated_polys, r_collider_list, true, false))
			return false;
		// sort splitter polys, find polys onplane to pivot and remove them from array

		v_add += Vector4i(0, 0, n_duplicated_polys, n_duplicated_polys);
		v_array_bounds.w += n_duplicated_polys;
		// update array bounds

        r_poly_indices_b[v_array_bounds.w] = r_poly_indices_b[hi_b];
        r_poly_indices_b[hi_b] = n_pivot;
		// sort b-array, update indices so the're pointing directly into array

		if(!_Sort_PolyList(m_poly_list[n_pivot].p_poly->t_Normal(), split_poly_list,
		   r_poly_indices_a, lo_a, hi_a, n_duplicated_polys, r_collider_list, false, true))
			return false;
		// sort splitted polys, find polys colliding onplane

		v_array_bounds.y += n_duplicated_polys;
		v_add += Vector4i(n_duplicated_polys, n_duplicated_polys, 0, 0);
		// sort a-array, update indices so the're pointing directly into array

		stack.pop();
		// remove current interval

		stack.push(Vector4i(hi_a /*- (lo_a == v_array_bounds.x)*/
			/*fixme: when lo_a == hi_a and lo_a == v_array_bounds.x, which can be zero, we'd get
			negative index*/, v_array_bounds.y, hi_b + 1, v_array_bounds.w) - v_add); // hi first!!! it's stack and i want to have it sorted ascendingly
		stack.push(Vector4i(v_array_bounds.x, lo_a - 1 + (hi_a == v_array_bounds.y), v_array_bounds.z, lo_b - 1) - v_add); // (hi_a == v_array_bounds.y) is because we have pivotless array, which needs special care
		// indices were offset by previous add on the beginning of this loop and
		// they were also offset in sorting process by newly created polygons
		if(stack.size() < 2)
			return false;
		// non-recursive recurse
	}
	// find colliders ...

	return true;
}

/*
 *	int CGeometry::Hollow(float f_hollow_size)
 *		- hollow geometry. it's done by copying every polygon, reverbing it and offseting
 *		  by f_hollow_size
 */
int CGeometry::Hollow(float f_hollow_size)
{
	return false;
}

/*
 *	static int qsort_compare_Vector2i(const void *p_a, const void *p_b)
 *		- function for comparison of Vector2i, used as pair of indices
 */
/*static int qsort_compare_Vector2i(const void *p_a, const void *p_b)
{
	if(((const Vector2i*)p_a)->x != ((const Vector2i*)p_b)->x)
		return ((const Vector2i*)p_a)->x - ((const Vector2i*)p_b)->x;
	return ((const Vector2i*)p_a)->y - ((const Vector2i*)p_b)->y;
}*/

/*
 *	static int qsort_compare_int(const void *p_a, const void *p_b)
 *		- function for comparison of int
 */
static int qsort_compare_int(const void *p_a, const void *p_b)
{
	return *((const int*)p_a) - *((const int*)p_b);
}

/*
 *	bool CGeometry::Boolean(const CGeometry &r_geometry, bool b_mesh_a, bool b_mesh_b,
 *		bool b_mesh_a_inside_b, bool b_mesh_b_inside_a, bool b_flip_a, bool b_flip_b)
 *		- calculate boolean operation on two meshes (this and r_geometry)
 *		- return false if not enough memory, true on success
 *		- not guaranteed to work on self (it has no sense doing booleans on self since
 *		  there are just two possible results - keep mesh intact or erase everything)
 *		- r_geometry is not modified
 *		- set b_mesh_a to true if you want to keep parts of mesh a (this)
 *		  lying outside mesh b (r_geometry), false if you want them to be removed
 *		- set b_mesh_b to true if you want to keep parts of mesh b (r_geometry)
 *		  lying outside mesh a (this), false if you want them to be removed
 *		- set b_mesh_a_inside_b to true if you want to keep parts of mesh a (this)
 *		  lying inside mesh b (r_geometry), false if you want them to be removed
 *		- set b_mesh_b_inside_a to true if you want to keep parts of mesh b (r_geometry)
 *		  lying inside mesh a (this), false if you want them to be removed
 *		- set b_flip_a to true if you want to flip normals (and vertex order) for polygons
 *		  of mesh a (this) lying inside mesh b (r_geometry), false to keep as is
 *		- set b_flip_b to true if you want to flip normals (and vertex order) for polygons
 *		  of mesh b (r_geometry) lying inside mesh a (this), false to keep as is
 *		- Boolean(x, true, false, false, true, false, true) performs mesh subtraction (a - b)
 *		- Boolean(x, false, true, true, false, true, false) performs mesh subtraction (b - a)
 *		- Boolean(x, true, true, false, false, false, false) performs mesh union (a + b)
 *		- Boolean(x, false, false, true, true, false, false) performs mesh intersection (a * b)
 *		- Boolean(x, true, false, true, false, false, false) slices mesh a by mesh b
 *		  (useful for lowpoly modelling)
 *		- Boolean(x, true, true, true, true, true, true) performs mesh nonequivalence (a xor b)
 *		  (it means it keeps everything aparts from intersection)
 */
bool CGeometry::Boolean(const CGeometry &r_geometry, bool b_mesh_a, bool b_mesh_b,
	bool b_mesh_a_inside_b, bool b_mesh_b_inside_a, bool b_flip_a, bool b_flip_b)
{
	int n_polys_a = m_poly_list.size();

	if(!(*this += r_geometry))
		return false;

	int n_polys_b = m_poly_list.size() - n_polys_a;

	std::vector<int> poly_indices_a;
	std::vector<int> poly_indices_b;

	for(int i = 0; i < n_polys_a; i ++) // todo - decide which mesh should go where (perform some testing)
		poly_indices_a.push_back(i);
	for(int i = n_polys_a; i < m_poly_list.size(); i ++)
		poly_indices_b.push_back(i);
	if(poly_indices_a.size() != n_polys_a ||
	   poly_indices_b.size() != n_polys_b)
		return false;
	// put polygon indices into some structure where we can insert and swap values

	for(std::vector<int>::const_iterator p_index_a = poly_indices_a.begin();
	   p_index_a < poly_indices_a.end(); p_index_a ++) {
		for(std::vector<int>::const_iterator p_index_b = poly_indices_b.begin();
		   p_index_b < poly_indices_b.end(); p_index_b ++) {
			if(!m_poly_list[*p_index_a].p_poly->b_Collide(*m_poly_list[*p_index_b].p_poly))
				continue;
			for(int i = 0, n_index_a = *p_index_a, n_index_b = *p_index_b; i < 2; i ++,
			   n_index_a = *p_index_b, n_index_b = *p_index_a) {
				CRefPolygon temp_poly[2];

				if(!(temp_poly[0] = *m_poly_list[n_index_a].p_poly))
					return false;
				if(!temp_poly[0].Split(m_poly_list[n_index_b].p_poly->t_Normal(),
				   temp_poly[1], plane_Front, f_epsilon))
					return false;
				// try to split polygon (first pass a, second pass b)

				if(!temp_poly[0].b_IsTiny() && !temp_poly[1].b_IsTiny()) {
					TPolygon t_new_poly = m_poly_list[n_index_a];
					t_new_poly.p_neighbour_index_list = 0;
					t_new_poly.n_neighbour_edge_num = 0;
					if(!(t_new_poly.p_poly = new CRefPolygon))
						return false;
					if(!m_poly_list[n_index_a].p_poly->Split(
					   m_poly_list[n_index_b].p_poly->t_Normal(),
					   *t_new_poly.p_poly, plane_Front, f_epsilon)) {
						delete[] t_new_poly.p_poly;
						return false;
					}
					// create a new polygon by splitting the old one by pivot plane

					m_b_poly_hash_dirty_flag = true;
					m_b_vertex_normal_smooth_group_dirty_flag = true;
					m_b_vertex_normal_smooth_all_dirty_flag = true;
					m_b_poly_neighbours_dirty_flag = true;
					// invalidate hash tables

					m_poly_list.push_back(t_new_poly);
					if(!m_poly_list.size()) {
						delete t_new_poly.p_poly;
						return false;
					}
					// add polygon to the list

					if(!i) {
						poly_indices_a.push_back(m_poly_list.size() - 1);
						if(!poly_indices_a.size())
							return false;
					} else {
						poly_indices_b.push_back(m_poly_list.size() - 1);
						if(!poly_indices_b.size())
							return false;
					}
					// this poly is in front of pivot polygon
				}
				// polygon from set a can be split into two valid polygons
			}
		}
	}
	// split all the polygons that does intersect or touch

	std::vector<int> poly_indices_a_inside;
	std::vector<int> poly_indices_b_inside;

	for(int n = 0; n < 2; n ++) {
		std::vector<int> *p_poly_indices_a;
		std::vector<int> *p_poly_indices_a_inside;
		std::vector<int> *p_poly_indices_b;
		std::vector<int> *p_poly_indices_b_inside;

		if(!n) {
			p_poly_indices_a = &poly_indices_a;
			p_poly_indices_a_inside = &poly_indices_a_inside;
			p_poly_indices_b = &poly_indices_b;
			p_poly_indices_b_inside = &poly_indices_b_inside;
		} else {
			p_poly_indices_b = &poly_indices_a;
			p_poly_indices_b_inside = &poly_indices_a_inside;
			p_poly_indices_a = &poly_indices_b;
			p_poly_indices_a_inside = &poly_indices_b_inside;
		}

		for(std::vector<int>::iterator p_index_a = p_poly_indices_a->begin();
		   p_index_a < p_poly_indices_a->end(); p_index_a ++) {
			float f_t = -1.0f;
			EPlanePos n_pos;
			Vector3f v_org = m_poly_list[*p_index_a].p_poly->v_Center();
			Vector3f v_dir = m_poly_list[*p_index_a].p_poly->t_Normal().v_normal;
			float f_polarity = 1.0f; // polarity = ray dot normal

			for(int n_test = 0; n_test < 2; n_test ++) {
				for(int m = 0; m < 2; m ++) {
					std::vector<int> *p_test_poly_indices = (m)? p_poly_indices_b_inside :
						p_poly_indices_b;
					for(std::vector<int>::const_iterator p_index =
					   p_test_poly_indices->begin(); p_index < p_test_poly_indices->end();) {
						CRefPolygon *p_poly_b = m_poly_list[*(p_index ++)].p_poly;
						if((float)fabs(p_poly_b->t_Normal().v_normal.f_Dot(v_dir)) < f_epsilon)
							continue;
						// can't test polygons, collinear with ray for intersection

						float f_cur_t, f_cur_u, f_cur_v;
						if(!p_poly_b->b_MT_RayHit(v_org, v_dir, f_cur_t, f_cur_u, f_cur_v) ||
						   f_cur_t < -f_epsilon /*.0f*/ || (f_cur_t > f_t && f_t > -.5f))
							continue;
						// test for ray intersection

						f_t = f_cur_t;
						if((n_pos = p_poly_b->t_Normal().n_Vector_Pos(v_org)) == plane_Onplane)
							n_pos = (p_poly_b->t_Normal().v_normal.f_Dot(v_org) > 0)? plane_Back : plane_Front;
						// record a new nearest intersection with triangle from b and see if it's front or back
					}
				}

				// todo - see if we can move if(n_test) break; here

				if(f_t < -.5f) {
					// if ray hit no polygon from the other group
					for(int m = 0, b_found = false; m < 2 && !b_found; m ++) {
						std::vector<int> *p_test_poly_indices = (m)? p_poly_indices_b_inside :
							p_poly_indices_b;
						for(std::vector<int>::const_iterator p_index =
						   p_test_poly_indices->begin(); p_index < p_test_poly_indices->end();) {
							CRefPolygon *p_poly_b = m_poly_list[*(p_index ++)].p_poly;
							Vector3f v_tar = p_poly_b->v_Center();
							Vector3f v_new_dir = v_tar - v_org;
							v_new_dir.Normalize();

							if((float)fabs(p_poly_b->t_Normal().v_normal.f_Dot(v_new_dir)) <
							   f_epsilon || -(m_poly_list[*p_index_a].p_poly->t_Normal(
							   ).v_normal.f_Dot(v_new_dir)) < f_epsilon)
								continue;
							// can't use this ray because of lack of precision
							// (ray collinear with either tested or ray-target surface)

							f_polarity = -(m_poly_list[*p_index_a].p_poly->t_Normal(
								).v_normal.f_Dot(v_new_dir));
							// determine ray polarity

							if(f_polarity < 0)
								continue;
							// todo - see why can't we use rays with negative polarity

							b_found = true;
							v_dir = v_new_dir;

							break;
						}
					}
					// find polygon we could hit (find new direction)
				} else {
					if(n_test)
						break;
				}
			}

			if(f_t > -.5f && ((f_polarity > 0 && n_pos == plane_Back) ||
			   (f_polarity < 0 && n_pos == plane_Front))) {
				p_poly_indices_a_inside->push_back(*p_index_a);
				if(!p_poly_indices_a_inside->size())
					return false;
				p_poly_indices_a->erase(p_index_a --);
				// it was inside -> add id to another list and erase it
			}
		}
	}
	// create lists of polygons that are inside / outside

	std::vector<int> polys_to_delete;

	if(!b_mesh_a) {
		polys_to_delete.insert(polys_to_delete.end(),
			poly_indices_a.begin(), poly_indices_a.end());
		if(!polys_to_delete.size() && poly_indices_a.size())
			return false;
	}
	poly_indices_a.clear();
	if(!b_mesh_b) {
		polys_to_delete.insert(polys_to_delete.end(),
			poly_indices_b.begin(), poly_indices_b.end());
		if(!polys_to_delete.size() && poly_indices_b.size())
			return false;
	}
	poly_indices_b.clear();
	if(b_mesh_a_inside_b) {
		if(b_flip_a) {
			for(std::vector<int>::const_iterator p_index = poly_indices_a_inside.begin();
			   p_index < poly_indices_a_inside.end(); p_index ++) {
				m_poly_list[*p_index].p_poly->Reverb();
				m_poly_list[*p_index].p_poly->r_t_Normal().v_normal *= -1;
				m_poly_list[*p_index].p_poly->r_t_Normal().f_dist *= -1;
				m_b_poly_neighbours_dirty_flag = true;
				m_b_vertex_normal_smooth_group_dirty_flag = true;
				m_b_vertex_normal_smooth_all_dirty_flag = true;
			}
			// flip normals for parts of a lying inside b if that was required
		}
	} else {
		polys_to_delete.insert(polys_to_delete.end(),
			poly_indices_a_inside.begin(), poly_indices_a_inside.end());
		if(!polys_to_delete.size() && poly_indices_a_inside.size())
			return false;
	}
	poly_indices_a_inside.clear();
	if(b_mesh_b_inside_a) {
		if(b_flip_b) {
			for(std::vector<int>::const_iterator p_index = poly_indices_b_inside.begin();
			   p_index < poly_indices_b_inside.end(); p_index ++) {
				m_poly_list[*p_index].p_poly->Reverb();
				m_poly_list[*p_index].p_poly->r_t_Normal().v_normal *= -1;
				m_poly_list[*p_index].p_poly->r_t_Normal().f_dist *= -1;
				m_b_poly_neighbours_dirty_flag = true;
				m_b_vertex_normal_smooth_group_dirty_flag = true;
				m_b_vertex_normal_smooth_all_dirty_flag = true;
			}
			// flip normals for parts of b lying inside a if that was required
		}
	} else {
		polys_to_delete.insert(polys_to_delete.end(),
			poly_indices_b_inside.begin(), poly_indices_b_inside.end());
		if(!polys_to_delete.size() && poly_indices_b_inside.size())
			return false;
	}
	poly_indices_b_inside.clear();
	// create list of polygons to delete ...

	qsort(&polys_to_delete[0], polys_to_delete.size(), sizeof(int), qsort_compare_int);
	// sort that list

#ifdef _DEBUG
	for(std::vector<int>::const_iterator p_prev_index = polys_to_delete.begin(),
	   p_index = polys_to_delete.begin() + 1; p_index < polys_to_delete.end();
	   p_prev_index = p_index ++)
		_ASSERTE(*p_index != *p_prev_index);
	// check for duplicate records // fixme - do they originate at all ?
#endif

	for(std::vector<int>::const_iterator p_index = polys_to_delete.end() - 1;
	   p_index >= polys_to_delete.begin(); p_index --)
		DeletePolygons(*p_index);
	// delete polygons (backwards so indices aren't affected by deleting array elements)
	// todo - see how does work reverse iterator

	return true;
}

bool CGeometry::Insert_EdgeQuads(int n_tex_channel, bool b_flip, int n_material)
{
	if(n_tex_channel < 0 || n_tex_channel >= 4)
		return false;

	if(m_b_poly_neighbours_dirty_flag && !FindNeighbours())
		return false;
	if(!UpdatePolygonHash())
		return false;

	for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = false;
	// none used

	for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin(),
	   p_end = m_vertex_list.end(); p_vertex < p_end; p_vertex ++) {
		std::vector<int> *p_hash =
			m_p_poly_hash_table[n_HashFunction(*p_vertex, m_n_poly_hash_table_size)];
		// list of all polygons, possibly referencing vertex p_vertex

		for(std::vector<int>::const_iterator p_index = p_hash->begin();
		   p_index < p_hash->end(); p_index ++) {
			CRefPolygon *p_poly = m_poly_list[*p_index].p_poly;
			for(int i = 0; i < p_poly->n_Vertex_Num(); i ++) {
				if(p_poly->r_p_Vertex(i) == *p_vertex) {
					if(!(*p_vertex)->n_flag)
						(*p_vertex)->n_flag = true; // now it's used
					else {
						m_b_poly_hash_dirty_flag = true;
						// we will modify vertex mapping -> set dirty flag

						if(!(p_poly->r_p_Vertex(i) = p_AddVertex(**p_vertex)))
							return false;
						p_poly->r_p_Vertex(i)->n_flag = false;
						// add a new unused vertex
					}
					p_poly->r_p_Vertex(i)->v_texcoord[n_tex_channel] = Vector4f(
						p_poly->t_Normal().v_normal.x, p_poly->t_Normal().v_normal.y,
						p_poly->t_Normal().v_normal.z, p_poly->t_Normal().f_dist);
					// set face normal
					break;
				}
			}
		}

	}
	// take care so every polygon has it's own vertices

	{int n_index = 0;
	for(std::vector<TPolygon>::iterator p_poly = m_poly_list.begin(),
	   p_end = m_poly_list.end(); p_poly < p_end; p_poly ++, n_index ++) {
		TPolygon *p_poly_0 = &*p_poly;

		for(int j = 0, n_prev_j = p_poly_0->n_neighbour_edge_num - 1;
		   j < p_poly_0->n_neighbour_edge_num; n_prev_j = j ++) {
			if(p_poly_0->p_neighbour_index_list[n_prev_j]->size() != 1 ||
			   (*p_poly_0->p_neighbour_index_list[n_prev_j])[0] < n_index) // using polyneighbours with vertex indices
				continue;
			// edges with neighbour count other than 1 aren't processed,
			// edges with polygons with index less than i were processed already

			TVertex *p_vertex_0_a = p_poly_0->p_poly->r_p_Vertex(n_prev_j);
			TVertex *p_vertex_0_b = p_poly_0->p_poly->r_p_Vertex(j); // using polyneighbours with vertex indices

			TPolygon *p_poly_1 = &m_poly_list[(*p_poly_0->p_neighbour_index_list[j])[0]];

			for(int k = 0, n_prev_k = p_poly_1->n_neighbour_edge_num - 1;
			   k < p_poly_1->n_neighbour_edge_num; n_prev_k = k ++) {
				if((*p_poly_1->p_neighbour_index_list[n_prev_k])[0] == n_index) {
					TVertex *p_vertex_1_a = p_poly_1->p_poly->r_p_Vertex(n_prev_k);
					TVertex *p_vertex_1_b = p_poly_1->p_poly->r_p_Vertex(k); // using polyneighbours with vertex indices
					// get pair of vertices that should correspond with vertices
					// from first polygon

					if((p_vertex_0_a->v_pos - p_vertex_1_a->v_pos).f_Length() < f_epsilon &&
					   (p_vertex_0_b->v_pos - p_vertex_1_b->v_pos).f_Length() < f_epsilon) {
						if((p_vertex_0_a->v_pos - p_vertex_1_b->v_pos).f_Length() < f_epsilon &&
						   (p_vertex_0_b->v_pos - p_vertex_1_a->v_pos).f_Length() < f_epsilon) {
							// we have problem, we need to decide which pair is closer
							// (it should be a-a, b-b)

							if((p_vertex_0_a->v_pos - p_vertex_1_b->v_pos).f_Length2() +
							   (p_vertex_0_b->v_pos - p_vertex_1_a->v_pos).f_Length2() <
							   (p_vertex_0_a->v_pos - p_vertex_1_a->v_pos).f_Length2() +
							   (p_vertex_0_b->v_pos - p_vertex_1_b->v_pos).f_Length2()) {
								// but it's a-b, b-a

								TVertex *p_tmp = p_vertex_1_a;
								p_vertex_1_a = p_vertex_1_b;
								p_vertex_1_b = p_tmp;
								// swap vertices
							}
						}
					} else {
						if((p_vertex_0_a->v_pos - p_vertex_1_b->v_pos).f_Length() < f_epsilon &&
						   (p_vertex_0_b->v_pos - p_vertex_1_a->v_pos).f_Length() < f_epsilon) {
							TVertex *p_tmp = p_vertex_1_a;
							p_vertex_1_a = p_vertex_1_b;
							p_vertex_1_b = p_tmp;
							// swap vertices
						} else {
							// we have problem, we need to decide which pair is closer
							// (it should be a-b, b-a)

							if((p_vertex_0_a->v_pos - p_vertex_1_b->v_pos).f_Length2() +
							   (p_vertex_0_b->v_pos - p_vertex_1_a->v_pos).f_Length2() <
							   (p_vertex_0_a->v_pos - p_vertex_1_a->v_pos).f_Length2() +
							   (p_vertex_0_b->v_pos - p_vertex_1_b->v_pos).f_Length2()) {
								// and it's a-b, b-a

								TVertex *p_tmp = p_vertex_1_a;
								p_vertex_1_a = p_vertex_1_b;
								p_vertex_1_b = p_tmp;
								// swap vertices
							}
						}
					}
					// swap vertices so 0_a corresponds to 1_a and 0_b to 1_b respectively

					TRefVertex t_ref_vertex[4];
					t_ref_vertex[0].p_geom = this;
					t_ref_vertex[0].p_vertex = p_vertex_0_a;
					t_ref_vertex[1].p_geom = this;
					t_ref_vertex[1].p_vertex = p_vertex_0_b;
					t_ref_vertex[2].p_geom = this;
					t_ref_vertex[2].p_vertex = p_vertex_1_b;
					t_ref_vertex[3].p_geom = this;
					t_ref_vertex[3].p_vertex = p_vertex_1_a;
					//
					TPolygon t_poly;
					t_poly.n_surface_flags = surf_Solid;
					t_poly.n_neighbour_edge_num = 0;
					t_poly.p_neighbour_index_list = 0;
					memset(t_poly.n_surface_map, 0, sizeof(t_poly.n_surface_map));
					t_poly.n_surface = m_n_surface_space;
					t_poly.n_material = n_material;
					if(!(t_poly.p_poly = new CRefPolygon))
						return false;
					if(!t_poly.p_poly->Add_Vertex(t_ref_vertex, 4)) {
						delete t_poly.p_poly;
						return false;
					}
					//
					if(b_flip)
						t_poly.p_poly->Reverb();
					t_poly.p_poly->Calc_Normal();
					//
					m_b_poly_hash_dirty_flag = true;
					m_b_vertex_normal_smooth_group_dirty_flag = true;
					m_b_vertex_normal_smooth_all_dirty_flag = true;
					m_b_poly_neighbours_dirty_flag = true;
					//
					m_poly_list.push_back(t_poly);
					if(!m_poly_list.size()) {
						delete t_poly.p_poly;
						return false;
					}
					// create a new quad

					break;
				}
				_ASSERTE(k + 1 != p_poly_1->n_neighbour_edge_num);
			}
		}
	}}

	m_n_surface_space ++;
	// we 've added a new quads

	return true;
}

struct TLinkedEdge {
	int n_polygon_index;
	TVertex *p_vertex[2];
	int n_vertex_index[2]; // vertex merge list
	int n_group;
	std::vector<struct TLinkedEdge*> *p_next_edge_list;
};

/*
 *	static int qsort_compare_linkededge(const void *p_a, const void *p_b)
 *		- function for comparison of TLinkedEdge
 *		- used for sorting with first vertex index as primary key
 */
static int qsort_compare_linkededge(const void *p_a, const void *p_b)
{
	const int *p_edge_a = ((const TLinkedEdge*)p_a)->n_vertex_index;
	const int *p_edge_b = ((const TLinkedEdge*)p_b)->n_vertex_index;

	if(p_edge_a[0] - p_edge_b[0])
		return p_edge_a[0] - p_edge_b[0];
	return p_edge_a[1] - p_edge_b[1];
}

/*
 *	static int bsearch_compare_linkededge(const void *p_a, const void *p_b)
 *		- function for comparison of TLinkedEdge
 *		- used for searching, secondary vertex of p_a must match primary of p_b
 */
static int bsearch_compare_linkededge(const void *p_a, const void *p_b)
{
	const int *p_edge_a = ((const TLinkedEdge*)p_a)->n_vertex_index;
	const int *p_edge_b = ((const TLinkedEdge*)p_b)->n_vertex_index;

	return p_edge_a[1] - p_edge_b[0];
}

static void Rotate_Insit_Left(int *p_array, int n_rotate, int n_length)
{
	while(n_rotate) {
		int n_shift = n_rotate;
		while(n_length % n_shift)
			n_shift --;
		n_rotate -= n_shift;

		for(int *p_ptr = p_array, *p_end = p_array + n_length - n_shift; p_ptr < p_end; p_ptr ++) {
			int n_temp = *p_ptr;
			*p_ptr = p_ptr[n_shift];
			p_ptr[n_shift] = n_temp;
		}
	}
}

static void Rotate_Insit_Right(int *p_array, int n_rotate, int n_length)
{
	while(n_rotate) {
		int n_shift = n_rotate;
		while(n_length % n_shift)
			n_shift --;
		n_rotate -= n_shift;

		for(int *p_ptr = p_array + n_length - n_shift - 1, *p_end = p_array; p_ptr >= p_end; p_ptr --) {
			int n_temp = *p_ptr;
			*p_ptr = p_ptr[n_shift];
			p_ptr[n_shift] = n_temp;
		}
	}
}

/*
 *	std::vector<int> *CGeometry::p_Outline_IndexList(const int *p_polygon_indices, int n_polygon_num)
 *		- create list of indices, describing (perhaps non-convex) outline of selected polygons
 *		- outline may contain double edges if selected polygons contain holes or form several
 *		  separate sub-surfaces
 */
std::vector<int> *CGeometry::p_Outline_IndexList(const int *p_polygon_indices, int n_polygon_num)
{
	if(!UpdatePolygonNeighbours())
		return 0;
	// need neighbour info

	std::vector<std::vector<TVertex*>*> *p_merge_table;
	
	if(!(p_merge_table = p_VertexMergeTable(true, f_epsilon)))
		return 0;
	// vertex merge table

	{int n_index = 0;
	for(std::vector<std::vector<TVertex*>*>::iterator p_list = p_merge_table->begin();
	   p_list < p_merge_table->end(); p_list ++, n_index ++) {
		for(std::vector<TVertex*>::iterator p_vertex = (*p_list)->begin(); p_vertex < (*p_list)->end();)
			(*(p_vertex ++))->n_flag = n_index;
		delete *p_list;
	}}
	delete p_merge_table;
	// mark vertex indices ...

	std::vector<TLinkedEdge> edge_list;

	{int n_index = 0;
	for(std::vector<TPolygon>::const_iterator p_polygon = (p_polygon_indices)?
	   m_poly_list.begin() + (n_index = (*p_polygon_indices)) : m_poly_list.begin();
	   n_polygon_num > 0; n_polygon_num --, (p_polygon_indices)? p_polygon = m_poly_list.begin() +
	   (n_index = (*(++ p_polygon_indices))) : (n_index ++, p_polygon ++)) { // todo - see what g++ thinks of such a construct
		for(int j = 0; j < (*p_polygon).n_neighbour_edge_num; j ++) {
			if((*p_polygon).p_neighbour_index_list[j]->size())
				continue;

			CRefPolygon *p_poly = (*p_polygon).p_poly;

			TVertex *p_edge_vertex[2] = {p_poly->r_p_Vertex((j + 1 /*+
				p_poly->n_Vertex_Num()*/) % p_poly->n_Vertex_Num()),
				p_poly->r_p_Vertex(j)}; // using polyneighbours with vertex indices

			TLinkedEdge t_edge = {n_index, {p_edge_vertex[0], p_edge_vertex[1]},
				{p_edge_vertex[0]->n_flag, p_edge_vertex[1]->n_flag}, -1, 0};
			// create an edge

			edge_list.push_back(t_edge);
			if(!edge_list.size())
				return 0;
			// add to list
		}
	}}
	// create list of edges, used just by a single polygon, i.e. outline edges

	qsort(&edge_list[0], edge_list.size(), sizeof(TLinkedEdge), qsort_compare_linkededge);
	// sort edges by first vertex index, then by second vertex index

	for(std::vector<TLinkedEdge>::iterator p_edge = edge_list.begin();
	   p_edge < edge_list.end(); p_edge ++) {
		(*p_edge).p_next_edge_list = 0;

		TLinkedEdge *p_neigh_edge; // t_odo - i screwed it up here, p_edge is name of iterator as well ... fix it
		if(!(p_neigh_edge = (TLinkedEdge*)bsearch(&(*p_edge), &edge_list[0],
		   edge_list.size(), sizeof(TLinkedEdge), bsearch_compare_linkededge)))
			continue;

		if(!((*p_edge).p_next_edge_list = new std::vector<TLinkedEdge*>)) {
			for(std::vector<TLinkedEdge>::iterator p_edge2 = edge_list.begin(); p_edge2 < p_edge; p_edge2 ++) {
				if((*p_edge2).p_next_edge_list)
					delete (*p_edge2).p_next_edge_list;
			}
			return 0;
		}

		for(TLinkedEdge *p_cur_edge = p_neigh_edge; p_cur_edge < edge_list.end() &&
		   p_cur_edge->n_vertex_index[0] == (*p_edge).n_vertex_index[1]; p_cur_edge ++) {
			(*p_edge).p_next_edge_list->push_back(p_cur_edge);
			if(!(*p_edge).p_next_edge_list->size()) {
				for(std::vector<TLinkedEdge>::iterator p_edge2 = edge_list.begin(); p_edge2 <= p_edge; p_edge2 ++) {
					if((*p_edge2).p_next_edge_list)
						delete (*p_edge2).p_next_edge_list;
				}
				return 0;
			}
		}
		for(TLinkedEdge *p_cur_edge = p_neigh_edge - 1; p_cur_edge >= edge_list.begin() &&
		   p_cur_edge->n_vertex_index[0] == (*p_edge).n_vertex_index[1]; p_cur_edge --) {
			(*p_edge).p_next_edge_list->push_back(p_cur_edge);
			if(!(*p_edge).p_next_edge_list->size()) {
				for(std::vector<TLinkedEdge>::iterator p_edge2 = edge_list.begin(); p_edge2 <= p_edge; p_edge2 ++) {
					if((*p_edge2).p_next_edge_list)
						delete (*p_edge2).p_next_edge_list;
				}
				return 0;
			}
		}
		// add all neighbour edges
	}
	// create lists of neighbour edges

	std::vector<int> *p_index_list;
	std::vector<int> outline_vertex_num_list;

	if(!(p_index_list = new std::vector<int>)) {
		for(std::vector<TLinkedEdge>::iterator p_edge = edge_list.begin();
		   p_edge < edge_list.end(); p_edge ++) {
			if((*p_edge).p_next_edge_list)
				delete (*p_edge).p_next_edge_list;
		}
		return 0;
	}

	{int n_index = 0;
	for(std::vector<TVertex*>::iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = n_index ++;
	}
	// vertex flags will contain vertex indices

	for(std::vector<TLinkedEdge>::iterator p_edge = edge_list.begin();
	   p_edge < edge_list.end(); p_edge ++) {
		if((*p_edge).n_group != -1)
			continue;

		TLinkedEdge *p_cur_edge, *p_first_edge;
		p_first_edge = &(*p_edge);
		p_cur_edge = p_first_edge;
		int n_vertex_num = p_index_list->size();

		p_index_list->push_back(p_cur_edge->p_vertex[0]->n_flag);
		p_index_list->push_back(p_cur_edge->p_vertex[1]->n_flag);
		if(p_index_list->size() < 2) {
			for(std::vector<TLinkedEdge>::iterator p_edge2 = edge_list.begin();
			   p_edge2 < edge_list.end(); p_edge2 ++) {
				if((*p_edge2).p_next_edge_list)
					delete (*p_edge2).p_next_edge_list;
			}
			delete p_index_list;
			return 0;
		}

		do {
			p_cur_edge->n_group = outline_vertex_num_list.size();

			if(!p_cur_edge->p_next_edge_list || !p_cur_edge->p_next_edge_list->size())
				break;

			TLinkedEdge *p_next_edge = 0;
			for(int j = 0; j < p_cur_edge->p_next_edge_list->size(); j ++) {
				if((*p_cur_edge->p_next_edge_list)[j] == p_first_edge) {
					p_next_edge = (*p_cur_edge->p_next_edge_list)[j];
					break;
				} else if((*p_cur_edge->p_next_edge_list)[j]->n_group == -1)
					p_next_edge = (*p_cur_edge->p_next_edge_list)[j];
			}
			// see if we can close the circle, otherwise choose next edge

			if(!p_next_edge)
				break;
			// can't find next edge, it's propably malformed surface / there are some pecission
			// issues - todo : be silent or export some reference variable with warning codes?

			if(p_cur_edge->p_vertex[1] != p_next_edge->p_vertex[0] /*&& !b_ignore_double_vertices*/)
				p_index_list->push_back(p_next_edge->p_vertex[0]->n_flag);
			if(p_next_edge != p_first_edge)
				p_index_list->push_back(p_next_edge->p_vertex[1]->n_flag);
			//
			if(p_index_list->size() < 1 + (p_next_edge != p_first_edge)) {
				for(std::vector<TLinkedEdge>::iterator p_edge2 = edge_list.begin();
				   p_edge2 < edge_list.end(); p_edge2 ++) {
					if((*p_edge2).p_next_edge_list)
						delete (*p_edge2).p_next_edge_list;
				}
				delete p_index_list;
				return 0;
			}
			// add one or two more vertices

			p_cur_edge = p_next_edge;
		} while(p_cur_edge != p_first_edge);

		p_index_list->erase(&p_index_list->back());
		n_vertex_num = p_index_list->size() - n_vertex_num;
		// calc number of added vertices

		outline_vertex_num_list.push_back(n_vertex_num);
		if(!outline_vertex_num_list.size()) {
			for(std::vector<TLinkedEdge>::iterator p_edge2 = edge_list.begin();
			   p_edge2 < edge_list.end(); p_edge2 ++) {
				if((*p_edge2).p_next_edge_list)
					delete (*p_edge2).p_next_edge_list;
			}
			delete p_index_list;
			return 0;
		}
		// add number of vertices for this loop into 
	}
	// create groups, forming (closed) loops and vector with numbers of their vertices

	for(std::vector<int>::iterator p_a_num = outline_vertex_num_list.begin(),
	   p_b_num = outline_vertex_num_list.begin() + 1; p_b_num < outline_vertex_num_list.end();
	   p_a_num = p_b_num ++) {
		// want to create double edge between vertex from first list
		// that means to re-order every group so part lies in the beginning
		// of list and the end lies on the end of the list

		_ASSERTE(*p_a_num);
		_ASSERTE(*p_b_num);

		int n_chosen_a = 0, n_chosen_b = 0;
		float f_min_dist2 = (m_vertex_list[(*p_index_list)[0]]->v_pos -
			m_vertex_list[(*p_index_list)[*p_a_num]]->v_pos).f_Length2();

		for(std::vector<int>::const_iterator p_index_a = p_index_list->begin(),
		   p_end_a = p_index_list->begin() + (*p_a_num); p_index_a < p_end_a; p_index_a ++) {
			for(std::vector<int>::const_iterator p_index_b = p_index_list->begin() + (*p_a_num),
			   p_end_a = p_index_list->begin() + (*p_a_num) + (*p_b_num); p_index_a < p_end_a; p_index_a ++) {
				float f_dist2 = (m_vertex_list[*p_index_a]->v_pos -
					m_vertex_list[*p_index_b]->v_pos).f_Length2();

				if(f_dist2 < f_min_dist2) {
					f_min_dist2 = f_dist2;

					n_chosen_a = p_index_a - p_index_list->begin();
					n_chosen_b = p_index_b - /*(*/p_index_list->begin() + (*p_a_num)/*)*/;
					// fixme - works ok without parenthesis, right? (save 1 mul ((*p_a_num) * sizeof(int)))

					if(f_min_dist2 < f_epsilon) {
						p_index_a = p_end_a; // want to quit the other loop right away
						break;
					}
				}
			}
		}
		// find two closest vertices in either loops

		for(std::vector<int>::iterator p_a = p_index_list->begin() + ((*p_a_num) - 1),
		   p_b = p_index_list->begin() + ((*p_a_num) - 1 + (*p_b_num)),
		   p_end = p_index_list->begin() + n_chosen_a; p_a > p_end; p_a --, p_b --) {
			int n_temp = *p_a;
			*p_a = *p_b;
			*p_b = n_temp;
		}
		p_index_list->insert(&(*p_index_list)[n_chosen_a + 1 + (*p_b_num)],
			(*p_index_list)[n_chosen_a]);
		if(!p_index_list->size()) {
			for(std::vector<TLinkedEdge>::iterator p_edge = edge_list.begin();
			   p_edge < edge_list.end(); p_edge ++) {
				if((*p_edge).p_next_edge_list)
					delete (*p_edge).p_next_edge_list;
			}
			delete p_index_list;
			return 0;
		}
		// rotate first loop and duplicate point where double edge begins

		n_chosen_b = (n_chosen_b + ((*p_a_num) - 1 - n_chosen_a)) % (*p_b_num);
		// b was rotated as well, we need to recalc chosen vertex index

		if(n_chosen_b < (*p_b_num) / 2) {
			// chosen b is in bottom half of index array, it's shorter to rotate
			// left and duplicate right

			p_index_list->insert(&(*p_index_list)[n_chosen_a + (*p_b_num) + 1],
				(*p_index_list)[n_chosen_a + 1 + n_chosen_b]);
			if(!p_index_list->size()) {
				for(std::vector<TLinkedEdge>::iterator p_edge = edge_list.begin();
				   p_edge < edge_list.end(); p_edge ++) {
					if((*p_edge).p_next_edge_list)
						delete (*p_edge).p_next_edge_list;
				}
				delete p_index_list;
				return 0;
			}
			// insert last b

			Rotate_Insit_Left(&(*p_index_list)[n_chosen_a + 1], n_chosen_b, (*p_b_num));
			// rotate left so n_chosen_b is first
		} else {
			// chosen b is in top half of index array, it's shorter to rotate
			// right and duplicate left

			Rotate_Insit_Right(&(*p_index_list)[n_chosen_a + 1],
				(*p_b_num) - 1 - n_chosen_b, (*p_b_num));
			// rotate right so n_chosen_b is last

			int n_what = (*p_index_list)[n_chosen_a + (*p_b_num)]; // for some reason doesn't work as function parameter (there's no lazy evaluation in c++, right?)
			p_index_list->insert(&(*p_index_list)[n_chosen_a + 1], n_what);
			if(!p_index_list->size()) {
				for(std::vector<TLinkedEdge>::iterator p_edge = edge_list.begin();
				   p_edge < edge_list.end(); p_edge ++) {
					if((*p_edge).p_next_edge_list)
						delete (*p_edge).p_next_edge_list;
				}
				delete p_index_list;
				return 0;
			}
			// insert first b
		}

		(*p_b_num) += (*p_a_num) + 2;
	}
	// now we need to add final double edges to connect our closed loops

	for(std::vector<TLinkedEdge>::iterator p_edge = edge_list.begin();
	   p_edge < edge_list.end(); p_edge ++) {
		if((*p_edge).p_next_edge_list)
			delete (*p_edge).p_next_edge_list;
	}

	return p_index_list;
}

/*
 *	unsigned int **CGeometry::p_IndexList(bool b_single_memblock)
 *		- return list of vertex indicex arrays for all polygons in geometry
 *		  (as long as number of polygons)
 *		- every i-th index array has as much items as i-th polygon vertices
 *		- if b_single_memblock is true, all the indices are stored in a single long memory
 *		  block, beginning at p_IndexList(true)[0] ... so call free once only; primary
 *		  use is int **pp = p_IndexList(true)[0]; glDrawArrays(GL_TRIANGLES, pp[0]);
 *		  (don't forget to call free(pp[0]); free(pp); after and to call MakeTris(); before)
 */
unsigned int **CGeometry::p_IndexList(bool b_single_memblock)
{
	unsigned int **p_index_list;
	if(!(p_index_list = new unsigned int*[m_poly_list.size()]))
		return 0;
	// alloc new pointer list

	int n_vertex_reference_num = 0;
	{int n_index = 0;
	for(std::vector<TPolygon>::const_iterator p_poly = m_poly_list.begin();
	   p_poly < m_poly_list.end(); p_poly ++, n_index ++) {
		n_vertex_reference_num += (*p_poly).p_poly->n_Vertex_Num();
		if(!b_single_memblock) {
			// either as a tiny block for every poly
			if(!(p_index_list[n_index] = new unsigned int[(*p_poly).p_poly->n_Vertex_Num()])) {
				for(unsigned int **p_list = p_index_list,
				   **p_end = p_index_list + n_index; p_list < p_end;)
					delete[] *p_list ++;
				delete p_index_list;
				return 0;
			}
		}
	}}
	if(b_single_memblock && m_poly_list.size()) {
		// or as a single long block of memory (usable for glDrawArrays(GL_TRIANGLES, ...))
		if(!(p_index_list[0] = new unsigned int[n_vertex_reference_num])) {
			delete p_index_list;
			return 0;
		}
		n_vertex_reference_num = 0;
		{int n_index = 0;
		for(std::vector<TPolygon>::const_iterator p_poly = m_poly_list.begin();
		   p_poly < m_poly_list.end(); p_poly ++, n_index ++) {
			p_index_list[n_index] = p_index_list[0] + n_vertex_reference_num;
			n_vertex_reference_num += (*p_poly).p_poly->n_Vertex_Num();
		}}
	}
	// alloc lists for polygon indices

	{int n_index = 0;
	for(std::vector<TVertex*>::const_iterator p_vertex = m_vertex_list.begin();
	   p_vertex < m_vertex_list.end();)
		(*(p_vertex ++))->n_flag = n_index ++;
	}
	// set vertex flags to vertex indices

	{int n_index = 0;
	for(std::vector<TPolygon>::const_iterator p_poly = m_poly_list.begin();
	   p_poly < m_poly_list.end(); p_poly ++, n_index ++) {
		unsigned int *p_poly_index = p_index_list[n_index];
		for(int n = 0; n < (*p_poly).p_poly->n_Vertex_Num(); n ++)
			*p_poly_index ++ = (*p_poly).p_poly->t_Vertex(n).n_flag;
		n_vertex_reference_num -= (*p_poly).p_poly->n_Vertex_Num();
	}}
	_ASSERTE(!n_vertex_reference_num);
	// fill polygon list array

	return p_index_list;
}

/*
 *	bool CGeometry::operator =(const CGeometry &r_geometry)
 *		- copy geometry
 *		- return false if didn't have enough memory
 */
bool CGeometry::operator =(const CGeometry &r_geometry)
{
	Delete();
	return (*this += r_geometry);
}

/*
 *	bool CGeometry::operator +=(const CGeometry &r_geometry)
 *		- add geometry into this one
 *		- return false if didn't have enough memory
 */
bool CGeometry::operator +=(const CGeometry &r_geometry)
{
	int n_vertex_index_offset = m_vertex_list.size();
	// index of first vertex from r_geometry in this->m_vertex_list

	for(std::vector<TVertex*>::const_iterator p_vertex = r_geometry.m_vertex_list.begin();
	   p_vertex < r_geometry.m_vertex_list.end(); p_vertex ++) {
		if(!p_AddVertex(**p_vertex))
			return false;
	}
	// add all vertices of r_geometry

	int n_surface_num = 0;
	// number of smoothing surfaces in r_geometry

	unsigned int **p_index_array = ((CGeometry&)r_geometry).p_IndexList(true);
	if(!p_index_array)
		return false;

	{int n_index = 0;
	for(std::vector<TPolygon>::const_iterator p_polygon = r_geometry.m_poly_list.begin();
	   p_polygon < r_geometry.m_poly_list.end(); p_polygon ++, n_index ++) {
		TPolygon t_polygon = *p_polygon;

		t_polygon.n_neighbour_edge_num = 0;
		t_polygon.p_neighbour_index_list = 0;

		if(t_polygon.n_surface > n_surface_num)
			n_surface_num = t_polygon.n_surface;
		t_polygon.n_surface += m_n_surface_space;

		t_polygon.p_poly = new CRefPolygon;
		if(!(*t_polygon.p_poly = *(*p_polygon).p_poly)) {
			if(r_geometry.m_poly_list.size())
				delete[] *p_index_array; // p_IndexList(true) ... single memblock
			delete[] p_index_array;
			m_n_surface_space += n_surface_num;
			// want to keep it a little consistent
			return false;
		}
		for(int i = 0; i < t_polygon.p_poly->n_Vertex_Num(); i ++) {
			t_polygon.p_poly->r_p_Vertex(i) = m_vertex_list[n_vertex_index_offset +
				p_index_array[n_index][i]];
			t_polygon.p_poly->SetVertexGeometryPointer(i, this);
		}
		// copy polygons, replace vertex references so they're pointing to this->m_vertex_list

		m_b_poly_hash_dirty_flag = true;
		m_b_vertex_normal_smooth_group_dirty_flag = true;
		m_b_vertex_normal_smooth_all_dirty_flag = true;
		m_b_poly_neighbours_dirty_flag = true;
		// invalidate hash tables

		m_poly_list.push_back(t_polygon);
		if(!m_poly_list.size()) {
			delete t_polygon.p_poly;
			if(r_geometry.m_poly_list.size())
				delete[] *p_index_array; // p_IndexList(true) ... single memblock
			delete[] p_index_array;
			m_n_surface_space += n_surface_num;
			// want to keep it a little consistent
			return false;
		}
	}}

	if(r_geometry.m_poly_list.size())
		delete[] p_index_array[0];
	delete[] p_index_array;
	// free index list

	m_n_surface_space += n_surface_num + 1;
	// increment surface counter

	return true;
}

/*
 *								=== ~CGeometry ===
 */
